! GENCLDIRS -- Generate initial class directories
! Nils M Holm, 1999,2000,2002
! See the file LICENSE for conditions of use.
!
! USAGE: GENCLDIRS <CLASSDIR

module gencldirs(t3x);

object	t[t3x];

!---- From TXTRN ------------------------------------------------------!

struct	SYMREC =
	SNAME,
	SVAL,
	SSIZE,
	SFLAGS;

struct SR2 =		! XSYMREC
	SNL, SNH,
	SVL, SVH,
	SZL, SZH,
	SFL, SFH;

const	CNST=	1,	! Symbol Flags
	GLOBL=	2,
	PROC=	4,
	PROTO=	8,
	SYS=	16,
	IPROC=	32,
	PUBLC=	64,
	EXTRN=	128,
	CLSS=	256,
	OBJCT=	512,
	IVAR=	1024,
	REDIR=	2048,
	LINKD=	4096;

const	CACHESPACE =	4096,
	NAMESPACE =	8192;

!----------------------------------------------------------------------!

var	Symbols[CACHESPACE];
var	Names::NAMESPACE+1;

length(s) do var k;
	k := 0;
	while (s::k) k := k+1;
	return k;
end

writep(f, s) t.write(T3X.SYSOUT, s, length(s));

wrout(s) writep(T3X.SYSOUT, s);

nl(f) do var b::3;
	writep(f, t.newline(b));
end

error(m) do
	writep(T3X.SYSERR, "GENCLDIR: ");
	writep(T3X.SYSERR, m);
	nl(T3X.SYSERR);
	halt 1;
end

writedir(tbl, n) do
	var	k, i, dir, dirname::256, name, s, sfx;
	var	fwerr;
	var	nb::2;

	fwerr := "file write error";
	s := @Names::(tbl::SNH << 8 | tbl::SNL);
	t.memcopy(dirname, s, length(s)+1);
	k := 0;
	while (dirname::k) k := k+1;
	sfx := ".dir";
	for (i=0, 5) dirname::(k+i) := sfx::i;
	name := dirname;
	wrout(name);
	nl(T3X.SYSOUT);
	dir := t.open(name, T3X.OWRITE);
	if (dir < 0) error("file creation failed");
	if (t.write(dir, "CDIR", 4) \= 4) error(fwerr);
	nb::0 := n & 255;
	nb::1 := n >> 8;
	if (t.write(dir, nb, 2) \= 2) error(fwerr);
	if (t.write(dir, tbl, n*SR2) \= n*SR2) error(fwerr);
	for (i=0, n*SR2, SR2) do
		s := @Names::(tbl::(i+SNH) << 8 | tbl::(i+SNL));
		k := 0;
		while (s::k) k := k+1;
		k := k+1;
		if (t.write(dir, s, k) \= k) error(fwerr);
	end
	t.close(dir);
end

gencldirs(n) do var i, j, f;
	for (i=0, n*SR2, SR2) do
		f := Symbols::(i+SFH) << 8 | Symbols::(i+SFL);
		if (f & CLSS) do
			j := i;
			while (1) do
				if (	\(Symbols::(j+SFH) << 8 |
					  Symbols::(j+SFL))
				)
					leave;
				if (j >= n*SR2)
					error("missing end of class");
				j := j+SR2;
			end
			writedir(@Symbols::i, (j+SR2-i)/SR2);
		end
	end
end

do
	var hd::4, i, n, p, nb::2;

	if (t.read(0, hd, 4) \= 4 \/ t.memcomp(hd, "CDIR", 4))
		error("magic match failed");
	if (t.read(0, nb, 2) \= 2) error("short file");
	n := nb::1 << 8 | nb::0;
	if (t.read(0, Symbols, n*SR2) \= n*SR2)
 		error("short file");
	if (t.read(0, Names, NAMESPACE+1) = NAMESPACE+1)
		error("file too long");
	p := 0;
	for (i=0, n*SR2, SR2) do
		Symbols::(i+SNL) := p & 255;
		Symbols::(i+SNH) := p >> 8;
		p := p+length(@Names::p)+1;
	end
	gencldirs(n);
end
