! T3X Runtime Class: I/O-Streams
! Nils M Holm, 1997,1998,2000,2002,2019
! See the file LICENSE for conditions of use.

module iostream();

public class iostream(t3x)

 object	t[t3x];

 var	Filedesc,	! descriptor of associated file / device
	Buffer,		! I/O buffer
	Length,		! length of I/O buffer
	Flags,		! state, inclusing user supplied flags
	Ptr,		! pointer to next char (read) / free space (write)
	Lim;		! number of characters in buffer

 ! mode/state flags
 public const	FREAD	= 0x0001,	! read-only
		FWRITE	= 0x0002,	! write-only
		FRDWR	= FREAD|FWRITE,	! read/write
		FKILLCR	= 0x0004,	! remove CRs from input
		FADDCR	= 0x0008,	! add CR before LF in output
		FTRANS	= FKILLCR|FADDCR;
 const		FEOF	= 0x0100,	! EOF detected
		FLASTW	= 0x0200;	! last access was of type 'write'

 ! seek modes
 public const	SEEK_SET = T3X.SEEK_SET,	! from beginning
		SEEK_FWD = T3X.SEEK_FWD,	! relative/forward
		SEEK_END = T3X.SEEK_END,	! from end (back)
		SEEK_BCK = T3X.SEEK_BCK;	! relative/back

 public create(fd, bufp, len, mode) do
	Filedesc := fd;
	Buffer := bufp;
	Flags := mode;
	Length := len;
	Ptr := 0;
	Lim := 0;
 end

 public open(name, bufp, len, fl) do var fd, mode, ifl;
	ifl := fl & FRDWR;
	mode := ifl = FREAD-> T3X.OREAD:
		ifl = FWRITE-> T3X.OWRITE:
		ifl = FREAD|FWRITE-> T3X.ORDWR: %1;
	if (mode < 0) return %1;
	fd := t.open(name, mode);
	if (fd < 0) return %1;
	return self.create(fd, bufp, len, fl);
 end

 public flush() do var k;
	if (Flags & (FWRITE|FLASTW) = FWRITE|FLASTW /\ Ptr) do
		k := t.write(Filedesc, Buffer, Ptr);
		if (k \= Ptr) return %1;
	end
	Ptr := 0;
	Lim := 0;
 end

 public close() do
	if (self.flush() = %1) return %1;
	t.close(Filedesc);
	Flags := 0;
 end

 public wrch(ch) do
	Flags := Flags | FLASTW;
	if (Flags & FADDCR /\ ch = '\n') self.wrch('\r');
	if (Ptr >= Length /\ self.flush() = %1) return %1;
	Buffer::Ptr := ch;
	Ptr := Ptr+1;
	return ch;
 end

 public write(bufp, len) do var i;
	Flags := Flags | FLASTW;
	i := 0;
	while (i < len) do
		if (Ptr >= Length /\ self.flush() = %1) return %1;
		if (Flags & FADDCR) do
			if (bufp::i = '\n') do
				Buffer::Ptr := '\r';
				Ptr := Ptr+1;
				if (Ptr >= Length /\ self.flush() = %1)
					return %1;
			end
		end
		Buffer::Ptr := bufp::i;
		Ptr := Ptr+1;
		i := i+1;
	end
	return i;
 end

 public writes(str) do var k;
	k := t.memscan(str, 0, 32767);
	if (k .> 32766) return %1;
	return self.write(str, k);
 end

 more() do var k;
	if (Flags & FREAD) do
		k := t.read(Filedesc, Buffer, Length);
		if (k <= 0) do
			Flags := Flags | FEOF;
			if (k < 0) return %1;
		end
		Ptr := 0;
		Lim := k;
		return k;
	end
 end

 public rdch() do var c;
	Flags := Flags & ~FLASTW;
	while (1) do
		if (Flags & FEOF) return %1;
		if (Ptr >= Lim /\ more() < 1) return %1;
		c := Buffer::Ptr;
		Ptr := Ptr+1;
		ie (Flags & FKILLCR)
			if (c \= '\r') leave;
		else
			leave;
	end
	return c;
 end

 _read(bufp, len, ckln) do var i;
	Flags := Flags & ~FLASTW;
	i := 0;
	while (i < len) do
		if (Ptr >= Lim /\ more() < 1) leave;
		ie (Flags & FKILLCR) do
			if (Buffer::Ptr \= '\r') do
				bufp::i := Buffer::Ptr;
				i := i+1;
			end
		end
		else do
			bufp::i := Buffer::Ptr;
			i := i+1;
		end
		Ptr := Ptr+1;
		if (ckln /\ bufp::(i-1) = '\n') leave;
	end
	if (ckln) bufp::i := 0;
	return i;
 end

 public read(bufp, len) return _read(bufp, len, 0);

 public reads(bufp, len) return _read(bufp, len, 1);

 public move(off, how) do var delta;
	ie (Flags & FLASTW) do
		if (self.flush() = %1) return %1;
	end
	else ie (how = SEEK_FWD) do
		delta := Lim - Ptr;
		if (self.flush() = %1) return %1;
		off := off - delta;
	end
	else ie (how = SEEK_BCK) do
		delta := Lim - Ptr;
		if (self.flush() = %1) return %1;
		off := off + delta;
	end
	else do
		if (self.flush() = %1) return %1;
	end
	return t.seek(Filedesc, off, how);
 end

 public eof() return (Flags & FEOF) \= 0;

 public reset() Flags := Flags & ~FEOF;

end

do end
