! T3X Runtime Class: String Functions
! Nils M Holm, 1999-2002,2019
! See the file LICENSE for conditions of use.

module string();

public class string(t3x)

 object	t[t3x];

 public const	MAXLEN = 32767;

 public length(s) return t.memscan(s, 0, MAXLEN);

 public copy(a, b) t.memcopy(a, b, t.memscan(b, 0, MAXLEN)+1);

 public ncopy(n, a, b) do var k;
	if (n < 0) return;
	k := t.memscan(b, 0, MAXLEN) + 1;
	t.memcopy(a, b, n < k-> n: k);
	a::(n-1) := 0;
 end

 public comp(a, b) do var k1, k2;
	k1 := t.memscan(a, 0, MAXLEN)+1;
	k2 := t.memscan(b, 0, MAXLEN)+1;
	return t.memcomp(a, b, k1>k2->k1:k2);
 end

 public scan(s, c) return t.memscan(s, c, t.memscan(s, 0, MAXLEN));

 public rscan(s, c) do var k, i;
	k := t.memscan(s, 0, MAXLEN);
	for (i=k-1, %1, %1) if (s::i = c) return i;
	return %1;
 end

 public find(a, b) do var ka, kb, i;
	ka := t.memscan(a, 0, MAXLEN);
	kb := t.memscan(b, 0, MAXLEN);
	for (i=0, ka-kb+1)
		if (\t.memcomp(@a::i, b, kb))
			return i;
	return %1;
 end

 public xlate(s, c1, c2) do var i;
	i := 0;
	while (s::i) do
		if (s::i = c1) s::i := c2;
		i := i+1;
	end
	return s;
 end

 _digits() return "0123456789ABCDEF";

 public numtostr(s, n, r) do
	var	digits;
 	var	i;
	var	ds::34, dp;

	digits := _digits();
	i := 0;
	! handle sign
	if (r < 0) do
		r := -r;
		if (n < 0) do
			n := -n;
			s::i := '-';
			i := i+1;
		end
	end
	dp := 0;
	! generate reverse numeric string
	while (1) do
		ds::dp := digits::(n mod r);
		dp := dp+1;
		n := n ./ r;
		if (\n) leave;
	end
	! reverse the string
	while (dp) do
		dp := dp-1;
		s::i := ds::dp;
		i := i+1;
	end
	s::i := 0;
	return s;
 end

 ! Template format:
 ! %[max][:F][U][LR][CDSX%]
 ! Fillchar
 ! Unsigned
 ! Left, Right
 ! Character, Decimal, String, heXadecimal, %

 public nformat(n, s, tmpl, args) do
	const	LEFT=-1, NONE=0, RIGHT=1;	! adjust modes
	const	MAXPAD = 256;	! max. padding characters
	var	adj;		! adjustment type, see above
	var	op, ap;
	var	i, j, k;
	var	max;		! max. field size
	var	p, x;
	var	unsigned;	! unsigned flag for %D (1=unsigned)
	var	nbuf::MAXPAD+1;
	var	fc;		! fill character (for padding)

	i := 0;		! template index
	j := 0;		! buffer index
	ap := 0;	! argument index
	while (tmpl::i) do
		ie (tmpl::i = '%') do
			! set defaults
			adj := NONE;
			max := -1;	! variable length
			unsigned := 0;
			fc := '\s';
			i := i+1;
			! accept max field length
			while ('0' <= tmpl::i /\ tmpl::i <= '9') do
				if (max = -1) max := 0;
				max := max*10 + tmpl::i - '0';
				i := i+1;
			end
			! accept fill character
			if (tmpl::i = ':') do
				fc := tmpl::(i+1);
				i := i+2;
			end
			! accept unsigned flag
			if (tmpl::i = 'u' \/ tmpl::i = 'U') do
				unsigned := 1;
				i := i+1;
			end
			! accept adjustment type
			ie (tmpl::i = 'l' \/ tmpl::i = 'L') do
				adj := LEFT;
				i := i+1;
			end
			else if (tmpl::i = 'r' \/ tmpl::i = 'R') do
				adj := RIGHT;
				i := i+1;
			end
			x := tmpl::i;	! get type
			if ('a' <= x /\ x <= 'z') x := x+('A'-'a');
			op := ap;
			ie (x = 'C') do
				nbuf::0 := args[ap];
				nbuf::1 := 0;
				p := nbuf;
				ap := ap+1;
			end
			else ie (x = 'D') do
				p := self.numtostr(nbuf, args[ap],
					unsigned-> 10: -10);
				if (adj = NONE) adj := RIGHT;
				ap := ap+1;
			end
			else ie (x = 'S') do
				p := args[ap];
				ap := ap+1;
				if (adj = NONE) adj := LEFT;
			end
			else ie (x = 'X') do
				p := self.numtostr(nbuf, args[ap],
					unsigned-> 16: -16);
				ap := ap+1;
				if (adj = NONE) adj := RIGHT;
			end
			! default: copy character after %
			else if (j .< n-1) do
				s::j := tmpl::i;
				j := j+1;
			end
			if (op \= ap) do	! op=ap means invalid type
				k := self.length(p);
				ie (max <= k) do
					! no padding required
					if (j+k .>= n-1) k := n-j-1;
					if (k) t.memcopy(@s::j, p, k);
					j := j+k;
				end
				else do
					if (j+max .>= n-1) max := n-j-1;
					if (max) t.memfill(@s::j, fc, max);
					if (j+k .>= n-1) k := n-j-1;
					ie (k /\ adj = LEFT)
						t.memcopy(@s::j, p, k);
					else if (k)
						t.memcopy(@s::(j+max-k), p, k);
					j := j+max;
				end
			end
		end
		else if (j .< n-1) do
			! no %-sequence: copy one char
			s::j := tmpl::i;
			j := j+1;
		end
		if (tmpl::i) i := i+1;
	end
	s::j := 0;
	return s;
 end

 public format(s, tmpl, args) return self.nformat(~0, s, tmpl, args);

 public strtonum(s, r, lp) do
	var	digits;
	var	g, i, j;
	var	v;
	var	lim;

	digits := _digits();
	r := r mod 17;
	i := 0;
	v := 0;
	g := 1;
	! skip leading white space
	while (s::i = '\s' \/ s::i = '\t') i := i+1;
	! ignore leading +
	ie (s::i = '+') do
		i := i+1;
	end
	! accept leading - or %
	else if (s::i = '-' \/ s::i = '%') do
		g := -1;
		i := i+1;
	end
	! if a limit is supplied, use it
	lim := lp /\ lp[0]-> lp[0]: MAXLEN;
	for (i=i, lim) do
		j := t.memscan(digits, s::i, r);
		if (j < 0) leave;
		v := v*r + j;
	end
	! if a limit was supplied, replace it with
	! the number of characters parsed
	if (lp) lp[0] := i;
	return v*g;
 end

 ! Template format:
 ! %[len][:delimiter][CDSWX%]
 ! Unsigned
 ! Character, Decimal, String, White, heXadecimal, %
 ! %[group] (literal []s)
 ! %W = %[\s\t]

 public parse(s, tmpl, args) do
	var	i, j, k, x;
	var	n, m;
	var	ap;
	var	len,		! max. length to parse
		dlm;		! delimiter
	var	cbuf::256;

	i := 0;		! template index
	j := 0;		! string (source) index
	k := self.length(s);
	ap := 0;	! argument index
	while (tmpl::i /\ j < k) do
		ie (tmpl::i = '%') do
			i := i+1;
			len := -1;	! no limit on length
			dlm := -1;	! no delimiter
			! accept length
			while ('0' <= tmpl::i /\ tmpl::i <= '9') do
				if (len = -1) len := 0;
				len := len*10 + tmpl::i - '0';
				i := i+1;
			end
			! accept delimiter
			if (tmpl::i = ':') do
				dlm := tmpl::(i+1);
				i := i+2;
			end
			x := tmpl::i;	! get type
			if ('a' <= x /\ x <= 'z') x := x+('A'-'a');
			ie (x = 'C') do
				args[ap][0] := s::j;
				ap := ap+1;
				j := j+1;
			end
			else ie (x = 'D') do
				n := len=-1-> 0: len;
				args[ap][0] := self.strtonum(@s::j, 10, @n);
				ap := ap+1;
				j := j+n;
			end
			else ie (x = 'S') do
				if (len = -1) do
					! no limit: copy rest of source
					if (dlm = -1) dlm := tmpl::(i+1);
					len := t.memscan(@s::j, dlm, k-j+1);
					If (len < 1) len := k-j+1;
				end
				if (len > k-j+1) len := k-j+1;
				t.memcopy(args[ap], @s::j, len);
				args[ap]::len := 0;
				ap := ap+1;
				j := j+len;
			end
			else ie (x = 'X') do
				n := len=-1-> 0: len;
				args[ap][0] := self.strtonum(@s::j, 16, @n);
				ap := ap+1;
				j := j+n;
			end
			else ie (x = 'W' \/ x = '[') do
				ie (x = 'W') do
					cbuf::0 := '\s';
					cbuf::1 := '\t';
					m := 2;
				end
				else do
					m := 0;
					i := i+1;
					while (tmpl::i /\ tmpl::i \= ']') do
						cbuf::m := tmpl::i;
						i := i+1;
						m := m+1;
					end
				end
				while (	j < k /\
					t.memscan(cbuf, s::j, m) >= 0
				)
					j := j+1;
			end
			else do
				! default: match character after %
				if (tmpl::i \= s::j) leave;
				j := j+1;
			end
		end
		else do
			! default: match literal character
			if (tmpl::i \= s::j) leave;
			j := j+1;
		end
		if (tmpl::i) i := i+1;
	end
	return ap;
 end

end

do end
