! MkTcode -- Generate Tcode Instruction Sets
! Nils M Holm, 2003
! See the file LICENSE for conditions of use.

module mktcode(t3x, iostream, string);

object	t[t3x], str[string];

const	IOBUFLEN = 1024;
object	In[iostream], Out1[iostream], Out2[iostream],
	Out3[iostream], Out4[iostream];
var	Inbuf::IOBUFLEN, Outbuf1::IOBUFLEN, Outbuf2::IOBUFLEN,
	Outbuf3::IOBUFLEN, Outbuf4::IOBUFLEN;
var	Lno;
var	Nlchar::3;
var	Inames[128], Ipool::1024, Ip, Pp;

error(msg) do var buf::128, nl::3;
	t.write(T3X.SYSERR, str.format(buf, "MKTCODE: %D: %S%S",
		[(Lno, msg, t.newline(nl))]), str.length(buf));
	halt 1;
end

init() do
	if (In.open("tcode.def", Inbuf, IOBUFLEN, iostream.FREAD) = %1)
		error("cannot open 'tcode.def'");
	if (Out1.open("tcode.t", Outbuf1, IOBUFLEN, iostream.FWRITE) = %1)
		error("cannot create 'tcode.t'");
	if (Out2.open("tcode.h", Outbuf2, IOBUFLEN, iostream.FWRITE) = %1)
		error("cannot create 'tcode.h'");
	if (Out3.open("rtcode.t", Outbuf3, IOBUFLEN, iostream.FWRITE) = %1)
		error("cannot create 'rtcode.t'");
	if (Out4.open("rtcode.h", Outbuf4, IOBUFLEN, iostream.FWRITE) = %1)
		error("cannot create 'rtcode.t'");
	Lno := 0;
	t.newline(Nlchar);
	Ip := 0;
	Pp := 0;
end

fini() do
	In.close();
	Out1.close();
	Out2.close();
	Out3.close();
	Out4.close();
end

split(line, args) do var p, q, n;
	p := 0;
	q := 0;
	n := 0;
	while (1) do
		if (line::p = 0 \/ line::p = '\t') do
			if (n > 3) error("too many arguments");
			args[n] := @line::q;
			n := n+1;
			if (line::p = 0) leave;
			line::p := 0;
			q := p+1;
		end
		p := p+1;
	end
	return n;
end

warning(ios, cs, ce) do var msg, i;
	msg := [
	"This file was automatically generated from 'tcode.def'.\t",
	"DO NOT REMOVE THIS FILE!\t\t\t\t\t",
	"A full T3X compiler is required to rebuild it.\t\t",
	0 ];
	i := 0; while(msg[i]) do
		send(ios, iostream, writes(cs));
		send(ios, iostream, writes(msg[i]));
		send(ios, iostream, writes(ce));
		send(ios, iostream, writes(Nlchar));
		i := i+1;
	end
end

locase(s) do var i;
	i := 0;
	while (s::i) do
		if ('A' <= s::i /\ s::i <= 'Z')
			s::i := s::i + 'a'-'A';
		i := i+1;
	end
end

generate() do
	var	line::128, buf::256;
	var	args[4], k, i;
	var	op;

	op := 0;
	warning(Out1, "! ", "!");
	warning(Out2, "/* ", " */");
	warning(Out3, "! ", "!");
	warning(Out4, "/* ", " */");
	Out1.writes(Nlchar);
	Out1.writes("module tcode();");
	Out1.writes(Nlchar);
	Out1.writes("public class tcode()");
	Out1.writes(Nlchar);
	Out1.writes("\t! The Tcode7 Instruction Set");
	Out1.writes(Nlchar);
	Out2.writes(Nlchar);
        Out2.writes("/* The Tcode7 Instruction Set */");
        Out2.writes(Nlchar);
	Out3.writes(Nlchar);
	Out3.writes("module rtcode();");
	Out3.writes(Nlchar);
	Out3.writes("public class rtcode()");
	Out3.writes(Nlchar);
	Out3.writes(" ! Tcode7 Instruction Name Lookup");
	Out3.writes(Nlchar);
	Out3.writes(" public lookup(i) do var names;");
	Out3.writes(Nlchar);
	Out3.writes("\tnames := [");
	Out3.writes(Nlchar);
	Out4.writes(Nlchar);
	Out4.writes("/* Tcode7 Instruction Name Lookup */");
	Out4.writes(Nlchar);
	Out4.writes("char\t*Tcode_Opcodes[] = {");
	Out4.writes(Nlchar);
	while (1) do
		k := In.reads(line, 128);
		if (In.eof() \/ k < 1) leave;
		Lno := Lno+1;
		line::(k-1) := 0;
		if (k > 1 /\ line::(k-2) = '\r') line::(k-2) := 0;
		if (line::0 = 0 \/ line::0 = '#') loop;
		k := split(line, args);
		if (k \= 4) error("wrong number of fields");
		Out1.writes(str.format(buf,
			"\tpublic const\tI%S\t= 0x%2:0X;\t! %S%S- %S%S",
				[(args[0]),
				 (args[1]::0 \= '0'-> op+0x80: op),
				 (args[2]::0 = '-'-> "": args[2]),
				 (args[2]::0 = '-'-> "": "\s"),
				 (args[3]),
				 (Nlchar)]));
		Out2.writes(str.format(buf,
			"#define\tI%S\t0x%2:0X\t/* %S%S- %S */%S",
				[(args[0]),
				 (args[1]::0 \= '0'-> op+0x80: op),
				 (args[2]::0 = '-'-> "": args[2]),
				 (args[2]::0 = '-'-> "": "\s"),
				 (args[3]),
				 (Nlchar)]));
		Out3.writes(str.format(buf,
			"\t\t\"%s\",%s", [(args[0], Nlchar)]));
		Out4.writes(str.format(buf,
			"\t\"%s\",%s", [(args[0], Nlchar)]));
		k := str.length(args[0])+1;
		locase(args[0]);
		t.memcopy(@Ipool::Pp, args[0], k);
		Inames[Ip] := @Ipool::Pp;
		Pp := Pp+k;
		Ip := Ip+1;
		op := op+1;
	end
	Out1.writes(str.format(buf,
		"\tpublic const\tIENDOFSET\s=\s0x%2:0X;%S", [(op),(Nlchar)]));
	Out2.writes(str.format(buf,
		"#define\tIENDOFSET\s0x%2:0X%S", [(op),(Nlchar)]));
	Out1.writes("end");
	Out1.writes(Nlchar);
	Out1.writes(Nlchar);
	Out1.writes("do end");
	Out1.writes(Nlchar);
	Out3.writes("\t0 ];");
	Out3.writes(Nlchar);
	Out3.writes(str.format(buf,
		"\tif (i.>255 \\/ i&127>=%d) return \"UNKNOWN\";", [(op)]));
	Out3.writes(Nlchar);
	Out3.writes("\treturn names[i&127];");
	Out3.writes(Nlchar);
	Out3.writes(" end");
	Out3.writes(Nlchar);
	Out3.writes("end");
	Out3.writes(Nlchar);
	Out3.writes("do end");
	Out3.writes(Nlchar);
	Out4.writes("(char*)0L};");
	Out4.writes(Nlchar);
	Out2.writes(Nlchar);
	for (i=0, Ip) do
		Out2.writes(str.format(buf, "void\ti%s();", [(Inames[i])]));
		Out2.writes(Nlchar);
	end
	Out2.writes(Nlchar);
	Out2.writes("void\t(*Tcode_instructions[])() = {");
	Out2.writes(Nlchar);
	for (i=0, Ip) do
		Out2.writes(str.format(buf, "\ti%s", [(Inames[i])]));
		if (i < Ip-1) Out2.writes(",");
		Out2.writes(Nlchar);
	end
	Out2.writes("};");
	Out2.writes(Nlchar);
end

do
	init();
	generate();
	fini();
end
