/*
 * TXX Interface Definitions
 * Nils M Holm 1997-2003,2019,2021
 * See the file LICENSE in the T3X documentation directory
 * ($T3XDIR/doc) for conditions of use.
 */

#define EXTERN extern

#ifdef MAIN
 #undef EXTERN
 #define EXTERN
#endif

#define uchar   unsigned char
#define ushort  unsigned short
#define uint    unsigned int
#define ulong   unsigned long

#ifdef __STDC__
 #ifndef COHERENT
  #define ANSIPROTO
 #endif
#else
 #ifdef __TURBOC__
  #define ANSIPROTO
 #endif
#endif

#ifdef __FreeBSD__
 #ifdef __i386__
  #define HOST32
  #define NL_IS_LF
 #else
  /* pre-3.0 FreeBSDs? */
  #ifdef i386
   #define HOST32
   #define NL_IS_LF
  #endif
  #ifdef __alpha__
   #define HOST64
   #define NL_IS_LF
  #endif
 #endif
#endif

#ifdef __NetBSD__
 #ifdef __alpha__
  #define HOST64
  #define NL_IS_LF
 #endif
 #ifdef i386
  #define HOST32
  #define NL_IS_LF
 #endif
 #ifdef mc68000
  #define HOST32
  #define NL_IS_LF
 #endif
#endif  

#ifdef __OpenBSD__
 #ifdef __i386__
  #define HOST32
  #define NL_IS_LF
 #endif
#endif  

#ifdef COHERENT
 #ifdef _I386
  #define HOST32
  #define NL_IS_LF
  #define NEED_RENAME
  #define void char
  #include <sys/types.h>
 #endif
#endif

#ifdef __linux__
 #ifdef __i386__
  #define HOST32
  #define NL_IS_LF
 #endif
#endif  

#ifdef dos
 #define HOST16
 #define NL_IS_CRLF
 #include <dos.h>
 #ifdef __TURBOC__
  #define off_t	ulong
  /* OOPS! open() and friends translate LF<-->CRLF */
  #define creat(X,Y)	_creat((X),0)
  #define open		_open
  #define close		_close
  #define read		_read
 #endif
#endif

/* If none of the above matches, choose one here. */
/* On x86-64 platforms use HOST32, HOST64 is for the alpha! */

/* #define HOST16 */
/* #define HOST32 */
/* #define HOST64 */

/* If none of the above matches, choose one here. */

/* #define NL_IS_LF */
/* #define NL_IS_CRLF */

#ifndef HOST16
 #ifndef HOST32
  #ifndef HOST64
   #error "Host machine word size undefined!"
  #endif
 #endif
#endif

#ifndef NL_IS_LF
 #ifndef NL_IS_CRLF
  #error "Newline sequence undefined!"
 #endif
#endif

#ifdef HOST16
 #define cell	short
 #define ucell	ushort
 #define BPW	2
 #define WORDCAST	(short)
 #ifdef __TURBOC__
  /* TurboC port of TXX uses large model */
  #define xcell		long
  #define xucell	ulong
 #else
  #define xcell		short
  #define xucell	ushort
 #endif
#endif

#ifdef HOST32
 #ifdef LIBRARY
  #define cell	int
  #define ucell	uint
  #define BPW	4
 #else
  #define cell	short
  #define ucell	ushort
  #define BPW	2
 #endif
 #define WORDCAST	(int)
 #define xcell	int
 #define xucell	uint
#endif

#ifdef HOST64
 #ifdef LIBRARY
  #define cell	long
  #define ucell	ulong
  #define BPW	8
 #else
  #define cell	short
  #define ucell	ushort
  #define BPW	2
 #endif
 #define WORDCAST	(long)
 #define xcell	long
 #define xucell	ulong
#endif

#ifndef LIBRARY
 #define ARG(C,N)	Stack[Sp+(C)-(N)]
 #define UARG(C,N)	Ustack[Sp+(C)-(N)]
 #define ARGP(C,N)	(char *) &Data[Ustack[Sp+(C)-(N)]]
 #define UARGP(C,N)	&Data[Ustack[Sp+(C)-(N)]]
#endif

#ifdef LIBRARY
 #define S0()			(__self)
 #define S1(a0)			(__self,a0)
 #define S2(a0,a1)		(__self,a1,a0)
 #define S3(a0,a1,a2)		(__self,a2,a1,a0)
 #define S4(a0,a1,a2,a3)	(__self,a3,a2,a1,a0)
 #define S5(a0,a1,a2,a3,a4)	(__self,a4,a3,a2,a1,a0)
 #define S6(a0,a1,a2,a3,a4,a5)	(__self,a5,a4,a3,a2,a1,a0)
 #define S7(a0,a1,a2,a3,a4,a5,a6)	\
				(__self,a6,a5,a4,a3,a2,a1,a0)
 #define SELF			xcell	__self;
#else
 #define S0()			()
 #define S1(a0)			(a0)
 #define S2(a0,a1)		(a1,a0)
 #define S3(a0,a1,a2)		(a2,a1,a0)
 #define S4(a0,a1,a2,a3)	(a3,a2,a1,a0)
 #define S5(a0,a1,a2,a3,a4)	(a4,a3,a2,a1,a0)
 #define S6(a0,a1,a2,a3,a4,a5)	(a5,a4,a3,a2,a1,a0)
 #define S7(a0,a1,a2,a3,a4,a5,a6)	\
				(a6,a5,a4,a3,a2,a1,a0)
 #define SELF
#endif

EXTERN ushort	*Ustack;
EXTERN short	*Stack;
EXTERN ushort	Sp;
EXTERN uchar	*Data;
EXTERN short	Acc;
EXTERN char	**Argvec;
EXTERN int	Argcount;

#define MAXINFPROC	64
struct infproc {
	char	*name;
	xcell	(*proc)();
};

#ifdef MAIN

void	libinit();

void	X_bpw(), X_close(), X_cvalist(), X_getarg(), X_getenv(),
	X_memcomp(), X_memcopy(), X_memfill(), X_memscan(),
	X_newline(), X_open(), X_read(), X_remove(),
	X_rename(),
	X_seek(), X_sys(), X_write();

void	(*Primitives[])() = {

	/*---TXX-PRIMITIVES-BEGIN---*/
	X_sys,
	X_bpw,
	X_memcopy,
	X_memcomp,
	X_memfill,
	X_memscan,
	X_getarg,
	X_getenv,
	X_open,
	X_read,
	X_write,
	X_newline,
	X_seek,
	X_close,
	X_rename,
	X_remove,
	X_cvalist,
	/*---TXX-PRIMITIVES-END---*/

	0
};

#endif	/* MAIN */
