! TXCGEXE -- Tcode->EXE translator
! Nils M Holm, 2000-2003,2019
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txcgexe(t3x, tcode, string, iostream);

object	t[t3x], str[string];

const	BPW =		2;

const	BUFSIZE	=	1024;	! I/O buffer size
const	MAXLABEL=	4096;	! max. label ID
const	MAXBSSREF=	512;	! max. data->BSS references
const	LIBMAX=		256;	! max space for library names
const	LIBPATHMAX=	1024;	! max space for library paths
const	DIRBUFSIZE=	1024;	! library directory buffer size
const	MAXINFPROC=	256;	! max. number of interface procedures
const	NAMEPOOLSIZE=	1024;	! interface procedure name pool size

struct INF =
	INF_ENTRY,
	INF_OFFSET,
	INF_NAME;

object	In[iostream], Out[iostream],
	Data[iostream];		! data segment stream
var	Rejected;
var	Inbuf::BUFSIZE;
var	Outbuf::BUFSIZE;
var	Databuf::BUFSIZE;
var	Addr;
var	Labels[MAXLABEL],
	Ltypes::MAXLABEL;	! label types: Text, Data, BSS
var	BSSrefs[MAXBSSREF], Bt;	! data->BSS references, top ptr
var	Ctop, Dtop, Btop, Bss;	! free address pointers, BSS start address
var	Firstpass;
var	Iqueue, Aqueue,		! 1-instruction queue (instr, arg)
	Queued,			! queue active indicator (1=active),
	Active,			! accumulator active flag (1=active),
	Depth;			! RT stack depth
var	Dataname;
var	K64k, KFFFF;		! 65536 constant
var	Tmpfile::256,		! tmp file name buffers
	Tmpdata::256;
var	Infprocs[INF*MAXINFPROC],	! interface procedures,
	Ninf;				! top ptr
var	Names::NAMEPOOLSIZE, Nt;	! interface name pool, top ptr
var	Infmap[MAXINFPROC];		! interface map
var	Verbose;

! variables to bind code sequences
var	LOAD_GLOB,
	LOAD_LOC,
	LOAD_IVAR,
	LOAD_LOC_VEC,
	LOAD_IVAR_VEC,
	LOAD_CONST,
	LOAD_SELF,
	PUSH_SEQ,
	POP_SEQ,
	DEREF_SEQ,
	DEREF_BYTE_SEQ,
	STORE_GLOB,
	STORE_LOC,
	STORE_IVAR,
	STORE_SEQ,
	STORE_BYTE,
	NORM_SEQ,
	INC_GLOB,
	INC_LOC,
	INC_IVAR,
	NEG_SEQ,
	NOT_SEQ,
	LNOT_SEQ,
	CMP_ZERO_SEQ,
	MUL_SEQ,
	DIV_SEQ,
	MOD_SEQ,
	UMUL_SEQ,
	UDIV_SEQ,
	ADD_SEQ,
	SUB_SEQ,
	AND_SEQ,
	OR_SEQ,
	XOR_SEQ,
	SHL_SEQ,
	SHR_SEQ,
	CMP_SEQ,
	SWAP_SEQ,
	BE_SEQ,
	BNE_SEQ,
	BL_SEQ,
	BG_SEQ,
	BLE_SEQ,
	BGE_SEQ,
	BUL_SEQ,
	BUG_SEQ,
	BULE_SEQ,
	BUGE_SEQ,
	JMP_SEQ,
	JMPS_SEQ,
	TRUE_SEQ,
	FALSE_SEQ,
	PROLOGUE_SEQ,
	EPILOGUE_SEQ,
	MPROLOGUE_SEQ,
	MEPILOGUE_SEQ,
	CALL_REG,
	CALL_SEQ,
	CALL_SYS,
	ADJUST_SEQ,
	CLEANUP_SEQ,
	HALT_SEQ;

! variables to bind addressing modes
struct	MODES =
	MODE_GLOB,	! acc/memory      (global)
	MODE_LOC,	! acc/[sp+offset] (local)
	MODE_IVAR,	! acc/[bx+offset] (instance)
	MODE_CONST,	! acc/immediate   (constant)
	MODE_SELF,	! acc/bx          (self)
	MODE_STACK;	! acc/[sp]        (popped value)

! Bind machine code fragments to variables.
! When a variable is bound to an array of sequences,
! then there is one entry for each of the above
! addressing modes. Otherwise, the instruction
! has an implied mode.
! The format of each entry is as follows:
!
!	member#		description
!	1		length of sequence
!	2		number of arguments	(0,1,2)
!	3		offset of argument 1	(optional)
!	4		offset of argument 2	(optional)
!	5		beginning of sequence
!
! Arguments are filled with patterns like 0xDDDD, 0xEEEE, 0xFFFF.
! If an argument is at the end of a sequence, the pattern is omitted.
!
mkopcodes() do
	LOAD_GLOB :=
	packed[ 2, 1, 2, 0x8B, 0x06 ];	! MOV AX,[EEEE]

	LOAD_LOC :=
	packed[ 2, 1, 2, 0x8B, 0x86 ];	! MOV AX,[BP+DDDD]

	LOAD_IVAR :=
	packed[ 2, 1, 2, 0x8B, 0x87 ];	! MOV AX,[BX+DDDD]

	LOAD_LOC_VEC :=
	packed[ 5, 1, 1, 0xB8, 0xDD, 0xDD,  	! MOV AX,DDDD
		0x01, 0xE8 ];			! ADD AX,BP

	LOAD_IVAR_VEC :=
	packed[ 5, 1, 1, 0xB8, 0xDD, 0xDD,  	! MOV AX,DDDD
		0x01, 0xD8 ];			! ADD AX,BX

	LOAD_CONST :=
	packed[ 1, 1, 1, 0xB8 ];	! MOV AX,EEEE

	LOAD_SELF :=
	packed[ 2, 0, 0x89, 0xD8 ];	! MOV AX,BX

	PUSH_SEQ :=
	packed[ 1, 0, 0x50 ];	! PUSH AX

	POP_SEQ :=
	packed[ 1, 0, 0x58 ];	! POP AX

	DEREF_SEQ := [
	packed[ 10, 1, 2, 0x8B, 0x36, 0xEE, 0xEE,	! MOV SI,[EEEE]
		0xD1, 0xE6,				! SHL SI,1
		0x03, 0xF0,				! ADD SI,AX
		0x8B, 0x04 ],				! MOV AX,[SI]

	packed[ 10, 1, 2, 0x8B, 0xB6, 0xDD, 0xDD,	! MOV SI,[BP+DDDD]
		0xD1, 0xE6,				! SHL SI,1
		0x03, 0xF0,				! ADD SI,AX
		0x8B, 0x04 ],				! MOV AX,[SI]

	packed[ 10, 1, 2, 0x8B, 0xB7, 0xDD, 0xDD,	! MOV SI,[BX+DDDD]
		0xD1, 0xE6,				! SHL SI,1
		0x03, 0xF0,				! ADD SI,AX
		0x8B, 0x04 ],				! MOV AX,[SI]

	packed[ 9, 1, 1, 0xBE, 0xFF, 0xFF,	! MOV DX,FFFF
		0xD1, 0xE6,			! SHL SI,1
		0x03, 0xF0,			! ADD SI,AX
		0x8B, 0x04 ],			! MOV AX,[SI]

	packed[ 8, 0, 0x8B, 0xF3,	! MOV SI,BX
		0xD1, 0xE6,		! SHL SI,1
		0x03, 0xF0,		! ADD SI,AX
		0x8B, 0x04 ],		! MOV AX,[SI]

	packed[ 7, 0, 0xD1, 0xE0,	! SHL AX,1
		0x5E,			! POP SI
		0x03, 0xF0,		! ADD SI,AX
		0x8B, 0x04 ],		! MOV AX,[SI]
	];

	DEREF_BYTE_SEQ := [
	packed[ 10, 1, 2,
		0x03, 0x06, 0xEE, 0xEE,	! ADD AX,[EEEE]
		0x89, 0xC6, 		! MOV SI,AX
		0x31, 0xC0,		! XOR AX,AX
		0x8A, 0x04 ],		! MOV AL,[SI]

	packed[ 10, 1, 2,
		0x03, 0x86, 0xDD, 0xDD,	! ADD AX,[BP+DDDD]
		0x89, 0xC6, 		! MOV SI,AX
		0x31, 0xC0,		! XOR AX,AX
		0x8A, 0x04 ],		! MOV AL,[SI]

	packed[ 10, 1, 2,
		0x03, 0x87, 0xDD, 0xDD,	! ADD AX,[BX+DDDD]
		0x89, 0xC6, 		! MOV SI,AX
		0x31, 0xC0,		! XOR AX,AX
		0x8A, 0x04 ],		! MOV AL,[SI]

	packed[ 10, 1, 2,
		0x81, 0xC0, 0xFF, 0xFF,	! ADD AX,FFFF
		0x89, 0xC6, 		! MOV SI,AX
		0x31, 0xC0,		! XOR AX,AX
		0x8A, 0x04 ],		! MOV AL,[SI]

	packed[ 8, 0, 0x01, 0xD8,	! ADD AX,BX
		0x89, 0xC6, 		! MOV SI,AX
		0x31, 0xC0,		! XOR AX,AX
		0x8A, 0x04 ],		! MOV AL,[SI]

	packed[ 7, 0, 0x5E,	! POP SI
		0x03, 0xF0,	! ADD SI,AX
		0x31, 0xC0,	! XOR AX,AX
		0x8A, 0x04 ]	! MOV AL,[SI]
	];

	STORE_GLOB :=
	packed[ 2, 1, 2, 0x89, 0x06 ];	! MOV [EEEE],AX

	STORE_LOC :=
	packed[ 2, 1, 2, 0x89, 0x86 ];	! MOV [BP+DDDD],AX

	STORE_IVAR :=
	packed[ 2, 1, 2, 0x89, 0x87 ];	! MOV [BX+DDDD],AX

	STORE_SEQ :=
	packed[ 3, 0, 0x5E,	! POP SI
		0x89, 0x04 ];	! MOV [SI],AX

	STORE_BYTE :=
	packed[ 3, 0, 0x5E,	! POP SI
		0x88, 0x04 ];	! MOV [SI],AL

	NORM_SEQ := [
	packed[ 8, 1, 2, 0x8B, 0x16, 0xEE, 0xEE,	! MOV DX,[EEEE]
		0xD1, 0xE2,				! SHL DX,1
		0x01, 0xD0 ],				! ADD AX,DX

	packed[ 8, 1, 2, 0x8B, 0x96, 0xDD, 0xDD,	! MOV DX,[BP+DDDD]
		0xD1, 0xE2,				! SHL DX,1
		0x01, 0xD0 ],				! ADD AX,DX

	packed[ 8, 1, 2, 0x8B, 0x97, 0xDD, 0xDD,	! MOV DX,[BX+DDDD]
		0xD1, 0xE2,				! SHL DX,1
		0x01, 0xD0 ],				! ADD AX,DX

	packed[ 7, 1, 1, 0xBA, 0xFF, 0xFF,	! MOV DX,FFFF
		0xD1, 0xE2,			! SHL DX,1
		0x01, 0xD0 ],			! ADD AX,DX

	packed[ 6, 0, 0x89, 0xDA,	! MOV DX,BX
		0xD1, 0xE2,		! SHL DX,1
		0x01, 0xD0 ],		! ADD AX,DX

	packed[ 5, 0, 0xD1, 0xE0,	! SHL AX,1
		0x5A,			! POP DX
		0x01, 0xD0 ]		! ADD AX,DX
	];

	INC_GLOB :=
	packed[ 6, 2, 2, 4, 0x81, 0x06 ];	! ADD WORD PTR [EEEE],FFFF

	INC_LOC :=
	packed[ 6, 2, 2, 4, 0x81, 0x86 ];	! ADD WORD PTR [BP+DDDD],FFFF

	INC_IVAR :=
	packed[ 6, 2, 2, 4, 0x81, 0x87 ];	! ADD WORD PTR [BX+DDDD],FFFF

	NEG_SEQ :=
	packed[ 2, 0, 0xF7, 0xD8 ];	! NEG AX

	NOT_SEQ :=
	packed[ 2, 0, 0xF7, 0xD0 ];	! NOT AX

	LNOT_SEQ :=
	packed[ 6, 0, 0xF7, 0xD8,	! NEG AX
		0x19, 0xC0,		! SBB AX,AX
		0xF7, 0xD0 ];		! NOT AX

	CMP_ZERO_SEQ :=
	packed[ 2, 0, 0x09, 0xC0 ];	! OR AX,AX

	MUL_SEQ := [
	packed[ 6, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0xF7, 0xE9 ],				! IMUL CX

	packed[ 6, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0xF7, 0xE9 ],				! IMUL CX

	packed[ 6, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0xF7, 0xE9 ],				! IMUL CX

	packed[ 5, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0xF7, 0xE9 ],			! IMUL CX

	packed[ 2, 0, 0xF7, 0xEB ],	! IMUL BX

	packed[ 3, 0, 0x59,	! POP CX
		0xF7, 0xE9 ]	! IMUL CX
	];

	DIV_SEQ := [
	packed[ 7, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0x99,					! CWD 
		0xF7, 0xF9 ],				! IDIV CX

	packed[ 7, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0x99,					! CWD 
		0xF7, 0xF9 ],				! IDIV CX

	packed[ 7, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0x99,					! CWD 
		0xF7, 0xF9 ],				! IDIV CX

	packed[ 6, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0x99,				! CWD 
		0xF7, 0xF9 ],			! IDIV CX

	packed[ 2, 0, 0xF7, 0xFB ],	! IDIV BX

	packed[ 6, 0, 0x59,	! POP CX
		0x87, 0xC8,	! XCHG CX,AX
		0x99,		! CWD 
		0xF7, 0xF9 ]	! IDIV CX
	];

	MOD_SEQ := [
	packed[ 10, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1,				! DIV CX
		0x89, 0xD0 ],				! MOV AX,DX

	packed[ 10, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1,				! DIV CX
		0x89, 0xD0 ],				! MOV AX,DX

	packed[ 10, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1,				! DIV CX
		0x89, 0xD0 ],				! MOV AX,DX

	packed[ 9, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0x31, 0xD2,			! XOR DX,DX
		0xF7, 0xF1,			! DIV CX
		0x89, 0xD0 ],			! MOV AX,DX

	packed[ 6, 0, 0x31, 0xD2,	! XOR DX,DX
		0xF7, 0xF3,		! DIV BX
		0x89, 0xD0 ],		! MOV AX,DX

	packed[ 9, 0, 0x59,	! POP CX
		0x87, 0xC8,	! XCHG CX,AX
		0x31, 0xD2,	! XOR DX,DX
		0xF7, 0xF1,	! DIV CX
		0x89, 0xD0 ]	! MOV AX,DX
	];

	UMUL_SEQ := [
	packed[ 6, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0xF7, 0xE1 ],				! MUL CX

	packed[ 6, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0xF7, 0xE1 ],				! MUL CX

	packed[ 6, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0xF7, 0xE1 ],				! MUL CX

	packed[ 5, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0xF7, 0xE1 ],			! MUL CX

	packed[ 2, 0, 0xF7, 0xE3 ],	! MUL BX

	packed[ 3, 0, 0x59,	! POP CX
		0xF7, 0xE1 ]	! MUL CX
	];

	UDIV_SEQ := [
	packed[ 8, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1 ],				! DIV CX

	packed[ 8, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1 ],				! DIV CX

	packed[ 8, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0x31, 0xD2,				! XOR DX,DX
		0xF7, 0xF1 ],				! DIV CX

	packed[ 7, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0x31, 0xD2,			! XOR DX,DX
		0xF7, 0xF1 ],			! DIV CX

	packed[ 4, 0, 0x31, 0xD2,	! XOR DX,DX
		0xF7, 0xF3 ],		! DIV BX

	packed[ 7, 0, 0x59,	! POP CX
		0x87, 0xC8,	! XCHG CX,AX
		0x31, 0xD2,	! XOR DX,DX
		0xF7, 0xF1 ]	! DIV CX
	];

	ADD_SEQ := [
	packed[ 2, 1, 2, 0x03, 0x06 ],	! ADD AX,[EEEE]

	packed[ 2, 1, 2, 0x03, 0x86 ],	! ADD AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x03, 0x87 ],	! ADD AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xC0 ],	! ADD AX,FFFF

	packed[ 2, 0, 0x01, 0xD8 ],	! ADD AX,BX

	packed[ 3, 0, 0x5A,	! POP DX
		0x01, 0xD0 ]	! ADD AX,DX
	];

	SUB_SEQ := [
	packed[ 2, 1, 2, 0x2B, 0x06 ],	! SUB AX,[EEEE]

	packed[ 2, 1, 2, 0x2B, 0x86 ],	! SUB AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x2B, 0x87 ],	! SUB AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xE8 ],	! SUB AX,FFFF

	packed[ 2, 0, 0x29, 0xD8 ],	! SUB AX,BX

	packed[ 5, 0, 0x5A,	! POP DX
		0x87, 0xD0,	! XCHG DX,AX
		0x29, 0xD0 ]	! SUB AX,DX
	];

	AND_SEQ := [
	packed[ 2, 1, 2, 0x23, 0x06 ],	! AND AX,[EEEE]

	packed[ 2, 1, 2, 0x23, 0x86 ],	! AND AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x23, 0x87 ],	! AND AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xE0 ],	! AND AX,FFFF

	packed[ 2, 0, 0x21, 0xD8 ],	! AND AX,BX

	packed[ 3, 0, 0x5A,	! POP DX
		0x21, 0xD0 ]	! AND AX,DX
	];

	OR_SEQ := [
	packed[ 2, 1, 2, 0x0B, 0x06 ],	! OR AX,[EEEE]

	packed[ 2, 1, 2, 0x0B, 0x86 ],	! OR AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x0B, 0x87 ],	! OR AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xC8 ],	! OR AX,FFFF

	packed[ 2, 0, 0x09, 0xD8 ],	! OR AX,BX

	packed[ 3, 0, 0x5A,	! POP DX
		0x09, 0xD0 ]	! OR AX,DX
	];

	XOR_SEQ := [
	packed[ 2, 1, 2, 0x33, 0x06 ],	! XOR AX,[EEEE]

	packed[ 2, 1, 2, 0x33, 0x86 ],	! XOR AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x33, 0x87 ],	! XOR AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xF0 ],	! XOR AX,FFFF

	packed[ 2, 0, 0x31, 0xD8 ],	! XOR AX,BX

	packed[ 3, 0, 0x5A,	! POP DX
		0x31, 0xD0 ]	! XOR AX,DX
	];

	SHL_SEQ := [
	packed[ 6, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE, 	! MOV CX,[EEEE]
		0xD3, 0xE0 ],				! SHL AX,CL

	packed[ 6, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0xD3, 0xE0 ],				! SHL AX,CL

	packed[ 6, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0xD3, 0xE0 ],				! SHL AX,CL

	packed[ 5, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0xD3, 0xE0 ],			! SHL AX,CL

	packed[ 4, 0, 0x89, 0xD9,	! MOV CX,BX
		0xD3, 0xE0 ],		! SHL AX,CL

	packed[ 5, 0, 0x59,	! POP CX
		0x87, 0xC8,	! XCHG CX,AX
		0xD3, 0xE0 ]	! SHL AX,CL
	];

	SHR_SEQ := [
	packed[ 6, 1, 2, 0x8B, 0x0E, 0xEE, 0xEE,	! MOV CX,[EEEE]
		0xD3, 0xE8 ],				! SHR AX,CL

	packed[ 6, 1, 2, 0x8B, 0x8E, 0xDD, 0xDD,	! MOV CX,[BP+DDDD]
		0xD3, 0xE8 ],				! SHR AX,CL

	packed[ 6, 1, 2, 0x8B, 0x8F, 0xDD, 0xDD,	! MOV CX,[BX+DDDD]
		0xD3, 0xE8 ],				! SHR AX,CL

	packed[ 5, 1, 1, 0xB9, 0xFF, 0xFF,	! MOV CX,FFFF
		0xD3, 0xE8 ],			! SHR AX,CL

	packed[ 4, 0,  0x89, 0xD9,	! MOV CX,BX
		0xD3, 0xE8 ],		! SHR AX,CL

	packed[ 5, 0, 0x59,	! POP CX
		0x87, 0xC8,	! XCHG CX,AX
		0xD3, 0xE8 ]	! SHR AX,CL
	];

	CMP_SEQ := [
	packed[ 2, 1, 2, 0x3B, 0x06 ],	! CMP AX,[EEEE]

	packed[ 2, 1, 2, 0x3B, 0x86 ],	! CMP AX,[BP+DDDD]

	packed[ 2, 1, 2, 0x3B, 0x87 ],	! CMP AX,[BX+DDDD]

	packed[ 2, 1, 2, 0x81, 0xF8 ],	! CMP AX,FFFF

	packed[ 2, 0, 0x39, 0xD8 ],	! CMP AX,BX

	packed[ 3, 0, 0x5A,	! POP DX
		0x39, 0xC2 ]	! CMP DX,AX
	];

	SWAP_SEQ := packed[ 4, 0, 0x89, 0xE6,	! MOV SI,SP
		0x87, 0x04 ];			! XCHG AX,[SI]

	BE_SEQ := packed[ 1, 1, 1, 0x74 ];	! JZ 02D4

	BNE_SEQ := packed[ 1, 1, 1, 0x75 ];	! JNZ 02D4

	BL_SEQ := packed[ 1, 1, 1, 0x7C ];	! JL 02D4

	BG_SEQ := packed[ 1, 1, 1, 0x7F ];	! JG 02D4

	BLE_SEQ := packed[ 1, 1, 1, 0x7E ];	! JLE 02D4

	BGE_SEQ := packed[ 1, 1, 1, 0x7D ];	! JGE 02D4

	BUL_SEQ := packed[ 1, 1, 1, 0x72 ];	! JB 02D4

	BUG_SEQ := packed[ 1, 1, 1, 0x77 ];	! JA 02D4

	BULE_SEQ := packed[ 1, 1, 1, 0x76 ];	! JBE 02D4

	BUGE_SEQ := packed[ 1, 1, 1, 0x73 ];	! JNB 02D4

	JMP_SEQ := packed[ 1, 1, 1, 0xE9 ];	! JMP CDCC

	JMPS_SEQ := packed[ 1, 1, 1, 0xEB ];	! JMPS 02D4

	TRUE_SEQ := packed[ 3, 0, 0xB8, 0xFF, 0xFF ];	! MOV AX,FFFF

	FALSE_SEQ := packed[ 2, 0, 0x31, 0xC0 ];	! XOR AX,AX

	PROLOGUE_SEQ := packed[ 3, 0, 0x55,	! PUSH BP
		0x89, 0xE5 ];			! MOV BP,SP

	EPILOGUE_SEQ := packed[ 2, 0, 0x5D,	! POP BP
		0xC3 ];				! RET 

	MPROLOGUE_SEQ := packed[ 7, 0, 0x55,	! PUSH BP
		0x89, 0xE5,			! MOV BP,SP
		0x53,				! PUSH BX
		0x8B, 0x5E, 0x04 ];		! MOV BX,[BP+04]

	MEPILOGUE_SEQ := packed[ 3, 0, 0x5B,	! POP BX
		0x5D,				! POP BP
		0xC3 ];				! RET 

	CALL_REG := packed[ 2, 0, 0xFF, 0xD0 ];	! CALL AX

	CALL_SEQ := packed[ 1, 1, 1, 0xE8 ];	! CALL CDCC

	CALL_SYS := packed[ 3, 1, 3, 0x2E,	! CS:
		0xFF, 0x16 ];			! CALL [EEEE]

	ADJUST_SEQ := packed[ 2, 1, 2, 0x81, 0xEC ];	! SUB SP,FFFF

	CLEANUP_SEQ := packed[ 2, 1, 2, 0x81, 0xC4 ];	! ADD SP,FFFF

	HALT_SEQ := packed[ 7, 1, 1, 0xB8, 0xFF, 0xFF,	! MOV AX,FFFF
		0xB4, 0x4C,				! MOV AH,4C
		0xCD, 0x21 ];				! INT 21
end

init() do var i, tmpname::240;
	Ninf := 0;
	Nt := 0;
	Bt := 0;
	Rejected := %1;
	Addr := 0;
	Active := 0;
	Queued := 0;
	Depth := 0;
	Dataname := 0;
	Verbose := 0;
	t.memfill(Infmap, 0, MAXINFPROC*t.bpw());
	K64k := 16384;		! split K64k := 16384 .* 4; to
	K64k := K64k .* 4;	! avoid constant expression folding
	KFFFF := K64k-1;
	t.memfill(Labels, -1, MAXLABEL*t.bpw());
	ie (t.getarg(Verbose->2: 1, tmpname, 240) > 1) do
		ie (tmpname::0 = '-' /\ tmpname::1 = 'T') do
			ie (tmpname::2)
				t.memcopy(tmpname, @tmpname::2,
					str.length(@tmpname::1));
			else if (t.getarg(2, tmpname, 240) < 2)
				tmpname::0 := 0;
		end
		else do
			tmpname::0 := 0;
		end
	end
	else do
		tmpname::0 := 0;
	end
	if (\tmpname::0 /\ t.getenv("TMP", tmpname, 240) < 0) do
		tmpname::0 := '.';
		tmpname::1 := 0;
	end
	i := str.length(tmpname);
	t.memcopy(Tmpfile, tmpname, i);
	t.memcopy(@Tmpfile::i, "/txcgexe.tmp", 13);
	t.memcopy(Tmpdata, tmpname, i);
	t.memcopy(@Tmpdata::i, "/txcgdata.tmp", 14);
	mkopcodes();
end

error(msg, name) do var buf::128;
	t.write(T3X.SYSERR, "TXCGEXE(", 8);
	t.write(T3X.SYSERR, Firstpass-> "1): ": "2): ", 4);
	t.write(T3X.SYSERR, str.numtostr(buf, Addr, 10), str.length(buf));
	t.write(T3X.SYSERR, ": ", 2);
	t.write(T3X.SYSERR, msg, str.length(msg));
	if (name) do
		t.write(T3X.SYSERR, ": ", 2);
		t.write(T3X.SYSERR, name, str.length(name));
	end
	t.write(T3X.SYSERR, t.newline(buf), str.length(buf));
	t.remove(Tmpfile);
	t.remove(Tmpdata);
	halt 1;
end

internal(msg) do
	t.write(T3X.SYSERR, "internal error: ", 16);
	error(msg, 0);
end

wrdata(n) do
	if (Dtop + Btop .>= KFFFF-256) error("data segment overflow", 0);
	Data.wrch(n);
	Data.wrch(n >> 8);
	Dtop := Dtop+BPW;
end

rdch() do var ch;
	if (Rejected >= 0) do
		var	o;
		o := Rejected;
		Rejected := %1;
		return o;
	end
	ch := In.rdch();
	if (Firstpass /\ \In.eof()) Out.wrch(ch);
	Addr := Addr+1;
	return ch;
end

wrch(ch) do
	if (Ctop .>= KFFFF-16) error("text segment overflow", 0);
	Ctop := Ctop+1;
	if (Firstpass) return;
	Out.wrch(ch);
end

! emit a code sequence with merged-in arguments
emit(code, args) do var k, n, i, j, p0, p1;
	p0 := -1;
	p1 := -1;
	k := code::0;
	n := code::1;
	i := 2;
	if (n) do
		p0 := code::2;
		i := 3;
		if (n > 1) do
			p1 := code::3;
			i := 4;
		end
	end
	if (p0 >= k) k := k + BPW;
	if (p1 >= k) k := k + BPW;
	for (j=0, k) do
		ie (j = p0) do
			wrch(args[0]);
			wrch(args[0] >> 8);
			j := j+1;
		end
		else ie (j = p1) do
			wrch(args[1]);
			wrch(args[1] >> 8);
			j := j+1;
		end
		else do
			wrch(code::(i+j));
		end
	end
end

! emit a PC-relative jump/call instruction
emitj(code, arg) do var k, n, i, j, p0, a;
	p0 := -1;
	k := code::0;
	n := code::1;
	i := 2;
	if (n) do
		p0 := code::2;
		i := 3;
	end
	if (p0 >= k) k := k + BPW;
	a := arg - Ctop - k;
	for (j=0, k) do
		ie (j = p0) do
			wrch(a);
			wrch(a >> 8);
			j := j+1;
		end
		else do
			wrch(code::(i+j));
		end
	end
end

! emit a conditional jump
emit8(code, arg) do
	wrch(code::3);	! assume 1,1,1 prefix
	wrch(arg);
end

mklabel(id, lt, a) do
	if (id >= MAXLABEL) error("mklabel: label too big", 0);
	if (Labels[id] \= -1) error("label redefinition", 0);
	Labels[id] := a;
	Ltypes::id := lt;
end

find(id) do
	if (id >= MAXLABEL) error("find: label too big", 0);
	if (Labels[id] = -1 /\ \Firstpass) error("undefined label", 0);
	return Labels[id] + (Ltypes::id = 'B'-> Bss: 0);
end

push() do
	if (\Active) internal("accumulator not active in PUSH()");
	emit(PUSH_SEQ, 0);
	Depth := Depth+1;
	Active := 0;
end

pop() do
	if (\Depth) do
		if (Active) do
			Active := 0;
			return;
		end
		internal("stack underflow in POP()");
	end
	emit(POP_SEQ, 0);
	Depth := Depth-1;
	Active := 1;
end

! load queued value into accumulator
spill(m, a) do
	if (Active) push();
	ie (m = tcode.ILDG) emit(LOAD_GLOB, [(find(a))]);
	else ie (m = tcode.ILDGV) emit(LOAD_CONST, [(find(a))]);
	else ie (m = tcode.ILDL) emit(LOAD_LOC, [(-a*BPW)]);
	else ie (m = tcode.ILDLV) emit(LOAD_LOC_VEC, [(-a*BPW)]);
	else ie (m = tcode.ILDI) emit(LOAD_IVAR, [(a*BPW)]);
	else ie (m = tcode.ILDIV) emit(LOAD_IVAR_VEC, [(a*BPW)]);
	else ie (m = tcode.ILDLAB) emit(LOAD_CONST, [(find(a))]);
	else ie (m = tcode.INUM) emit(LOAD_CONST, [(a)]);
	else ie (m = tcode.ISELF) emit(LOAD_SELF, 0);
	else internal("unknown instruction in LOAD()");
	Active := 1;
	Queued := 0;
end

! generate load instruction
load(m, a) do
	ie (\Queued) do
		Iqueue := m;
		Aqueue := a;
		Queued := 1;
		return;
	end
	else do
		spill(Iqueue, Aqueue);
		Iqueue := m;
		Aqueue := a;
		Queued := 1;
	end
end

unary(op) do
	if (Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("accumulator not active in UNARY()");
		pop();
	end
	ie (op = tcode.INEG) emit(NEG_SEQ, 0);
	else ie (op = tcode.IBNOT) emit(NOT_SEQ, 0);
	else ie (op = tcode.ILNOT) emit(LNOT_SEQ, 0);
	else internal("unknown instruction in UNARY()");
end

emitb(seq, mode) emit(seq[mode], [(Aqueue)]);

binary(op) do var mode, op2, flg, lab, rel;
	if (\Active /\ Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("too few operands in BINARY()");
		pop();
	end
	ie (Queued) do
		ie (Iqueue = tcode.ILDG) mode := MODE_GLOB;
		else ie (Iqueue = tcode.ILDGV) mode := MODE_CONST;
		else ie (Iqueue = tcode.ILDL) mode := MODE_LOC;
		else ie (Iqueue = tcode.ILDLV) internal("@local in BINARY()");
		else ie (Iqueue = tcode.ILDI) mode := MODE_IVAR;
		else ie (Iqueue = tcode.ILDIV) internal("@ivar in BINARY()");
		else ie (Iqueue = tcode.ILDLAB) mode := MODE_CONST;
		else ie (Iqueue = tcode.INUM) mode := MODE_CONST;
		else ie (Iqueue = tcode.ISELF) mode := MODE_SELF;
		else internal("bad mode in BINARY()");
		! these instructions reference labels
		if (	Iqueue = tcode.ILDG \/ Iqueue = tcode.ILDGV \/
			Iqueue = tcode.ILDLAB
		) 
			Aqueue := find(Aqueue);
	end
	else do
		if (\Depth) internal("stack underflow in BINARY()");
		mode := MODE_STACK;
		Depth := Depth-1;
	end
	Queued := 0;
	Active := 1;
	! convert Tcode-style local/ivar addresses into real addresses
	ie (mode = MODE_LOC) Aqueue := -Aqueue*BPW;
	else if (mode = MODE_IVAR) Aqueue := Aqueue*BPW;
	ie (op = tcode.INORMB) emitb(ADD_SEQ, mode);
	else ie (op = tcode.INORM) emitb(NORM_SEQ, mode);
	else ie (op = tcode.IDEREF) emitb(DEREF_SEQ, mode);
	else ie (op = tcode.IDREFB) emitb(DEREF_BYTE_SEQ, mode);
	else ie (op = tcode.IMUL) emitb(MUL_SEQ, mode);
	else ie (op = tcode.IDIV) emitb(DIV_SEQ, mode);
	else ie (op = tcode.IMOD) emitb(MOD_SEQ, mode);
	else ie (op = tcode.IUMUL) emitb(UMUL_SEQ, mode);
	else ie (op = tcode.IUDIV) emitb(UDIV_SEQ, mode);
	else ie (op = tcode.IADD) emitb(ADD_SEQ, mode);
	else ie (op = tcode.ISUB) emitb(SUB_SEQ, mode);
	else ie (op = tcode.IBAND) emitb(AND_SEQ, mode);
	else ie (op = tcode.IBOR) emitb(OR_SEQ, mode);
	else ie (op = tcode.IBXOR) emitb(XOR_SEQ, mode);
	else ie (op = tcode.IBSHL) emitb(SHL_SEQ, mode);
	else ie (op = tcode.IBSHR) emitb(SHR_SEQ, mode);
	else do
		emitb(CMP_SEQ, mode);
		op2 := rdch();
		ie (op2 = tcode.IBRT \/ op2 = tcode.IBRF) do
			pop();
			flg := op2=tcode.IBRF;
			lab := rdch();
			lab := rdch() << 8 | lab;
			ie (op = tcode.IEQU) rel := flg-> BE_SEQ: BNE_SEQ;
			else ie (op = tcode.INEQU)
				rel := flg-> BNE_SEQ: BE_SEQ;
			else ie (op = tcode.ILESS)
				rel := flg-> BL_SEQ: BGE_SEQ;
			else ie (op = tcode.IGRTR)
				rel := flg-> BG_SEQ: BLE_SEQ;
			else ie (op = tcode.ILTEQ)
				rel := flg-> BLE_SEQ: BG_SEQ;
			else ie (op = tcode.IGTEQ)
				rel := flg-> BGE_SEQ: BL_SEQ;
			else ie (op = tcode.IULESS)
				rel := flg-> BUL_SEQ: BUGE_SEQ;
			else ie (op = tcode.IUGRTR)
				rel := flg-> BUG_SEQ: BULE_SEQ;
			else ie (op = tcode.IULTEQ)
				rel := flg-> BULE_SEQ: BUG_SEQ;
			else ie (op = tcode.IUGTEQ)
				rel := flg-> BUGE_SEQ: BUL_SEQ;
			else internal("unknown relation in BINARY()");
			emit8(rel, 3);
			emitj(JMP_SEQ, find(lab));
		end
		else do
			Rejected := op2;
			ie (op = tcode.IEQU) rel := BNE_SEQ;
			else ie (op = tcode.INEQU) rel := BE_SEQ;
			else ie (op = tcode.ILESS) rel := BGE_SEQ;
			else ie (op = tcode.IGRTR) rel := BLE_SEQ;
			else ie (op = tcode.ILTEQ) rel := BG_SEQ;
			else ie (op = tcode.IGTEQ) rel := BL_SEQ;
			else ie (op = tcode.IULESS) rel := BUGE_SEQ;
			else ie (op = tcode.IUGRTR) rel := BULE_SEQ;
			else ie (op = tcode.IULTEQ) rel := BUG_SEQ;
			else ie (op = tcode.IUGTEQ) rel := BUL_SEQ;
			else internal("unknown relation in BINARY()");
			emit8(rel, 5);
			emit(TRUE_SEQ, 0);
			emit8(JMPS_SEQ, 2);
			emit(FALSE_SEQ, 0);
		end
	end
end

store(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("too few operands in STORE()");
		pop();
	end
	ie (op = tcode.ISAVG) emit(STORE_GLOB, [(find(arg))]);
	else ie (op = tcode.ISAVL) emit(STORE_LOC, [(-arg*BPW)]);
	else ie (op = tcode.ISAVI) emit(STORE_IVAR, [(arg*BPW)]);
	else ie (op = tcode.ISTORB) emit(STORE_BYTE, 0);
	else ie (op = tcode.ISTORE) emit(STORE_SEQ, 0);
	else internal("bad mode in STORE()");
	if (op = tcode.ISTORE \/ op = tcode.ISTORB) Depth := Depth-1;
	pop();
	! Leftover values are OK when computing dynamic tables
	! if (Depth) internal("leftover values in STORE()");
end

pcall(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	ie (op = tcode.ICALR) do
		if (\Active) do
			if (\Depth)
				internal("no active reg in PCALL(CALR)");
			pop();
		end
		emit(CALL_REG, 0);
	end
	else ie (op = tcode.ICALL) do
		if (Active) push();
		emitj(CALL_SEQ, find(arg));
	end
	else ie (op = tcode.IICALL) do
		if (Active) push();
		ie (Firstpass) do
			emit(LOAD_CONST, [0]);
			emitj(CALL_SEQ, 0);
		end
		else do
			if (arg .>= MAXINFPROC \/ \Infmap[arg])
				error("bad interface mapping", 0);
			emit(LOAD_CONST, [(Infmap[arg][INF_OFFSET])]);
			emitj(CALL_SEQ, Infmap[arg][INF_ENTRY]);
		end
	end
	else do
		if (Active) push();
		emit(CALL_SYS, [(arg*2+4+16)]);
	end
	Active := 1;
end

pclean(arg) do
	Depth := Depth-arg;
	if (arg) emit(CLEANUP_SEQ, [(arg*BPW)]);
end

branch(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	ie (op = tcode.IJUMP) do
		emitj(JMP_SEQ, find(arg));
		if (Depth \/ Active) pop();
	end
	else do
		if (\Active) do
			if (\Depth) internal("too few operands in BRANCH()");
			pop();
		end
		emit(CMP_ZERO_SEQ, 0);
		if (op = tcode.IBRT \/ op = tcode.IBRF) pop();
		emit8(op=tcode.IBRT \/ op=tcode.INBRT-> BE_SEQ: BNE_SEQ, 3);
		emitj(JMP_SEQ, find(arg));
	end
end

genstr(s, k) do var i;
	if (Dtop + Btop + k .>= KFFFF-256)
		error("data segment overflow", 0);
	for (i=0, k) Data.wrch(s::i);
	Data.wrch(0);
	Dtop := Dtop + k+1;
	if (Dtop & 1) do
		Data.wrch(0);
		Dtop := Dtop+1;
	end
end

mapinf(id, name, len) do var i, ip;
	for (i=0, Ninf) do
		ip := @Infprocs[i*INF];
		if (	ip[INF_NAME]::0 = name::0 /\
			\t.memcomp(ip[INF_NAME], name, len+1)
		)
			leave;
	end
	if (i >= Ninf) error("unresolved interface procedure", name);
	if (id >= MAXINFPROC) error("interface ID out of range", name);
	Infmap[id] := ip;
end

w(s) do var b::4;
	t.write(T3X.SYSERR, s, str.length(s));
	t.write(T3X.SYSERR, t.newline(b), str.length(b));
end

generate(op, a1, a2, sbuf) do
	ie (	op = tcode.ILDG \/ op = tcode.ILDGV \/
		op = tcode.ILDL \/ op = tcode.ILDLV \/
		op = tcode.ILDI \/ op = tcode.ILDIV \/
		op = tcode.ILDLAB \/ op = tcode.INUM \/
		op = tcode.ISELF
	) do
		load(op, a1);
	end
	else ie (op = tcode.INEG \/ op = tcode.ILNOT \/
		op = tcode.IBNOT
	) do
		unary(op);
	end
	else ie (op = tcode.INORMB \/ op = tcode.INORM \/
		op = tcode.IMUL \/ op = tcode.IDIV \/
		op = tcode.IMOD \/ op = tcode.IADD \/
		op = tcode.ISUB \/ op = tcode.IBAND \/
		op = tcode.IBOR \/ op = tcode.IBXOR \/
		op = tcode.IBSHL \/ op = tcode.IBSHR \/
		op = tcode.IEQU \/ op = tcode.INEQU \/
		op = tcode.ILESS \/ op = tcode.IGRTR \/
		op = tcode.ILTEQ \/ op = tcode.IGTEQ \/
		op = tcode.IUMUL \/ op = tcode.IUDIV \/
		op = tcode.IULESS \/ op = tcode.IUGRTR \/
		op = tcode.IULTEQ \/ op = tcode.IUGTEQ \/
		op = tcode.IDREFB \/ op = tcode.IDEREF
	) do
		binary(op);
	end
	else ie (op = tcode.ISAVG \/ op = tcode.ISAVL \/ op = tcode.ISAVI \/
		op = tcode.ISTORE \/ op = tcode.ISTORB
	) do
		store(op, a1);
	end
	else ie (op = tcode.ICALL \/ op = tcode.ICALR \/
		op = tcode.ISYS \/ op = tcode.IICALL
	) do
		if (op = tcode.ISYS)
			error("The SYS mechanism is not supported by TXCGEE",
				0);
		pcall(op, a1);
	end
	else ie (op = tcode.IPOP) do
		if (Queued) spill(Iqueue, Aqueue);
		pop();
	end
	else ie (op = tcode.ICLEAN) do
		pclean(a1);
	end
	else ie (op = tcode.ISTACK) do
		emit(ADJUST_SEQ, [(a1*BPW)]);
	end
	else ie (op = tcode.IINCG) do
		emit(INC_GLOB, [(find(a1)),(a2)]);
	end
	else ie (op = tcode.IINCI) do
		emit(INC_IVAR, [(a1*BPW),(a2)]);
	end
	else ie (op = tcode.IINCL) do
		emit(INC_LOC, [(-a1*BPW),(a2)]);
	end
	else ie (op = tcode.IUNEXT \/ op = tcode.IDNEXT) do
		if (Queued) spill(Iqueue, Aqueue);
		if (\Active \/ \Depth)
			internal("too few operands in GENERATE(?NEXT)");
		emit(CMP_SEQ[MODE_STACK], 0);
		Depth := Depth-1;
		emit8(op=tcode.IUNEXT-> BL_SEQ: BG_SEQ, 3);
		emitj(JMP_SEQ, find(a1));
		pop();
	end
	else ie (op = tcode.IBRF \/ op = tcode.IBRT \/ op = tcode.INBRF \/
		op = tcode.INBRT \/
		op = tcode.IJUMP
	) do
		branch(op, a1);
	end
	else ie (op = tcode.IHDR) do
		emit(PROLOGUE_SEQ, 0);
		Active := 0;
	end
	else ie (op = tcode.IEND) do
		emit(EPILOGUE_SEQ, 0);
		if (Depth) error("unbalanced stack at END", 0);
		Active := 0;
	end
	else ie (op = tcode.IMHDR) do
		emit(MPROLOGUE_SEQ, 0);
		Active := 0;
	end
	else ie (op = tcode.IENDM) do
		emit(MEPILOGUE_SEQ, 0);
		if (Depth) error("unbalanced stack at END", 0);
		Active := 0;
	end
	else ie (op = tcode.IINIT) do
		if (a1 \= 7) error("unsupported Tcode version", 0);
	end
	else ie (op = tcode.ICLAB) do
		if (Queued) spill(Iqueue, Aqueue);
		if (Firstpass) mklabel(a1, 'C', Ctop);
	end
	else ie (op = tcode.IDLAB) do
		if (Dataname) do
			if (Firstpass) mklabel(Dataname, 'D', Dtop);
		end
		Dataname := a1;
	end
	else ie (op = tcode.IVEC) do
		if (Firstpass) do
			mklabel(Dataname, 'B', Btop);
			if (Dtop + Btop + a1 .>= KFFFF-256)
				error("data segment overflow", 0);
			Btop := Btop + (a1=0-> BPW: a1*BPW);
		end
		Dataname := 0;
	end
	else ie (op = tcode.IDATA \/ op = tcode.IDREF \/ op = tcode.ICREF) do
		if (Firstpass) do
			if (Dataname) mklabel(Dataname, 'D', Dtop);
			ie (op = tcode.IDATA) do
				wrdata(a1);
			end
			else do
				! memorize data->BSS references
				if (Ltypes::a1 = 'B') do
					if (Bt >= MAXBSSREF)
						error(
						"too many BSS references", 0);
					BSSrefs[Bt] := Dtop;
					Bt := Bt+1;
				end
				wrdata(find(a1));
			end
		end
		Dataname := 0;
	end
	else ie (op = tcode.ISTR) do
		if (Firstpass) do
			mklabel(Dataname, 'D', Dtop);
			genstr(sbuf, a1);
		end
		Dataname := 0;
	end
	else ie (op = tcode.IHALT) do
		emit(HALT_SEQ, [(a1)]);
	end
	else ie (op = tcode.IDUP) do
		if (Queued) spill(Iqueue, Aqueue);
		push();
		Active := 1;
	end
	else ie (op = tcode.ISWAP) do
		if (Queued) spill(Iqueue, Aqueue);
		if (Depth < 1) internal("too few operands in GENERATE(SWAP)");
		emit(SWAP_SEQ, 0);
	end
	else ie (op = tcode.IIPROC) do
		if (\Firstpass) mapinf(a2, sbuf, a1);
	end
	else ie (op = tcode.IHINT \/ op = tcode.ILIB \/
		op = tcode.IGLUE \/ op = tcode.ICMAP \/
		op = tcode.IPUB \/ op = tcode.IEXT \/
		op = tcode.IIREF
	) do
		! ignore
	end
	else ie (op = tcode.ICALX) do
		error("unresolved CALX (external reference)", 0);
	end
	else ie (op = tcode.IICALX) do
		error("unresolved ICALX (external interface)", 0);
	end
	else do
		internal("unknown instruction in GENERATE()");
	end
end

xlate() do
	var	op, o2, a1, a2, i, j, ch;
	var	sbuf[256];

	op := rdch();
	if (op \= tcode.IINIT) error("magic match failed", 0);
	while (\In.eof()) do
		o2 := op & ~128;
		if (o2 >= tcode.IENDOFSET) error("invalid opcode", 0);
		if (op & 128) do
			a1 := rdch();
			a1 := rdch() << 8 | a1;
			if (a1 > 32767) a1 := a1-K64k;	! 32-bit hack!
		end
		ie (	op = tcode.IINCL \/ op = tcode.IINCG \/
			op = tcode.IINCI \/ op = tcode.IINIT \/
			op = tcode.ICMAP
		) do
			a2 := rdch();
			a2 := rdch() << 8 | a2;
			if (a2 > 32767) a2 := a2-K64k;	! 32-bit hack!
		end
		else if (op >= tcode.IINIT \/ op = tcode.ISTR \/
			op = tcode.ILIB
		) do
			if (op \= tcode.ISTR /\ op \= tcode.ILIB) do
				a2 := a1;
				a1 := rdch();
				a1 := rdch() << 8 | a1;
				if (a1 > 32767) a1 := a1-K64k;	! 32-bit hack!
			end
			for (i=0, a1) sbuf::i := rdch();
			sbuf::i := 0;
		end
		generate(op, a1, a2, sbuf);
		op := rdch();
	end
end

addinfproc(ep, off, name, len) do var n;
	if (len + 1 + Nt >= NAMEPOOLSIZE) error("too many library names", 0);
	if (Ninf >= MAXINFPROC) error("too many interface procedures", 0);
	n := @Infprocs[Ninf*INF];
	Ninf := Ninf + 1;
	n[INF_ENTRY] := ep;
	n[INF_OFFSET] := off;
	n[INF_NAME] := @Names::Nt;
	t.memcopy(@Names::Nt, name, len+1);
	Nt := Nt + len + 1;
end

finddir(buf, len) do var i;
	for (i=len-6, -1, -1) do
		if (buf::i = 'T' /\ \t.memcomp(@buf::i, "T3XLIB", 6))
			return i+6;
	end
	return -1;
end

! adjust addresses in jump table of library
relocate(buf, n) do var i, r, p;
	for (i=0, n) do
		r := @buf::(i*2+6);
		p := r::1 << 8 | r::0;
		p := p+Ctop;
		r::0 := p;
		r::1 := p>>8;
	end
end

loadlib(path) do var buf::DIRBUFSIZE, ld, k, i, msg, n, nproc;
	msg := "corrupt library file";
	ld := t.open(path, T3X.OREAD);
	if (ld < 0) return -1;
	! read last block of library
	k := t.read(ld, buf, DIRBUFSIZE);
	if (k = DIRBUFSIZE) do
		t.seek(ld, DIRBUFSIZE, T3X.SEEK_END);
		k := t.read(ld, buf, DIRBUFSIZE);
		if (k \= DIRBUFSIZE) error("file read error", path);
	end
	i := finddir(buf, k);
	if (i < 0) error(msg, path);
	n := 0;
	! extract symbol table and build INFPROC table
	if (\Firstpass) do
		nproc := 0;
		while (buf::i) do
			if (i >= k) error(msg, path);
			addinfproc(Ctop+3, n, @buf::(i+1), buf::i-2);
			i := i + buf::i;
			n := n+BPW;
			nproc := nproc + 1;
		end
	end
	t.seek(ld, 0, T3X.SEEK_SET);
	k := t.read(ld, buf, DIRBUFSIZE);
	if (\Firstpass) relocate(buf, nproc);
	! copy library text to output file
	while (1) do
		i := finddir(buf, k);
		ie (i < 0) do
			ie (k < DIRBUFSIZE) do
				internal("LOADLIB: missing signature");
			end
			else do
				if (\Firstpass) Out.write(buf, DIRBUFSIZE/2);
				Ctop := Ctop + DIRBUFSIZE/2;
				t.memcopy(buf, @buf::(DIRBUFSIZE/2),
					DIRBUFSIZE/2);
				k := t.read(ld, @buf::(DIRBUFSIZE/2),
					DIRBUFSIZE/2);
				if (k >= 0) k := k + DIRBUFSIZE/2;
			end
		end
		else do
			! cut off symbol table
			if (\Firstpass) Out.write(buf, i-6);
			Ctop := Ctop+i-6;
			leave;
		end
	end
	t.close(ld);
end

loadlibs() do
	var	path::LIBMAX, base::LIBMAX, buf::LIBMAX;
	var	ld, k, i, n, p;

	ld := t.open("LIBDIR", T3X.OREAD);
	if (ld = -1) return;
	if (t.getenv("T3XDIR", base, LIBMAX) < 0) base::0 := 0;
	n := t.read(ld, buf, LIBMAX);
	t.close(ld);
	if (n >= LIBMAX-1) error("too many libraries", 0);
	for (i=0, n) if (buf::i = '\n' \/ buf::i = '\r') buf::i := 0;
	i := 0;
	while (i < n) do
		str.nformat(LIBMAX, path, "%S/lib/%S.x", [(base, @buf::i)]);
		if (str.length(path) >= LIBMAX-1)
			error("library path too long", path);
		k := str.length(@buf::i);
		p := str.length(path) - k - 2;
		i := i + k+1;
		while (buf::i = 0 /\ i < n) i := i+1;
		if (loadlib(@path::p) < 0) do
			if (loadlib(path) < 0)
				error("no such extension library",
					@path::p);
		end
	end
end

pass1() do
	Firstpass := 1;
	Bss := 0;
	Ctop := 16;	! 16 bytes reserved for segment setup
	Dtop := 0;
	Btop := 0;
	In.create(T3X.SYSIN, Inbuf, BUFSIZE, iostream.FREAD);
	if (Out.open(Tmpfile, Outbuf, BUFSIZE, iostream.FWRITE) = %1)
		error("cannot create temporary file", 0);
	if (Data.open(Tmpdata, Databuf, BUFSIZE, iostream.FWRITE) = %1)
		error("cannot create data file", 0);
	loadlibs();
	xlate();
	Out.close();
	Data.close();
end

wrwrd(x) do
	Out.wrch(x);
	Out.wrch(x >> 8);
end

wrbs(k, v) do var i;
	for (i=0, k) Out.wrch(v::i);
end

! emit segment setup code
! tlen = text segment size in paragraphs (16-byte blocks)
segsetup(tlen) do
	wrbs(2, packed[ 0x8C, 0xC8 ]);		! MOV AX,CS
	wrbs(1, packed[ 0x05 ]); wrwrd(tlen);	! ADD AX,tlen
	wrbs(2, packed[ 0x8E, 0xD0 ]);		! MOV SS,AX
	wrbs(3, packed[ 0xBC, 0xFE, 0xFF ]);	! MOV SP,$FFFE
	wrbs(2, packed[ 0x8E, 0xD8 ]);		! MOV DS,AX
	wrbs(2, packed[ 0x8C, 0xC2 ]);		! MOV DX,ES	! save PSP
	wrbs(2, packed[ 0x8E, 0xC0 ]);		! MOV ES,AX
end

! generate EXE header
writeexehd() do
	const	delta = 32;     ! size of non-{TEXT,DATA} area (header)
	var	k, m, n;
	var	ltext, ldata;

	ltext := Ctop + 15 & ~15;	! round up to paragraph boundary
	ldata := Bss;
	m := ltext mod 512;
	n := (ldata+delta) mod 512;
	k := ltext ./ 512 + (ldata+delta) ./ 512;	! total size / 512
	ie (m+n = 0) do		! fix last page
		k := k-1;
		m := 0;
	end
	else ie (m+n = 512) do
		m := m+n-512;
	end
	else ie (m+n > 512) do
		k := k+1;
		m := m+n-512;
	end
	else do
		m := m+n;
	end
	! write EXE header
	wrbs(2, packed[ 0x4D, 0x5A ]);	! Magic
	wrwrd(m);			! Size MOD 512
	wrwrd(k+1);			! Size / 512 + 1
	wrwrd(0);			! Num. of RELOC entries
	wrwrd(2);			! (header size + 15) / 16
	wrwrd(0);			! MINALLOC
	wrwrd(~0);			! MAXALLOC
	wrwrd(ltext./16+2);		! Initial SS, ignored
	wrwrd(~2);			! Initial SP
	wrwrd(0);			! Checksum (0=ignore)
	wrwrd(0);			! IEP, IP
	wrwrd(0);			! IEP, CS
	wrwrd(28);			! Offset of RELOC
	wrwrd(0);			! Number of overlays
	wrbs(4, packed[ 0xFF, 0xFF, 0xFF, 0xFF ]);
	return ltext;
end

! fill text segment up to next paragraph boundary
fill(len) do var z::16, f;
	t.memfill(z, 0x5A, 16);
	if (len /\ len mod 16) do
		f := 16 - len mod 16;
		Out.write(z, f);
		Ctop := Ctop + f;
	end
end

! append data segment
appenddata(file) do var ch, k, n, bp, a;
	if (In.open(file, Inbuf, BUFSIZE, iostream.FREAD) = %1)
		error("cannot reopen data segment file", 0);
	a := 0;
	bp := 1;
	n := Bt-> BSSrefs[0]: -1;
	while (1) do
		ch := In.rdch();
		if (In.eof()) leave;
		if (a = n) do
			! address references BSS:
			! fix up by adding BSS start address
			n := bp >= Bt-> -1: BSSrefs[bp];
			bp := bp+1;
			k := (In.rdch() << 8 | ch) + Bss;
			if (Out.wrch(k & 255) < 0)
				error("file write error", 0);
			ch := k>>8;
			a := a+1;
		end
		if (Out.wrch(ch) < 0) error("file write error", 0);
		a := a+1;
	end
	In.close();
end

pass2() do var ltext, i, nl::4;
	Bss := Dtop;
	Out.create(T3X.SYSOUT, Outbuf, BUFSIZE, iostream.FWRITE);
	ltext := writeexehd();
	segsetup(ltext >> 4);
	if (In.open(Tmpfile, Inbuf, BUFSIZE, iostream.FREAD) = %1)
		error("cannot reopen temporary file", 0);
	Firstpass := 0;
	Ctop := 16;
	loadlibs();
	Addr := 0;
	Active := 0;
	xlate();
	In.close();
	t.remove(Tmpfile);
	fill(Ctop);
	appenddata(Tmpdata);
	t.remove(Tmpdata);
	Out.flush();
	t.newline(nl);
	if (Verbose) do var b::100;
		t.write(T3X.SYSERR,
			str.format(b, "text=%UD data=%UD bss=%UD%S",
			[(Ctop, Dtop, Btop, nl)]), str.length(b));
	end
end

do var abuf::4;
	init();
	if (t.getarg(1, abuf, 4) = 2 /\ \t.memcomp(abuf, "-v", 3))
		Verbose := 1;
	pass1();	! build data segment
	pass2();	! build text segment and emit EXE file
end
