! TXCGUC -- Tcode->Ugly C translator
! Nils M Holm, 2000-2004,2019
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txcguc(t3x, tcode, string, iostream);

object	t[t3x], str[string];

const	BUFSIZE =	1024;	! I/O buffer size
const	MAXINF =	256;	! max. interface procedures
const	MAXLABEL =	4096;	! max. number of labels
const	NAMEPOOLSIZE =	1024;	! interface name pool size
const	MAXPROTO =	1024;	! max. functions

const	STACKSIZE =		4096;

struct	INFPROC =
		INFLABEL,
		INFNARG,
		INFCMAP,
		INFNAME;

var	Phase;
object	In[iostream], Out[iostream], Tmp[iostream];
var	Rejected;
var	Inbuf::BUFSIZE;
var	Outbuf::BUFSIZE;
var	Tmpbuf::BUFSIZE;
var	Tmpfile::256;
var	Segment;
var	Addr;
var	Lbpw;
var	Tlabel;
var	Iqueue, Aqueue,			! 1-instruction queue (instr, arg)
	Queued,				! queue active indicator (1=active),
	Active,				! accumulator active flag (1=active),
	Depth;				! RT stack depth
var	Tprefix, Aprefix;
var	K64k;				! 65536
var	Infprocs[MAXINF*INFPROC],	! interface procedure table
	Infmax;				! # of interfaces
var	Names::NAMEPOOLSIZE, Ntop;	! Name pool, top of pool
var	Lname[MAXLABEL],		! Label/address pool
	Laddr[MAXLABEL];
var	Prototbl[MAXPROTO],		! Prototype table
	Ptop;				! top of Prototbl
var	Ltop;				! top of L/A pool
var	Dtop;				! top of data area
var	Dsize;				! size of data area
var	Verbose;

var	BANNER,
	HEAD,
	TAIL,
	DECL_WORD,
	DECL_CREF,
	DECL_DREF,
	DECL_CODE_LABEL,
	DECL_TMP_LABEL,
	LOAD_GLOB,
	LOAD_LOC,
	LOAD_IVAR,
	LOAD_GLOB_VEC,
	LOAD_LOC_VEC,
	LOAD_IVAR_VEC,
	LOAD_DLAB,
	LOAD_CLAB,
	LOAD_CONST,
	LOAD_SELF,
	PUSH_SEQ,
	POP_SEQ,
	DEREF_SEQ,
	DEREF_BYTE_SEQ,
	STORE_GLOB,
	STORE_LOC,
	STORE_IVAR,
	STORE_SEQ,
	STORE_BYTE,
	NORM_SEQ,
	NORM_BYTE_SEQ,
	INC_GLOB,
	INC_LOC,
	INC_IVAR,
	NEG_SEQ,
	NOT_SEQ,
	LNOT_SEQ,
	EXT_BOOL_SEQ,
	SAVE_CMP_RESULT,
	MUL_SEQ,
	DIV_SEQ,
	MOD_SEQ,
	UMUL_SEQ,
	UDIV_SEQ,
	ADD_SEQ,
	SUB_SEQ,
	AND_SEQ,
	OR_SEQ,
	XOR_SEQ,
	SHL_SEQ,
	SHR_SEQ,
	CMP_EQ_SEQ,
	CMP_NE_SEQ,
	CMP_LT_SEQ,
	CMP_GT_SEQ,
	CMP_LE_SEQ,
	CMP_GE_SEQ,
	CMP_ULT_SEQ,
	CMP_UGT_SEQ,
	CMP_ULE_SEQ,
	CMP_UGE_SEQ,
	SWAP_SEQ,
	BE_SEQ,
	BNE_SEQ,
	BL_SEQ,
	BG_SEQ,
	BLE_SEQ,
	BGE_SEQ,
	BUL_SEQ,
	BUG_SEQ,
	BULE_SEQ,
	BUGE_SEQ,
	JMP_SEQ,
	TRUE_SEQ,
	FALSE_SEQ,
	EPILOGUE_SEQ,
	MPROLOGUE_SEQ,
	MEPILOGUE_SEQ,
	CALL_REG,
	CALL_SEQ,
	ADJUST_SEQ,
	CLEANUP_SEQ,
	HALT_SEQ;

! variables to bind addressing modes
struct	MODES = MODE_GLOB,		! acc/memory
		MODE_LOC,		! acc/(local)
		MODE_IVAR,		! acc/(instance)
		MODE_CONST,		! acc/immediate
		MODE_SELF,		! acc/(self)
		MODE_STACK;		! acc/(popped value)

decl mkops2(0);

mkops() do
	BANNER := [
		"/*",
		" * File generated by TXCGUC, do not edit",
		" */",
	0 ];
	HEAD := [
		"#ifdef plan9",
		" #include <u.h>",
		" #include <libc.h>",
		"#else",
		" #include <stdlib.h>",
		" #include <string.h>",
		"#endif",
		"#define BPW	4",
		"#define XBPW	2",
		"#define CELLS	%1",
		"#define B	unsigned char",
		"#define C	int",
		"#define U	unsigned int",
		"C	A;",
		"B	*BP;",
		"C	*P;",
		"C	T;",
		"C	X;",
		"U	S = CELLS;",
		"U	F = CELLS;",
		"C	*I = 0;",
		"C	M[CELLS];",
		"U	*UM = (U*)M;",
		"#define push(X)	M[--S] = X",
		"#define pop()	M[S++]",
		"#ifdef plan9",
		" void exit(int n) {",
		"	char b[4];",
		"	sprint(b, \"%%d\", n & 255);",
		"	exits(n? b: nil);",
		" }",
		"#endif",
		"void init(void) {",
	0 ];
	TAIL := [ "}", 0 ];
	DECL_WORD := [ "M[%1] = %2;", 0 ];
	DECL_CREF := [ "{ void P%2(void); M[%1] = (C)&P%2; }", 0 ];
	DECL_DREF := [ "M[%1] = (C)&M[%2];", 0 ];
	DECL_CODE_LABEL := [ "A%1:", 0 ];
	DECL_TMP_LABEL := [ "T%1:", 0 ];
	LOAD_GLOB := [ "A = M[%1];", 0 ];
	LOAD_LOC := [ "A = M[F-(%1)];", 0];
	LOAD_IVAR := [ "A = I[%1];", 0 ];
	LOAD_GLOB_VEC := [ "A = (C)&M[%1];", 0 ];
	LOAD_LOC_VEC := [ "A = (C)&M[F-(%1)];", 0];
	LOAD_IVAR_VEC := [ "A = (C)&I[%1];", 0 ];
	LOAD_CONST := [ "A = %1;", 0 ];
	LOAD_CLAB := [ "A = (C)&P%1;", 0 ];
	LOAD_DLAB := [ "A = (C)&M[%1];", 0 ];
	LOAD_SELF := [ "A = (C)I;", 0 ];
	PUSH_SEQ := [ "push(A);", 0 ];
	POP_SEQ := [ "A = pop();" ,0 ];
	DEREF_SEQ := [
		[ "P = (C*)A + M[%1]; A = *P;", 0 ],
		[ "P = (C*)A + M[F-(%1)]; A = *P;", 0 ],
		[ "P = (C*)A + I[%1]; A = *P;", 0 ],
		[ "P = (C*)A + %1; A = *P;", 0 ],
		[ "P = (C*)A + I; A = *P;", 0 ],
		[ "P = (C*)pop() + A; A = *P;", 0 ],
	0 ];
	DEREF_BYTE_SEQ := [ 
		[ "BP = (B*)A + M[%1]; A = (C)*BP;", 0 ],
		[ "BP = (B*)A + M[F-(%1)]; A = (C)*BP;", 0 ],
		[ "BP = (B*)A + I[%1]; A = (C)*BP;", 0 ],
		[ "BP = (B*)A + %1; A = (C)*BP;", 0 ],
		[ "BP = (B*)A + I; A = (C)*BP;", 0 ],
		[ "BP = (B*)A + pop(); A = (C)*BP;", 0 ],
	0 ];
	STORE_GLOB := [ "M[%1] = A;", 0 ];
	STORE_LOC := [ "M[F-(%1)] = A;", 0 ];
	STORE_IVAR := [ "I[%1] = A;", 0 ];
	STORE_SEQ := [ "P = (C*)pop(); *P = A;", 0 ];
	STORE_BYTE := [ "BP = (B*)pop(); *BP = (B)A;", 0 ];
	NORM_SEQ := [
		[ "A += M[%1] << XBPW;", 0 ],
		[ "A += M[F-(%1)] << XBPW;", 0 ],
		[ "A += I[%1] << XBPW;", 0 ],
		[ "A += %1 << XBPW;", 0 ],
		[ "A += (C)I << XBPW;", 0 ],
		[ "A = (A << XBPW) + pop();", 0 ],
	0 ];
	NORM_BYTE_SEQ := [
		[ "A += M[%1];", 0 ],
		[ "A += M[F-(%1)];", 0 ],
		[ "A += I[%1];", 0 ],
		[ "A += %1;", 0 ],
		[ "A += (C)I;", 0 ],
		[ "A += pop();", 0 ],
	0 ];
	INC_GLOB := [ "M[%1] += %2;", 0 ];
	INC_LOC := [ "M[F-(%1)] += %2;", 0 ];
	INC_IVAR := [ "I[%1] += %2;", 0 ];
	NEG_SEQ := [ "A = -A;", 0 ];
	NOT_SEQ := [ "A = ~A;", 0 ];
	LNOT_SEQ := [ "A = A?0:-1;", 0 ];
	EXT_BOOL_SEQ := [ "A = T?-1:0;", 0 ];
	SAVE_CMP_RESULT := [ "T = A;", 0 ];
	mkops2();
end

mkops2() do
	MUL_SEQ := [
		[ "A *= M[%1];", 0 ],
		[ "A *= M[F-(%1)];", 0 ],
		[ "A *= I[%1];", 0 ],
		[ "A *= %1;", 0 ],
		[ "A *= (C)I;", 0 ],
		[ "A *= pop();", 0 ],
	0 ];
	DIV_SEQ := [
		[ "A /= M[%1];", 0 ],
		[ "A /= M[F-(%1)];", 0 ],
		[ "A /= I[%1];", 0 ],
		[ "A /= %1;", 0 ],
		[ "A /= (C)I;", 0 ],
		[ "A = pop() / A;", 0 ],
	0 ];
	MOD_SEQ := [
		[ "A %%= M[%1];", 0 ],
		[ "A %%= M[F-(%1)];", 0 ],
		[ "A %%= I[%1];", 0 ],
		[ "A %%= %1;", 0 ],
		[ "A %%= (C)I;", 0 ],
		[ "A = pop() %% A;", 0 ],
	0 ];
	UMUL_SEQ := [
		[ "A = (U)A * UM[%1];", 0 ],
		[ "A = (U)A * UM[F-(%1)];", 0 ],
		[ "A = (U)A * (U)I[%1];", 0 ],
		[ "A = (U)A * (U)%1;", 0 ],
		[ "A = (U)A * (U)I;", 0 ],
		[ "A = (U)A * (U)pop();", 0 ],
	0 ];
	UDIV_SEQ := [
		[ "A = (U)A / UM[%1];", 0 ],
		[ "A = (U)A / UM[F-(%1)];", 0 ],
		[ "A = (U)A / (U)I[%1];", 0 ],
		[ "A = (U)A / (U)%1;", 0 ],
		[ "A = (U)A / (U)I;", 0 ],
		[ "A = (U)pop() / (U)A;", 0 ],
	0 ];
	ADD_SEQ := NORM_BYTE_SEQ;
	SUB_SEQ := [
		[ "A -= M[%1];", 0 ],
		[ "A -= M[F-(%1)];", 0 ],
		[ "A -= I[%1];", 0 ],
		[ "A -= %1;", 0 ],
		[ "A -= (C)I;", 0 ],
		[ "A = pop() - A;", 0 ],
	0 ];
	AND_SEQ := [
		[ "A &= M[%1];", 0 ],
		[ "A &= M[F-(%1)];", 0 ],
		[ "A &= I[%1];", 0 ],
		[ "A &= %1;", 0 ],
		[ "A &= (C)I;", 0 ],
		[ "A &= pop();", 0 ],
	0 ];
	OR_SEQ := [
		[ "A |= M[%1];", 0 ],
		[ "A |= M[F-(%1)];", 0 ],
		[ "A |= I[%1];", 0 ],
		[ "A |= %1;", 0 ],
		[ "A |= (C)I;", 0 ],
		[ "A |= pop();", 0 ],
	0 ];
	XOR_SEQ := [
		[ "A ^= M[%1];", 0 ],
		[ "A ^= M[F-(%1)];", 0 ],
		[ "A ^= I[%1];", 0 ],
		[ "A ^= %1;", 0 ],
		[ "A ^= (C)I;", 0 ],
		[ "A ^= pop();", 0 ],
	0 ];
	SHL_SEQ := [
		[ "A = (U)A << UM[%1];", 0 ],
		[ "A = (U)A << UM[F-(%1)];", 0 ],
		[ "A = (U)A << (U)I[%1];", 0 ],
		[ "A = (U)A << (U)%1;", 0 ],
		[ "A = (U)A << (U)I;", 0 ],
		[ "A = (U)pop() << (U)A;", 0 ],
	0 ];
	SHR_SEQ := [
		[ "A = (U)A >> UM[%1];", 0 ],
		[ "A = (U)A >> UM[F-(%1)];", 0 ],
		[ "A = (U)A >> (U)I[%1];", 0 ],
		[ "A = (U)A >> (U)%1;", 0 ],
		[ "A = (U)A >> (U)I;", 0 ],
		[ "A = (U)pop() >> (U)A;", 0 ],
	0 ];
	CMP_EQ_SEQ := [
		[ "T = A == M[%1];", 0 ],
		[ "T = A == M[F-(%1)];", 0 ],
		[ "T = A == I[%1];", 0 ],
		[ "T = A == %1;", 0 ],
		[ "T = A == (C)I;", 0 ],
		[ "T = pop() == A;", 0 ],
	0 ];
	CMP_NE_SEQ := [
		[ "T = A != M[%1];", 0 ],
		[ "T = A != M[F-(%1)];", 0 ],
		[ "T = A != I[%1];", 0 ],
		[ "T = A != %1;", 0 ],
		[ "T = A != (C)I;", 0 ],
		[ "T = pop() != A;", 0 ],
	0 ];
	CMP_LT_SEQ := [
		[ "T = A < M[%1];", 0 ],
		[ "T = A < M[F-(%1)];", 0 ],
		[ "T = A < I[%1];", 0 ],
		[ "T = A < %1;", 0 ],
		[ "T = A < (C)I;", 0 ],
		[ "T = pop() < A;", 0 ],
	0 ];
	CMP_GT_SEQ := [
		[ "T = A > M[%1];", 0 ],
		[ "T = A > M[F-(%1)];", 0 ],
		[ "T = A > I[%1];", 0 ],
		[ "T = A > %1;", 0 ],
		[ "T = A > (C)I;", 0 ],
		[ "T = pop() > A;", 0 ],
	0 ];
	CMP_LE_SEQ := [
		[ "T = A <= M[%1];", 0 ],
		[ "T = A <= M[F-(%1)];", 0 ],
		[ "T = A <= I[%1];", 0 ],
		[ "T = A <= %1;", 0 ],
		[ "T = A <= (C)I;", 0 ],
		[ "T = pop() <= A;", 0 ],
	0 ];
	CMP_GE_SEQ := [
		[ "T = A >= M[%1];", 0 ],
		[ "T = A >= M[F-(%1)];", 0 ],
		[ "T = A >= I[%1];", 0 ],
		[ "T = A >= %1;", 0 ],
		[ "T = A >= (C)I;", 0 ],
		[ "T = pop() >= A;", 0 ],
	0 ];
	CMP_ULT_SEQ := [
		[ "T = (U)A < UM[%1];", 0 ],
		[ "T = (U)A < UM[F-(%1)];", 0 ],
		[ "T = (U)A < (U)I[%1];", 0 ],
		[ "T = (U)A < (U)%1;", 0 ],
		[ "T = (U)A < (U)I;", 0 ],
		[ "T = (U)pop() < (U)A;", 0 ],
	0 ];
	CMP_UGT_SEQ := [
		[ "T = (U)A > UM[%1];", 0 ],
		[ "T = (U)A > UM[F-(%1)];", 0 ],
		[ "T = (U)A > (U)I[%1];", 0 ],
		[ "T = (U)A > (U)%1;", 0 ],
		[ "T = (U)A > (U)I;", 0 ],
		[ "T = (U)pop() > (U)A;", 0 ],
	0 ];
	CMP_ULE_SEQ := [
		[ "T = (U)A <= UM[%1];", 0 ],
		[ "T = (U)A <= UM[F-(%1)];", 0 ],
		[ "T = (U)A <= (U)I[%1];", 0 ],
		[ "T = (U)A <= (U)%1;", 0 ],
		[ "T = (U)A <= (U)I;", 0 ],
		[ "T = (U)pop() <= (U)A;", 0 ],
	0 ];
	CMP_UGE_SEQ := [
		[ "T = (U)A >= UM[%1];", 0 ],
		[ "T = (U)A >= UM[F-(%1)];", 0 ],
		[ "T = (U)A >= (U)I[%1];", 0 ],
		[ "T = (U)A >= (U)%1;", 0 ],
		[ "T = (U)A >= (U)I;", 0 ],
		[ "T = (U)pop() >= (U)A;", 0 ],
	0 ];
	SWAP_SEQ := [ "X = A; A = M[S]; M[S] = X; ", 0 ];
	BE_SEQ := [ "if (!T) goto %A%2;", 0 ];
	BNE_SEQ := [ "if (T) goto %A%2;", 0 ];
	JMP_SEQ := [ "goto %A%2;", 0 ];
	TRUE_SEQ := [ "A = -1;", 0 ];
	FALSE_SEQ := [ "A = 0;", 0 ];
	EPILOGUE_SEQ := [
		"F = pop();",
		"++S;",
		"}",
	0 ];
	MEPILOGUE_SEQ := [
		"I = (C*)pop();",
		"F = pop();",
		"++S;",
		"}",
	0 ];
	CALL_REG := [ "(*(void(*)(void))A)()", 0 ];
	CALL_SEQ := [ "P%1();", 0 ];
	ADJUST_SEQ := [ "S -= %1;", 0 ];
	CLEANUP_SEQ := [ "S += %1;", 0 ];
	HALT_SEQ := [ "exit(%1);", 0 ];
end

decl	lookup(1);

reset() do
	Addr := 0;
	Tlabel := 0;
	Active := 0;
	Queued := 0;
	Depth := 0;
	Dtop := 0;
end

error(msg) do var buf::128;
	t.write(T3X.SYSERR, "TXCGUC: ", 8);
	t.write(T3X.SYSERR, str.numtostr(buf, Addr, 10), str.length(buf));
	t.write(T3X.SYSERR, ": ", 2);
	t.write(T3X.SYSERR, msg, str.length(msg));
	t.write(T3X.SYSERR, t.newline(buf), str.length(buf));
	halt 1;
end

init() do var i;
	K64k := 16384;	! avoid constant folding
	K64k := 4 * K64k;
	Rejected := %1;
	Lbpw := t.bpw();
	Aprefix := "A";
	Tprefix := "T";
	Infmax := 0;
	Ltop := 0;
	Dtop := 0;
	Ntop := 0;
	Verbose := 0;
	if (	t.getenv("TMP", Tmpfile, 240) < 1 /\
		t.getenv("tmp", Tmpfile, 240) < 1
	) do
		Tmpfile::0 := '.';
		Tmpfile::1 := 0;
	end
	i := str.length(Tmpfile);
	if (i >= 256-13) error("TMP path too long");
	t.memcopy(@Tmpfile::i, "/_txcguc.tmp", 12);
	reset();
	mkops();
end

internal(msg) do
	t.write(T3X.SYSERR, "Internal error: ", 16);
	error(msg);
end

rdch() do var ch;
	if (Rejected >= 0) do var o;
		o := Rejected;
		Rejected := %1;
		return o;
	end
	ch := In.rdch();
	Addr := Addr+1;
	if (Phase = 0 /\ ch \= -1) Tmp.wrch(ch);
	return ch;
end

reject(op) Rejected := op;

rdarg() do var a;
	a := rdch();
	a := rdch() << 8 | a;
	if (a > 32767) a := a - K64k;
	return a;
end

wrch(ch) Out.wrch(ch);

! emit a code sequence with arguments merged in
_emit(vcode, args) do var i, j, c, b::33, code, radix;
	j := 0;
	while (vcode[j]) do
		code := vcode[j]; j := j+1;
		i := 0;
		while (1) do
			c := code::i; i := i+1;
			if (\c) leave;
			ie (c = '%') do
				radix := -10;
				c := code::i; i := i+1;
				if (c = '+') do
					radix := 10;
					c := code::i; i := i+1;
				end
				ie ('1' <= c /\ c <= '9')
					Out.writes(str.numtostr(b,
						args[c-'1'], radix));
				else ie ('A' <= c /\ c <= 'Z')
					Out.writes(args[c-'A']);
				else ie (\c)
					i := i-1;
				else
					Out.wrch(c);
			end
			else do
				Out.wrch(c);
			end
		end
		Out.wrch('\n');
	end
end

emit(code, args) if (Phase = 2) _emit(code, args);

emitd(code, args) if (Phase = 1) _emit(code, args);

dgrow(n) do
	Dtop := Dtop + n;	! XXX size check?
end

emitl(seq, arg) if (Phase = 2) emit(seq, [(-arg)]);

emiti(seq, arg) if (Phase = 2) emit(seq, [(arg)]);

emitl2(seq, arg, a2) if (Phase = 2) emit(seq, [(-arg, a2)]);

emiti2(seq, arg, a2) if (Phase = 2) emit(seq, [(arg, a2)]);

newlabel() do
	Tlabel := Tlabel + 1;
	return Tlabel;
end

push() do
	if (\Active) internal("accumulator not active in PUSH()");
	emit(PUSH_SEQ, 0);
	Depth := Depth+1;
	Active := 0;
end

pop() do
	if (\Depth) do
		if (Active) do
			Active := 0;
			return;
		end
		internal("stack underflow in POP()");
	end
	emit(POP_SEQ, 0);
	Depth := Depth-1;
	Active := 1;
end

loadlabel(l) do var a;
	a := lookup(l);
	ie (a < 0)
		emit(LOAD_CLAB, [(l)]);
	else
		emit(LOAD_DLAB, [(a)]);
end

! value has been queued into accumulator by LOAD
spill(m, a) do
	if (Active) push();
	ie (m = tcode.ILDG) emit(LOAD_GLOB, [(lookup(a))]);
	else ie (m = tcode.ILDGV) emit(LOAD_GLOB_VEC, [(lookup(a))]);
	else ie (m = tcode.ILDL) emit(LOAD_LOC, [(a)]);
	else ie (m = tcode.ILDLV) emit(LOAD_LOC_VEC, [(a)]);
	else ie (m = tcode.ILDI) emiti(LOAD_IVAR, a);
	else ie (m = tcode.ILDIV) emiti(LOAD_IVAR_VEC, a);
	else ie (m = tcode.ILDLAB) loadlabel(a);
	else ie (m = tcode.INUM) emit(LOAD_CONST, [(a)]);
	else ie (m = tcode.ISELF) emit(LOAD_SELF, 0);
	else internal("unknown instruction in LOAD()");
	Active := 1;
	Queued := 0;
end

! generate load instruction
load(m, a) do
	ie (\Queued) do
		Iqueue := m;
		Aqueue := a;
		Queued := 1;
		return;
	end
	else do
		spill(Iqueue, Aqueue);
		Iqueue := m;
		Aqueue := a;
		Queued := 1;
	end
end

unary(op) do
	if (Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("accumulator not active in UNARY()");
		pop();
	end
	ie (op = tcode.INEG) emit(NEG_SEQ, 0);
	else ie (op = tcode.IBNOT) emit(NOT_SEQ, 0);
	else ie (op = tcode.ILNOT) emit(LNOT_SEQ, 0);
	else internal("unknown instruction in UNARY()");
end

emitb(seq, mode)
	if (Phase = 2)
		_emit(seq[mode],
			[(mode = MODE_GLOB-> lookup(Aqueue): Aqueue)]);

cmp_regs(op, mode) do var op2, rel, flg, lab, t1;
	ie (op = tcode.IEQU) rel := CMP_EQ_SEQ;
	else ie (op = tcode.INEQU) rel := CMP_NE_SEQ;
	else ie (op = tcode.ILESS) rel := CMP_LT_SEQ;
	else ie (op = tcode.IGRTR) rel := CMP_GT_SEQ;
	else ie (op = tcode.ILTEQ) rel := CMP_LE_SEQ;
	else ie (op = tcode.IGTEQ) rel := CMP_GE_SEQ;
	else ie (op = tcode.IULESS) rel := CMP_ULT_SEQ;
	else ie (op = tcode.IUGRTR) rel := CMP_UGT_SEQ;
	else ie (op = tcode.IULTEQ) rel := CMP_ULE_SEQ;
	else ie (op = tcode.IUGTEQ) rel := CMP_UGE_SEQ;
	else internal("unknown relation in BINARY()");
	emitb(rel, mode);
	op2 := rdch();
	ie (op2 = tcode.IBRT \/ op2 = tcode.IBRF) do
		pop();
		lab := rdarg();
		emit(op2 = tcode.IBRF-> BE_SEQ: BNE_SEQ,
			[(Aprefix, lab)]);
	end
	else do
		reject(op2);
		emit(EXT_BOOL_SEQ, 0);
	end
end

wm(s) do var nl::4;
	t.write(T3X.SYSERR, s, str.length(s));
	t.write(T3X.SYSERR, t.newline(nl), str.length(nl));
end

undefd_addr() do
	wm("TXNCG: Addresses of local variables and anonymous vectors");
	wm("	   are not supported in binary operations.");
	wm("	   This may be considered a bug of TXCGUC.");
	wm("	   See the documentation for details.");
	internal("unsupported addressing mode in binary()");
end

binary(op) do var mode;
	if (\Active /\ Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("too few operands in BINARY()");
		pop();
	end
	ie (Queued) do
		ie (Iqueue = tcode.ILDG) mode := MODE_GLOB;
		else ie (Iqueue = tcode.ILDGV) mode := MODE_CONST;
		else ie (Iqueue = tcode.ILDL) mode := MODE_LOC;
		else ie (Iqueue = tcode.ILDLV) undefd_addr();
		else ie (Iqueue = tcode.ILDI) mode := MODE_IVAR;
		else ie (Iqueue = tcode.ILDIV) undefd_addr();
		else ie (Iqueue = tcode.ILDLAB) undefd_addr();
		else ie (Iqueue = tcode.INUM) mode := MODE_CONST;
		else ie (Iqueue = tcode.ISELF) mode := MODE_SELF;
		else internal("bad mode in BINARY()");
	end
	else do
		if (\Depth) internal("stack underflow in BINARY()");
		mode := MODE_STACK;
		Depth := Depth-1;
	end
	Queued := 0;
	Active := 1;
	! convert local/ivar Tcode addresses into real addresses
	ie (op = tcode.INORMB) emitb(NORM_BYTE_SEQ, mode);
	else ie (op = tcode.INORM) emitb(NORM_SEQ, mode);
	else ie (op = tcode.IDEREF) emitb(DEREF_SEQ, mode);
	else ie (op = tcode.IDREFB) emitb(DEREF_BYTE_SEQ,mode);
	else ie (op = tcode.IMUL) emitb(MUL_SEQ, mode);
	else ie (op = tcode.IDIV) emitb(DIV_SEQ, mode);
	else ie (op = tcode.IMOD) emitb(MOD_SEQ, mode);
	else ie (op = tcode.IUMUL) emitb(UMUL_SEQ, mode);
	else ie (op = tcode.IUDIV) emitb(UDIV_SEQ, mode);
	else ie (op = tcode.IADD) emitb(ADD_SEQ, mode);
	else ie (op = tcode.ISUB) emitb(SUB_SEQ, mode);
	else ie (op = tcode.IBAND) emitb(AND_SEQ, mode);
	else ie (op = tcode.IBOR) emitb(OR_SEQ, mode);
	else ie (op = tcode.IBXOR) emitb(XOR_SEQ, mode);
	else ie (op = tcode.IBSHL) emitb(SHL_SEQ, mode);
	else ie (op = tcode.IBSHR) emitb(SHR_SEQ, mode);
	else cmp_regs(op, mode);
end

store(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	if (\Active) do
		if (\Depth) internal("too few operands in STORE()");
		pop();
	end
	ie (op = tcode.ISAVG) emit(STORE_GLOB, [(lookup(arg))]);
	else ie (op = tcode.ISAVL) emitl(STORE_LOC, -arg);
	else ie (op = tcode.ISAVI) emiti(STORE_IVAR, arg);
	else ie (op = tcode.ISTORB) emit(STORE_BYTE, 0);
	else ie (op = tcode.ISTORE) emit(STORE_SEQ, 0);
	else internal("bad mode in STORE()");
	if (op = tcode.ISTORE \/ op = tcode.ISTORB) Depth := Depth-1;
	pop();
	! Leftover values are OK when computing dynamic tables
	! if (Depth) internal("leftover values in STORE()");
end

findinfproc(id) do var y, i;
	for (i=0, Infmax, INFPROC) do
		y := @Infprocs[i];
		if (y[INFLABEL] = id) return y;
	end
	error("undefined interface procedure");
end

addinfproc(lab, len, narg, cmap, name) do var inf;
	if (Phase \= 0) return;
	if (len + Ntop + 1 >= NAMEPOOLSIZE \/
		Infmax + INFPROC >= MAXINF
	)
		error("too many interface procedures");
	inf := @Infprocs[Infmax];
	Infmax := Infmax + INFPROC;
	inf[INFLABEL] := lab;
	inf[INFNARG] := narg;
	inf[INFCMAP] := cmap;
	inf[INFNAME] := @Names::Ntop;
	Ntop := Ntop + len+1;
	t.memcopy(inf[INFNAME], name, len+1);
end

lookup(n) do var i;
	for (i=0, Ltop)
		if (Lname[i] = n)
			return Laddr[i];
	return %1;
end

addsym(n) do
	if (Phase \= 1) return;
	if (lookup(n) \= -1) error("redefinition of label");
	if (Ltop >= MAXLABEL) error("too many labels");
	Lname[Ltop] := n;
	Laddr[Ltop] := Dtop;
	Ltop := Ltop+1;
end

pclean(arg) do
	Depth := Depth-arg;
	if (arg) emiti(CLEANUP_SEQ, arg);
end

fixargs() do var op, narg, i, nra;
	op := rdch();
	if (op \= tcode.ICLEAN)
		internal("CLEAN instruction required by fixargs()");
	narg := rdarg();
	if (\narg) return 0;
	pclean(nra);
	return narg;
end

declinfprocs() do var y, i, j, b::100, cmbit;
	for (i=0, Infmax, INFPROC) do
		y := @Infprocs[i];
		Out.writes("C ");
		Out.writes(y[INFNAME]);
		Out.writes("(C *SELF");
		cmbit := 1;
		for (j=1, y[INFNARG]) cmbit := cmbit << 1;
		for (j=0, y[INFNARG]) do
			Out.writes(str.format(b, ", C%sa%d", [(
				y[INFCMAP] & cmbit-> " *": " ", j)]));
			cmbit := cmbit >> 1;
		end
		Out.writes(");\n");
	end
end

callinfproc(id) do var ip, i, cmbit, buf::100;
	if (Phase \= 2) return;
	ip := findinfproc(id);
	Out.writes("A = ");
	Out.writes(ip[INFNAME]);
	Out.writes("((C*)M[S]");
	cmbit := 1;
	for (i=1, ip[INFNARG]) cmbit := cmbit << 1;
	for (i=0, ip[INFNARG]) do
		Out.writes(str.format(buf, ", %sM[S+%d]", [(
			ip[INFCMAP] & cmbit-> "(C*)": "",
			i+1)]));
		cmbit := cmbit >> 1;
	end
	Out.writes(");\n");
end

pcall(op, arg) do var narg;
	if (Queued) spill(Iqueue, Aqueue);
	ie (op = tcode.ICALR) do
		if (\Active) do
			if (\Depth)
				internal("no active register in PCALL(CALR)");
			pop();
		end
		emit(CALL_REG, 0);
	end
	else ie (op = tcode.ICALL) do
		if (Active) push();
		emit(CALL_SEQ, [(arg)]);
	end
	else ie (op = tcode.ISYS) do
		internal("The SYS instruction is not supported by TXCGUC");
	end
	else do ! op = tcode.IICALL
		if (Active) push();
		callinfproc(arg);
	end
	Active := 1;
end

branch(op, arg) do
	if (Queued) spill(Iqueue, Aqueue);
	ie (op = tcode.IJUMP) do
		if (arg \= 0) emit(JMP_SEQ, [(Aprefix, arg)]);
		if (Depth \/ Active) pop();
	end
	else do
		if (\Active) do
			if (\Depth) internal("too few operands in BRANCH()");
			pop();
		end
		emit(SAVE_CMP_RESULT, 0);
		if (op = tcode.IBRT \/ op = tcode.IBRF) pop();
		emit(op=tcode.IBRT \/
			op=tcode.INBRT-> BNE_SEQ:
			BE_SEQ, [(Aprefix, arg)]);
	end
end

loop_regs(op, a1) do
	if (Phase = 2) do
		_emit(op=tcode.IUNEXT-> CMP_GE_SEQ[MODE_STACK]:
			CMP_LE_SEQ[MODE_STACK], 0);
		emit(BNE_SEQ, [(Aprefix, a1)]);
	end
	Depth := Depth-1;
	pop();
end

genstr(s, k) do var i, j, b::100;
	if (Phase \= 1) do
		dgrow((k+Lbpw)/Lbpw);
		return;
	end
	Out.writes(str.format(b, "memcpy(&M[%d], \"", [(Dtop)]));
	for (i=0, k) do
		ie (s::i < 32 \/ s::i > 126) do
			Out.wrch('\\');
			str.numtostr(b, s::i, 8);
			for (j=str.length(b), 3) Out.wrch('0');
			Out.writes(b);
		end
		else ie (s::i = '"') do
			Out.writes("\\\"");
		end
		else ie (s::i = '\\') do
			Out.writes("\\\\");
		end
		else do
			Out.wrch(s::i);
		end
	end
	Out.writes("\\0");
	k := k+1;
	while (k & Lbpw-1) do
		Out.writes("\\0");
		k := k+1;
	end
	Out.writes(str.format(b, "\", %d);\n", [(k)]));
	dgrow(k/Lbpw);
end

addproto(id) do var p;
	if (Ptop >= MAXPROTO) error("too many functions");
	Prototbl[Ptop] := id;
	Ptop := Ptop + 1;
end

beginproc(label, pub) do var b::100, i, op, narg;
	op := rdch();
	ie (op = tcode.IHINT)
		narg := rdarg();
	else
		reject(op);
	if (Phase = 1) do
		addproto(label);
		return;
	end
	if (Phase = 0) return;
	Out.writes(str.format(b, "void P%d(void) {\n", [(label)]));
	Out.writes("--S;\n");
	Out.writes("push(F);\n");
	Out.writes("F = S;\n");
	if (pub) do
		Out.writes("push((C)I);\n");
		Out.writes("I = (C*)M[F+2];\n");
	end
end

generate(op, a1, a2, sbuf) do
	ie (	op = tcode.ILDG \/
		op = tcode.ILDGV \/
		op = tcode.ILDL \/
		op = tcode.ILDLV \/
		op = tcode.ILDI \/
		op = tcode.ILDIV \/
		op = tcode.ILDLAB \/
		op = tcode.INUM \/
		op = tcode.ISELF
	) do
		load(op, a1);
	end
	else ie (op = tcode.INEG \/
		op = tcode.ILNOT \/
		op = tcode.IBNOT
	) do
		unary(op);
	end
	else ie (op = tcode.INORMB \/
		op = tcode.INORM \/
		op = tcode.IMUL \/
		op = tcode.IDIV \/
		op = tcode.IMOD \/
		op = tcode.IADD \/
		op = tcode.ISUB \/
		op = tcode.IBAND \/
		op = tcode.IBOR \/
		op = tcode.IBXOR \/
		op = tcode.IBSHL \/
		op = tcode.IBSHR \/
		op = tcode.IEQU \/
		op = tcode.INEQU \/
		op = tcode.ILESS \/
		op = tcode.IGRTR \/
		op = tcode.ILTEQ \/
		op = tcode.IGTEQ \/
		op = tcode.IUMUL \/
		op = tcode.IUDIV \/
		op = tcode.IULESS \/
		op = tcode.IUGRTR \/
		op = tcode.IULTEQ \/
		op = tcode.IUGTEQ \/
		op = tcode.IDREFB \/
		op = tcode.IDEREF
	) do
		binary(op);
	end
	else ie (op = tcode.ISAVG \/
		op = tcode.ISAVL \/
		op = tcode.ISAVI \/
		op = tcode.ISTORE \/
		op = tcode.ISTORB
	) do
		store(op, a1);
	end
	else ie (op = tcode.ICALL \/
		op = tcode.ICALR \/
		op = tcode.IICALL \/
		op = tcode.ISYS
	) do
		pcall(op, a1);
	end
	else ie (op = tcode.IPOP) do
		if (Queued) spill(Iqueue, Aqueue);
		pop();
	end
	else ie (op = tcode.ICLEAN) do
		pclean(a1);
	end
	else ie (op = tcode.ISTACK) do
		emitl(ADJUST_SEQ, -a1);
	end
	else ie (op = tcode.IINCG) do
		emit(INC_GLOB, [(lookup(a1)),(a2)]);
	end
	else ie (op = tcode.IINCI) do
		emiti2(INC_IVAR, a1, a2);
	end
	else ie (op = tcode.IINCL) do
		emitl2(INC_LOC, -a1, a2);
	end
	else ie (op = tcode.IUNEXT \/
		op = tcode.IDNEXT
	) do
		if (Queued) spill(Iqueue, Aqueue);
		if (\Active \/ \Depth)
			internal("too few operands in GENERATE(*NEXT)");
		loop_regs(op, a1);
	end
	else ie (op = tcode.IBRF \/
		op = tcode.IBRT \/
		op = tcode.INBRF \/
		op = tcode.INBRT \/
		op = tcode.IJUMP
	) do
		branch(op, a1);
	end
	else ie (op = tcode.IHDR) do
		internal("HDR without CLAB");
	end
	else ie (op = tcode.IEND) do
		emit(EPILOGUE_SEQ, 0);
		if (Depth) internal("unbalanced stack at END");
		Active := 0;
	end
	else ie (op = tcode.IMHDR) do
		internal("MHDR without CLAB");
	end
	else ie (op = tcode.IENDM) do
		emit(MEPILOGUE_SEQ, 0);
		if (Depth) internal("unbalanced stack at ENDM");
		Active := 0;
	end
	else ie (op = tcode.IINIT) do
		if (a1 \= 7) error("unsupported Tcode version");
	end
	else ie (op = tcode.ICLAB) do
		if (Queued) spill(Iqueue, Aqueue);
		op := rdch();
		ie (op = tcode.IHDR) do
			beginproc(a1, 0);
			Active := 0;
		end
		else ie (op = tcode.IMHDR) do
			beginproc(a1, 1);
			Active := 0;
		end
		else ie (a1 = 0) do
			reject(op);
			if (Phase = 2)
				Out.writes(
				"void _TXSTART_(void) {\ninit();\n");
			Active := 0;
		end
		else do
			reject(op);
			emit(DECL_CODE_LABEL, [(a1)]);
		end
	end
	else ie (op = tcode.IDLAB) do
		addsym(a1);
	end
	else ie (op = tcode.IVEC) do
		dgrow(a1=0-> 1: a1);
	end
	else ie (op = tcode.IDATA \/
		op = tcode.IDREF \/
		op = tcode.ICREF
	) do
		ie (op = tcode.IDATA) do
			emitd(DECL_WORD, [(Dtop, a1)]);
			dgrow(1);
		end
		else ie (op = tcode.IDREF) do
			emitd(DECL_DREF, [(Dtop, lookup(a1))]);
			dgrow(1);
		end
		else do
			emitd(DECL_CREF, [(Dtop, a1)]);
			dgrow(1);
		end
	end
	else ie (op = tcode.ISTR) do
		genstr(sbuf, a1);
	end
	else ie (op = tcode.IGLUE) do
		! nop
	end
	else ie (op = tcode.IHALT) do
		emit(HALT_SEQ, [(a1)]);
	end
	else ie (op = tcode.IDUP) do
		if (Queued) spill(Iqueue, Aqueue);
		push();
		Active := 1;
	end
	else ie (op = tcode.ISWAP) do
		if (Queued) spill(Iqueue, Aqueue);
		if (Depth < 1) internal("too few operands in GENERATE(SWAP)");
		emit(SWAP_SEQ, 0);
	end
	else ie (op = tcode.IIPROC) do var n, m;
		op := rdch();
		if (op \=tcode.ICMAP) internal("no CMAP after IPROC");
		n := rdarg();
		m := rdarg();
		addinfproc(a1, a2, n, m, sbuf);
	end
	else ie (op = tcode.IHINT \/ op = tcode.ICMAP \/
		op = tcode.ILIB \/ op = tcode.IPUB \/
		op = tcode.IEXT
	) do
		! ignore
	end
	else ie (op = tcode.ICALX) do
		error("unresolved CALX (external reference)");
	end
	else ie (op = tcode.IICALX) do
		error("unresolved ICALX (external interface)");
	end
	else do
		internal("unknown instruction in GENERATE()");
	end
end

xlate() do
	var	op, a1, a2, i, j, ch, k;
	var	sbuf::256;

	reset();
	op := rdch();
	if (op \= tcode.IINIT) error("magic match failed");
	while (\In.eof()) do
		if (op & 127 >= tcode.IENDOFSET) error("invalid opcode");
		if (op & 128) do
			a1 := rdarg();
		end
		ie (    op = tcode.IINCL \/ op = tcode.IINCG \/
			op = tcode.IINCI \/ op = tcode.IINIT \/
			op = tcode.ICMAP
		) do
			a2 := rdarg();
		end
		else if (op >= tcode.IINIT \/ op = tcode.ISTR \/
			op = tcode.ILIB
		) do
			ie (op = tcode.ISTR \/ op = tcode.ILIB)
				a2 := a1;
			else
				a2 := rdarg();
			for (i=0, a2) sbuf::i := rdch();
			sbuf::i := 0;
		end
		generate(op, a1, a2, sbuf);
		op := rdch();
	end
end

declprotos() do var i, b::100;
	for (i=0, Ptop)
		Out.writes(str.format(b, "void P%d(void);\n",
			[(Prototbl[i])]));
end

phase0() do
	Out.create(T3X.SYSOUT, Outbuf, BUFSIZE, IOSTREAM.FWRITE);
	if (Tmp.open(Tmpfile, Tmpbuf, BUFSIZE, IOSTREAM.FWRITE))
		error("failed to create work file");
	In.create(T3X.SYSIN, Inbuf, BUFSIZE, IOSTREAM.FREAD);
	Phase := 0;
	xlate();
	Tmp.close();
	In.flush();
	Dsize := Dtop;
end

phase1() do
	if (In.open(Tmpfile, Inbuf, BUFSIZE, IOSTREAM.FREAD))
		error("failed to re-open work file");
	Phase := 1;
	emitd(BANNER, 0);
	emitd(HEAD, [(Dsize + STACKSIZE)]);
	xlate();
	emitd(TAIL, 0);
	In.close();
end

phase2() do
	if (In.open(Tmpfile, Inbuf, BUFSIZE, IOSTREAM.FREAD))
		error("failed to re-open work file");
	Phase := 2;
	declinfprocs();
	declprotos();
	xlate();
	emit(TAIL, 0);
	In.close();
	Out.flush();
end

do var buf::4;
	init();
	if (t.getarg(1, buf, 4) = 2 /\ \t.memcomp(buf, "-v", 3))
		Verbose := 1;
	phase0();
	phase1();
	phase2();
	t.remove(Tmpfile);
	if (Verbose) do var b::100;
		t.write(T3X.SYSERR, str.format(b, "data=%UD%S",
			[(Dsize*Lbpw, t.newline(buf))]), str.length(b));
	end
end
