# install script for @{IDENT}
# Nils M Holm, 2003, 2019
# See the file LICENSE for conditions of use.

# TXBASE	where to install the package
# TXDIR		don't change
# BINDIR	location of the binaries (tx, txx, etc)
# INCDIR	where the local include files go
# MANDIR	where to place the man page
# OWNER		owner (uid) of the T3X files
# GROUP		group (gid) of the T3X files

if [ "$INSTALL_PREFIX" = "" ]; then
	PREFIX=/usr/local
else
	PREFIX=$INSTALL_PREFIX
fi
TXBASE=$PREFIX/t3x
TXDIR=$TXBASE/r7
BINDIR=$PREFIX/bin
INCDIR=`eval echo @{INCDIR}`
MANDIR=`eval echo @{MANDIR}`
if [ "@{MANSECT}" != "" ]; then
	MANPATH="$MANDIR/man@{MANSECT}"
else
	MANPATH=$MANDIR
fi
OWNER=bin
GROUP=bin

TXCM="../tools/txcm -b $TXDIR"

shell_fn_test() {
	return 203
}
shell_fn_test >/dev/null 2>&1
if [ $? != 203 ]; then
	echo
	echo "The installation script requires a Bourne-compatible shell"
	echo "that supports shell functions."
	echo
	echo "On your system, you may try one out of these shells:"
	for p in /bin /usr/bin /usr/local/bin ${PREFIX}/bin; do
		for s in ksh bash; do
			if [ -f $p/$s ]; then echo $p/$s; fi
		done
	done
	echo
	exit 1
fi

id | grep root >/dev/null
if [ $? != 0 ]; then
	echo 
	echo "The install script must be run by root."
	echo
	exit
fi

(cd ../classes; for cls in char iostream memory string system ttyctl xmem ; do
	if [ -f $cls.cls -a ! -f ${cls}.ndx ]; then
		echo $cls:t3x >$cls.ndx
	fi
 done)

(cd ../classes; echo util:t3x,string,iostream >util.ndx)

install() {
	if [	"x$1" != "x-o" -o \
		"x$3" != "x-g" -o \
		"x$6" != "x-m" ]
	then
		echo install: bad arguments
		exit 1
	fi
	if [ "x$5" = "x-d" ]; then
		if [ ! -d $8 ]; then
			mkdir $8
			chown $2 $8
			chgrp $4 $8
			chmod $7 $8
		fi
	elif [ "x$5" = "x-C" ]; then
		if [ -d $9 ]; then
			file=$9/`basename $8`
		else
			file=$9
		fi
		cp $8 $file
		chown $2 $file
		chgrp $4 $file
		chmod $7 $file
	else
		echo install: bad arguments
		exit 1
	fi
}

BINLINK="@{SYMLINK_CMD}"
ln -fs Makefile .symtest.$$ >/dev/null 2>&1
if [ $? != 0 ]; then
	BINLINK=binlink
fi
rm -f .symtest.$$

binlink() {
	echo "#!/bin/sh" >$2
	echo "$1 \$*" >>$2
	chmod 0755 $2
	chown bin $2
	chgrp bin $2
}


@{PREPARE_INSTALL}

if [ ! -d $PREFIX ]; then install -o $OWNER -g $GROUP -d -m 0755 $PREFIX; fi
if [ "$INSTALL_PREFIX" != "" -a ! -d $PREFIX/man ];
	then install -o $OWNER -g $GROUP -d -m 0755 $PREFIX/man
fi
if [ ! -d $INCDIR ]; then
	install -o $OWNER -g $GROUP -d -m 0755 $INCDIR
fi
if [ ! -d $BINDIR ]; then install -o $OWNER -g $GROUP -d -m 0755 $BINDIR; fi
if [ ! -d $MANDIR ]; then install -o $OWNER -g $GROUP -d -m 0755 $MANDIR; fi
if [ ! -d $MANPATH ]; then install -o $OWNER -g $GROUP -d -m 0755 $MANPATH; fi

install -o $OWNER -g $GROUP -d -m 0755 $TXBASE
install -o $OWNER -g $GROUP -d -m 0755 $TXDIR
install -o $OWNER -g $GROUP -d -m 0755 $TXDIR/bin
install -o $OWNER -g $GROUP -d -m 0755 $TXDIR/classes
install -o $OWNER -g $GROUP -d -m 0755 $TXDIR/lib
install -o $OWNER -g $GROUP -d -m 0755 $TXDIR/doc

install -o $OWNER -g $GROUP -C -m 0644 t3x.dir $TXDIR/classes
install -o $OWNER -g $GROUP -C -m 0755 txdpf $TXDIR/bin
install -o $OWNER -g $GROUP -C -m 0755 txld $TXDIR/bin
install -o $OWNER -g $GROUP -C -m 0755 txopt $TXDIR/bin
install -o $OWNER -g $GROUP -C -m 0755 txtrn $TXDIR/bin
install -o $OWNER -g $GROUP -C -m 0755 txx $TXDIR/bin
$BINLINK $TXDIR/bin/txx $BINDIR/txx

install -o $OWNER -g $GROUP -C -m 0644 ../compiler/tcode.dir $TXDIR/classes
touch $TXDIR/classes/tcode.cls
touch $TXDIR/classes/CLASSNDX
(cd ../compiler; $TXCM -qi t3x)
(cd ../classes; $TXCM -i char)
(cd ../classes; $TXCM -i iostream)
(cd ../classes; $TXCM -i memory)
(cd ../classes; $TXCM -i string)
(cd ../classes; $TXCM -i util)
(cd ../classes; $TXCM -i system)
(cd ../classes; $TXCM -i ttyctl)
(cd ../classes; $TXCM -i xmem)

install -o $OWNER -g $GROUP -C -m 0755 ../native/txcg@{TARGET} $TXDIR/bin
if [ @{TARGET} = 386a -o @{TARGET} = 386c -o @{TARGET} = 386e ]; then
	rm -f $TXDIR/bin/txcg386
	@{LINK_CMD} $TXDIR/bin/txcg@{TARGET} $TXDIR/bin/txcg386
fi
rm -f $TXDIR/bin/txcgl
@{LINK_CMD} $TXDIR/bin/txcg@{TARGET} $TXDIR/bin/txcgl
install -o $OWNER -g $GROUP -C -m 0755 ../native/txcgexe $TXDIR/bin
rm -f $TXDIR/bin/txcg86
@{LINK_CMD} $TXDIR/bin/txcgexe $TXDIR/bin/txcg86
install -o $OWNER -g $GROUP -C -m 0755 ../native/txcguc $TXDIR/bin

install -o $OWNER -g $GROUP -C -m 0755 ../native/ts86 $TXDIR/bin
$BINLINK $TXDIR/bin/ts86 $BINDIR/ts86
# should be done by TXCM, but sources are in wrong directory
install -o $OWNER -g $GROUP -C -m 0644 ../native/t3x.x $TXDIR/lib
install -o $OWNER -g $GROUP -C -m 0644 ../native/system.x $TXDIR/lib
install -o $OWNER -g $GROUP -C -m 0644 ../native/ttyctl.x $TXDIR/lib
install -o $OWNER -g $GROUP -C -m 0644 ../native/xmem.x $TXDIR/lib

if [ -f ../tide/tide ]; then
        (cd ../tide;
	 for cls in *.cls; do
		 $TXCM -i `basename $cls .cls`
	 done)
	if [ ! -d $TXDIR/tide ]; then
		install -o $OWNER -g $GROUP -d -m 0755 $TXDIR/tide
	fi
	install -o $OWNER -g $GROUP -C -m 0644 ../tide/helpfile $TXDIR/tide
	install -o $OWNER -g $GROUP -C -m 0644 ../tide/tide_rc \
						$TXDIR/tide.rc
	install -o $OWNER -g $GROUP -C -m 0644 ../tide/default.cs $TXDIR/tide
	install -o $OWNER -g $GROUP -C -m 0644 ../tide/monochrm.cs $TXDIR/tide
	install -o $OWNER -g $GROUP -C -m 0755 ../tide/tide $TXDIR/bin/tide
	install -o $OWNER -g $GROUP -C -m 0755 ../tide/tideconf \
						$TXDIR/bin/tideconf
	$BINLINK $TXDIR/bin/tide $BINDIR/tide
fi

chown bin $TXDIR/classes/* ; chgrp bin $TXDIR/classes/*
chown bin $TXDIR/lib/* ; chgrp bin $TXDIR/lib/*

install -o $OWNER -g $GROUP -C -m 0755 ../tools/tx $TXDIR/bin
$BINLINK $TXDIR/bin/tx $BINDIR/tx
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txmake $TXDIR/bin
$BINLINK $TXDIR/bin/txmake $BINDIR/txmake
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txhelp $TXDIR/bin
$BINLINK $TXDIR/bin/txhelp $BINDIR/txhelp
install -o $OWNER -g $GROUP -C -m 0644 ../tools/txprint.rules $TXDIR
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txcm $TXDIR/bin
$BINLINK $TXDIR/bin/txcm $BINDIR/txcm
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txdump $TXDIR/bin
$BINLINK $TXDIR/bin/txdump $BINDIR/txdump
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txprint $TXDIR/bin
$BINLINK $TXDIR/bin/txprint $BINDIR/txprint
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txsdfmt $TXDIR/bin
$BINLINK $TXDIR/bin/txsdfmt $BINDIR/txsdfmt
install -o $OWNER -g $GROUP -C -m 0755 ../tools/txsum $TXDIR/bin
$BINLINK $TXDIR/bin/txsum $BINDIR/txsum

install -o $OWNER -g $GROUP -C -m 0644 ../ports/@{PORT}/txrc \
	$TXDIR/system.txrc
install -o $OWNER -g $GROUP -C -m 0644 ../ports/@{PORT}/txrc.dos \
	$TXDIR/dos.txrc
install -o $OWNER -g $GROUP -C -m 0644 ../ports/@{PORT}/txrc.uc \
	$TXDIR/uc.txrc
if [ -f ../ports/@{PORT}/txrc.static ]; then
	install -o $OWNER -g $GROUP -C -m 0644 ../ports/@{PORT}/txrc.static \
		$TXDIR/static.txrc
fi
install -o $OWNER -g $GROUP -C -m 0644 ../ports/@{PORT}/txrc.dos \
	$TXDIR/dos.txrc

install -o $OWNER -g $GROUP -C -m 0644 ../LICENSE $TXDIR/doc
install -o $OWNER -g $GROUP -C -m 0644 ../docs/t3x-book.txt $TXDIR/doc
(cd ../docs; for docfile in *.sd; do
	install -o $OWNER -g $GROUP -C -m 0644 $docfile $TXDIR/doc
 done)
install -o $OWNER -g $GROUP -C -m 0644 ../docs/sdtoc $TXDIR/doc/.sdtoc

install -o $OWNER -g $GROUP -C -m 0644 txx.h @{INCDIR}

install -o $OWNER -g $GROUP -C -m 0644 ../docs/t3x.7 $MANPATH/t3x.7

strip $TXDIR/bin/* >/dev/null 2>&1

@{FINISH_INSTALL}

echo
echo "Don't forget to define T3XDIR = $TXDIR ."
echo
echo "You may want to install $TXDIR/system.txrc as ~/.txrc, too."
echo

