/*
 * XMEM -- External Memory Interface.
 * Nils M Holm, 2000,2002,2003,2019.
 * See the file LICENSE for conditions of use.
 */

#ifdef EXT_XMEM

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef dos
 #include <alloc.h>
#endif

#include "../compiler/txx.h"

#define MAXBLOCKS	16

uchar	*Blocks[MAXBLOCKS];

xcell xmem_init S0()
SELF
{
	int	i;

	for (i=0; i<MAXBLOCKS; i++) Blocks[i] = NULL;
	return 0;
}

xcell xmem_alloc S1(size)
SELF
xucell	size;
{
	int	i;

	size &= 0xffff;
	for (i=0; i<MAXBLOCKS; i++)
		if (Blocks[i] == NULL) break;
	if (i >= MAXBLOCKS) return -1;
	if ((Blocks[i] = (uchar *) malloc(size)) == NULL)
		return -1;
	return i;
}

xcell xmem_free S1(id)
SELF
xcell	id;
{
	if (id < 0 || id >= MAXBLOCKS || Blocks[id] == NULL) return -1;
	free(Blocks[id]);
	Blocks[id] = NULL;
	return 0;
}

xcell xmem_get S2(id, x)
SELF
xcell	id;
xucell	x;
{
	if (id < 0 || id >= MAXBLOCKS || Blocks[id] == NULL) return -1;
	return Blocks[id][x & 0xffff];
}

xcell xmem_put S3(id, x, v)
SELF
xcell	id;
xucell	x, v;
{
	if (id < 0 || id >= MAXBLOCKS || Blocks[id] == NULL) return -1;
	return Blocks[id][x & 0xffff] = v;
}

xcell xmem_read S4(id, x, v, k)
SELF
xcell	id;
xucell	x;
uchar	*v;
xucell	k;
{
	if (id < 0 || id >= MAXBLOCKS || Blocks[id] == NULL) return -1;
	memcpy(v, &Blocks[id][x & 0xffff], k);
	return 0;
}

xcell xmem_write S4(id, x, v, k)
SELF
xcell	id;
xucell	x;
uchar	*v;
xucell	k;
{
	if (id < 0 || id >= MAXBLOCKS || Blocks[id] == NULL) return -1;
	memcpy(&Blocks[id][x & 0xffff], v, k);
	return 0;
}

xcell xmem_copy S4(id, d, s, k)
SELF
xcell	id;
xucell	d;
xucell	s;
xucell	k;
{
	if (id < 0 || id >= MAXBLOCKS || Blocks[id] == NULL) return -1;
	memcpy(&Blocks[id][d & 0xffff], &Blocks[id][s & 0xffff], k);
	return 0;
}

xcell xmem_fini S0()
SELF
{
	return 0;
}

#ifndef LIBRARY

struct infproc	InfXMem[] = {
	{ "xmem_alloc",	xmem_alloc	},
	{ "xmem_fini",	xmem_fini	},
	{ "xmem_free",	xmem_free	},
	{ "xmem_get",	xmem_get	},
	{ "xmem_init",	xmem_init	},
	{ "xmem_put",	xmem_put	},
	{ "xmem_read",	xmem_read	},
	{ "xmem_write",	xmem_write	},
	{ "xmem_copy",	xmem_copy	},
	{ NULL,		NULL		}
};

#endif

#else	/* EXT_XMEM */

#include "../compiler/txx.h"

#ifndef LIBRARY
struct infproc	InfXMem[MAXINFPROC] = { { 0L, 0L} };
#endif

#endif	/* EXT_XMEM */
