#!/bin/sh

# Installation Script for the Generic Bytecode Port
# Nils M Holm, 2003, 2019
# See the file LICENSE for conditions of use.

# install script for the Tcode port

# TXBASE	where to install the package
# TXDIR		don't change
# BINDIR	location of the binaries (tx, txx, etc)
# INCDIR	where the local include files go
# DOCDIR	online documents
# MANDIR	base directory of the manual pages
# MANSECT	in which section to place the man page
# OWNER		owner:group of the T3X files
# GROUP		group (gid) of the T3X files

TXBASE=/usr/local/t3x
TXDIR=$TXBASE/r7
BINDIR=/usr/local/bin
DOCDIR=$TXDIR/doc
MANDIR=/usr/local/man
MANSECT=7
OWNER=bin
GROUP=bin
MANPATH="$MANDIR/man$MANSECT"

TXCM="../compiler/txx ../tools/txcm"

rm -f $BINDIR/tx $BINDIR/txx $BINDIR/txcm $BINDIR/txdump $BINDIR/txprint \
	$BINDIR/txsdfmt $BINDIR/txsum $BINDIR/txmake $BINDIR/txhelp \
	$BINDIR/ts86 $BINDIR/tide $BINDIR/tideconf

mkdir $TXBASE && chmod 0755 $TXBASE && chown $OWNER:$GROUP $TXBASE
mkdir $TXDIR && chmod 0755 $TXDIR && chown $OWNER:$GROUP $TXDIR
mkdir $TXDIR/bin && chmod 0755 $TXDIR/bin && chown $OWNER:$GROUP $TXDIR/bin
mkdir $TXDIR/classes && chmod 0755 $TXDIR/classes && \
	chown $OWNER:$GROUP $TXDIR/classes
mkdir $TXDIR/lib && chmod 0755 $TXDIR/lib && chown $OWNER:$GROUP $TXDIR/lib
mkdir $DOCDIR && chmod 0755 $DOCDIR && chown $OWNER:$GROUP $DOCDIR
mkdir $MANDIR && chmod 0755 $MANDIR && chown $OWNER:$GROUP $MANDIR
mkdir $MANPATH && chmod 0755 $MANPATH && chown $OWNER:$GROUP $MANPATH

touch $TXDIR/classes/CLASSNDX
cp tcode.dir $TXDIR/classes
touch $TXDIR/classes/tcode.cls
(cd ../compiler; $TXCM -qi t3x)
for cls in char iostream memory string system ttyctl xmem; do
	(cd ../classes; echo $cls:t3x >$cls.ndx)
	(cd ../classes; $TXCM -i $cls)
done
(cd ../classes; echo util:t3x,string,iostream >util.ndx)
(cd ../classes; $TXCM -i util)
chmod 0644 $TXDIR/classes/*
chown $OWNER:$GROUP $TXDIR/classes/*

cp ../native/t3x.x $T3XDIR/lib
cp ../native/system.x $T3XDIR/lib
cp ../native/ttyctl.x $T3XDIR/lib
cp ../native/xmem.x $T3XDIR/lib

chown $OWNER:$GROUP $TXDIR//b*
chmod 0644 $TXDIR/lib/*

cp ../compiler/txx $TXDIR/bin
cp ../tools/tx $TXDIR/bin
cp ../tools/txcm $TXDIR/bin
cp ../tools/txdump $TXDIR/bin
cp ../tools/txprint $TXDIR/bin
cp ../tools/txsdfmt $TXDIR/bin
cp ../tools/txsum $TXDIR/bin
if [ -f ../tools/txmake ]; then
	cp ../tools/txmake $TXDIR/bin
fi
if [ -f ../tools/txhelp ]; then
	cp ../tools/txhelp $TXDIR/bin
fi
cp ../compiler/txtrn $TXDIR/bin
cp ../compiler/txld $TXDIR/bin
cp ../compiler/txopt $TXDIR/bin
cp ../compiler/txdpf $TXDIR/bin
cp ../native/txcgexe $TXDIR/bin
cp ../native/txcguc $TXDIR/bin
cp ../native/ts86 $TXDIR/bin
chown $OWNER:$GROUP $TXDIR/bin/*
chmod 0644 $TXDIR/bin/*
chmod 0755 $TXDIR/bin/txx $TXDIR/bin/tx

cp $TXDIR/bin/txx $BINDIR/txx && chown $OWNER:$GROUP $BINDIR/txx && \
	chmod 0755 $BINDIR/txx

if [ -f ../tools/tx2 ]; then
	echo "#! /bin/sh" >$BINDIR/tx
	echo "$TXDIR/bin/txx $TXDIR/bin/tx \$*" >>$BINDIR/tx
	chown $OWNER:$GROUP $BINDIR/tx
	chmod 0755 $BINDIR/tx
	chmod 0644 $T3XDIR/bin/tx
else
	cp $TXDIR/bin/tx $BINDIR/tx && chown $OWNER:$GROUP $BINDIR/tx && \
		chmod 0755 $BINDIR/tx
fi

echo "#! /bin/sh" >$BINDIR/txcm
echo "$TXDIR/bin/txx $TXDIR/bin/txcm \$*" >>$BINDIR/txcm
echo "#! /bin/sh" >$BINDIR/txdump
echo "$TXDIR/bin/txx $TXDIR/bin/txdump \$*" >>$BINDIR/txdump
echo "#! /bin/sh" >$BINDIR/txprint
echo "$TXDIR/bin/txx $TXDIR/bin/txprint \$*" >>$BINDIR/txprint
echo "#! /bin/sh" >$BINDIR/txsdfmt
echo "$TXDIR/bin/txx $TXDIR/bin/txsdfmt \$*" >>$BINDIR/txsdfmt
echo "#! /bin/sh" >$BINDIR/txsum
echo "$TXDIR/bin/txx $TXDIR/bin/txsum \$*" >>$BINDIR/txsum
if [ -f ../tools/txmake ]; then
	echo "#! /bin/sh" >$BINDIR/txmake
	echo "$TXDIR/bin/txx $TXDIR/bin/txmake \$*" >>$BINDIR/txmake
 	chown $OWNER:$GROUP $BINDIR/txmake
 	chmod 0755 $BINDIR/txmake
fi
if [ -f ../tools/txhelp ]; then
	echo "#! /bin/sh" >$BINDIR/txhelp
	echo "$TXDIR/bin/txx $TXDIR/bin/txhelp \$*" >>$BINDIR/txhelp
 	chown $OWNER:$GROUP $BINDIR/txhelp
 	chmod 0755 $BINDIR/txhelp
fi
echo "#! /bin/sh" >$BINDIR/ts86
echo "$TXDIR/bin/txx $TXDIR/bin/ts86 \$*" >>$BINDIR/ts86
chown $OWNER:$GROUP $BINDIR/txcm $BINDIR/txdump $BINDIR/txprint \
	$BINDIR/txsdfmt $BINDIR/txsum $BINDIR/ts86
chmod 0755 $BINDIR/txcm $BINDIR/txdump $BINDIR/txprint $BINDIR/txsdfmt \
	$BINDIR/txsum $BINDIR/ts86

cp ../tools/txprint.rules $TXDIR/txprint.rules && \
	chown $OWNER:$GROUP $TXDIR/txprint.rules && \
	chmod 0644 $TXDIR/txprint.rules && \

cp ../LICENSE $DOCDIR
cp ../docs/*.sd $DOCDIR
cp ../docs/sdtoc $DOCDIR/.sdtoc
cp ../docs/t3x-book.txt $DOCDIR
chown $OWNER:$GROUP $DOCDIR/*
chmod 0644 $DOCDIR/*

if [ -f ../tide/tide ]; then
	(cd ../tide;
	 for cls in *.cls; do
		$TXCM -i `basename $cls .cls`
 	 done)
	mkdir $TXDIR/tide && \
		chmod 0755 $TXDIR/tide && \
		chown $OWNER:$GROUP $TXDIR/tide
	cp ../tide/helpfile $TXDIR/tide
	sed -e 's/native = 1/native = 0/' <../tide/tide_rc >$TXDIR/tide.rc
	chmod 0644 $TXDIR/tide.rc && chown $OWNER:$GROUP $TXDIR/tide.rc
	cp ../tide/*.cs $TXDIR/tide
	chmod 0644 $TXDIR/tide/* && chown $OWNER:$GROUP $TXDIR/tide/*
	cp ../tide/tide $TXDIR/bin/tide
	chmod 0644 $TXDIR/bin/tide && chown $OWNER:$GROUP $TXDIR/bin/tide
	echo "#! /bin/sh" >$BINDIR/tide
	echo "$TXDIR/bin/txx $TXDIR/bin/tide \$*" >>$BINDIR/tide
 	chown $OWNER:$GROUP $BINDIR/tide
 	chmod 0755 $BINDIR/tide
	cp ../tide/tideconf $TXDIR/bin/tideconf
	chmod 0644 $TXDIR/bin/tideconf
	chown $OWNER:$GROUP $TXDIR/bin/tideconf
fi

cp ../ports/Generic-unix/txrc $T3XDIR/system.txrc
chown $OWNER:$GROUP $T3XDIR/system.txrc
chmod 0644 $T3XDIR/system.txrc

cp ../ports/Generic-unix/txrc.dos $T3XDIR/dos.txrc
chown $OWNER:$GROUP $T3XDIR/dos.txrc
chmod 0644 $T3XDIR/dos.txrc

cp ../ports/Generic-unix/txrc.uc $T3XDIR/uc.txrc
chown $OWNER:$GROUP $T3XDIR/uc.txrc
chmod 0644 $T3XDIR/uc.txrc

cp ../docs/t3x.7 $MANPATH/t3x.7
chown $OWNER:$GROUP $MANPATH/t3x.7
chmod 0644 $MANPATH/t3x.7

echo
echo "Don't forget to define T3XDIR = $TXDIR ."
echo
echo "You may want to install $TXDIR/system.txrc as ~/.txrc, too."
echo
