#! /bin/sh

# TX.SH -- A Bourne clone of TX
# Nils M Holm, 2000,2001,2002,2003,2019
# See the file LICENSE for conditions of use
#
# -- GENERIC (bytecode) version --
#
# This is an attempt to simulate the TX compiler driver in a shell script.
# Not all options of the original version are supported.

T3X_VERSION="T3X 7.4.5"

if [ "$T3XDIR" = "" ]; then
	T3XDIR=/usr/local/t3x/r7
fi

BINDIR=$T3XDIR/bin
CLASSDIR=$T3XDIR/classes

TXOPT=""
TXDPF=""
VERBOSE=0
PRESERVE=0
LOAD=1
BANGHDR=NA
MAKECLASS=0
OUTFILE=NA
GENDOSEXE=0
LONGUSAGE=0
LICENSE=0

USAGE=\
"Usage: tx [-?h] [-Ados] [-B dir] [-C dir] [-o file] [-ml] [-DOceipv] file"

while true; do
	case "x$1" in
		x-\?|x-h)
			LONGUSAGE=1
			break
			;;
		x-Ados)	GENDOSEXE=1
			;;
		x-B)	if [ "$2" != "" ]; then
				shift
			fi
			BINDIR=$1
			;;
		x-C)	if [ "$2" != "" ]; then
				shift
			fi
			CLASSDIR=$1
			;;
		x-o)	if [ "$2" != "" ]; then
				shift
			fi
			OUTFILE=$1
			;;
		x-O0)	TXOPT=NA
			;;
		x-D0)	TXDPF=NA
			;;
		x-O|x-O1)
			TXOPT=$TXOPT_DFLT
			;;
		x-D|x-D1)
			TXDPF=$TXDPF_DFLT
			;;
		x-V)	LICENSE=1
			break
			;;
		x-c)	LOAD=0
			;;
		x-i)	MAKECLASS=1
			;;
		x-e)	BANGHDR="#! $BINDIR/txx"
			;;
		x-U)		# ignore -U
			shift	# argument
			;;
		x-ml)	# ignore
			;;
		x-p)	PRESERVE=1
			;;
		x-v)	VERBOSE=1
			echo "$T3X_VERSION"
			;;
		x-vv)	VERBOSE=2
			echo "$T3X_VERSION"
			;;
		x-*|x)
			echo $USAGE
			exit 1
			;;
		x*)	break
			;;
	esac
	shift
done

LIBS="t3x `(cd $CLASSDIR; ls *.cls | sed -e 's/\.cls//' -e 's/t3x//')`"

if [ "$TXOPT" != "NA" ]; then
	TXOPT="$BINDIR/txx $BINDIR/txopt"
fi
if [ "$TXDPF" != "NA" ]; then
	TXDPF="$BINDIR/txx $BINDIR/txdpf"
fi
TXTRN="$BINDIR/txx $BINDIR/txtrn"
TXLD="$BINDIR/txx $BINDIR/txld"
TXCM="$BINDIR/txx $BINDIR/txcm"
TXCGEXE="$BINDIR/txx $BINDIR/txcgexe"

if [ $LONGUSAGE = 1 ]; then
echo
echo $USAGE
cat <<EOT

-Ados   generate DOS executable
-B dir  search binaries in 'dir'
-C dir  search classes in 'dir'
-D[01]  disable/enable dead procedure filter
-O[01]  disable/enable tcode optimization
-U file load user configuration (IGNORED)
-V      display conditions of use
-c      compile only (don't load)
-e      create #! header (Unix only)
-h      display this summary (also -?)
-i      create an installable class
-ml     generate native code (IGNORED)
-o file direct output to 'file'
-p      preserve CLASSDIR and LIBDIR
-v      verbose operation (more v's = more output)

EOT
exit 0
fi

if [ $LICENSE = 1 ]; then
echo
cat <<EOT
T3X -- A Compiler for the Minimal Procedural Language T3X
Nils M Holm 1996-2003

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
EOT
echo
exit 0
fi

FILE=$1
OFILE=`basename $FILE .t`
TMP=tx$$.tmp

if [ ! -f $FILE ]; then
	echo "TX: $FILE: no such file"
	exit 1
fi

if [ ! -f $OFILE.t ]; then
	echo "TX: $FILE: file has no .t suffix"
	exit 1
fi

if [ "$OUTFILE" != "NA" ]; then
	OFILE=$OUTFILE
fi

if [ $PRESERVE = 0 ]; then
	rm -f CLASSDIR LIBDIR
fi

if [ $VERBOSE != 0 ]; then
	echo $TXTRN $FILE '->' $OFILE
fi
$TXTRN <$FILE >$OFILE
if [ $? != 0 ]; then
	rm -f $TMP $OFILE
	echo "TX: error in compilation. stopped."
	exit 1
fi

if [ $MAKECLASS = 1 ]; then
	rm -f $OFILE.cls
	mv $OFILE $OFILE.cls
	if [ $VERBOSE != 0 ]; then
		echo $TXCM -qfx $OFILE CLASSDIR
	fi
	$TXCM -qfx $OFILE CLASSDIR
	if [ $VERBOSE != 0 ]; then
		echo $TXCM -o $OFILE.ndx -n $OFILE CLASSDIR
	fi
	$TXCM -o $OFILE.ndx -n $OFILE CLASSDIR
	if [ $PRESERVE = 0 ]; then
		rm -f CLASSDIR
	fi
	exit 0
fi

if [ $LOAD = 0 ]; then
	exit 0
fi

if [ -f CLASSDIR ]; then
	rm -f $TMP
	strings CLASSDIR >CLASSDIR.names 2>/dev/null
	for libname in $LIBS; do
		if grep "\*$libname\*" CLASSDIR.names >/dev/null; then
			if [ $VERBOSE = 2 ]; then
				echo "(load) $CLASSDIR/$libname.cls"
			fi
			cat $CLASSDIR/$libname.cls >>$TMP
		fi
	done
	rm -f CLASSDIR.names
	cat $OFILE >>$TMP
	mv -f $TMP $OFILE
fi

if [ $VERBOSE != 0 ]; then
	echo $TXLD $FILE '->' $OFILE
fi
mv -f $OFILE $TMP
$TXLD <$TMP >$OFILE
if [ $? != 0 ]; then
	rm -f $TMP $OFILE
	echo "TX: error in compilation. stopped."
	exit 1
fi

if [ "$TXOPT" != "NA" ]; then
	if [ $VERBOSE != 0 ]; then
		echo $TXOPT $FILE '->' $OFILE
	fi
	mv -f $OFILE $TMP
	$TXOPT <$TMP >$OFILE
	if [ $? != 0 ]; then
		rm -f $TMP $OFILE
		echo "TX: error in compilation. stopped."
		exit 1
	fi
fi

if [ "$TXDPF" != "NA" ]; then
	if [ $VERBOSE != 0 ]; then
		echo $TXDPF $FILE '->' $OFILE
	fi
	mv -f $OFILE $TMP
	$TXDPF <$TMP >$OFILE
	if [ $? != 0 ]; then
		rm -f $TMP $OFILE
		echo "TX: error in compilation. stopped."
		exit 1
	fi
fi

if [ "$GENDOSEXE" = "1" ]; then
	if [ $VERBOSE != 0 ]; then
		echo $TXCGEXE $OFILE '->' $OFILE.exe
	fi
	$TXCGEXE <$OFILE >$OFILE.exe
	if [ $? != 0 ]; then
		rm -f $TMP $OFILE $OFILE.exe
		echo "TX: error in compilation. stopped."
		exit 1
	fi
	rm $OFILE
fi

if [ "$BANGHDR" != "NA" ]; then
	mv -f $OFILE $TMP
	echo "$BANGHDR" >$OFILE
	cat $TMP >>$OFILE
	chmod +x $OFILE
fi

if [ $PRESERVE = 0 ]; then
	rm -f CLASSDIR LIBDIR
fi
rm -f $TMP
