! TIDE -- Configuration class
! Nils M Holm, 2000,2002
! See the file LICENSE for conditions of use.

module config();

public class config(t3x, iostream, string, ttyctl)

 object	t[t3x], str[string];

 const	BUFSIZE = 256;

 public struct	CONF =
		C_NAME,
		C_TYPE,
		C_DEFAULT,
		C_VAR,
		C_LIMIT;

 public const	T_STRING = 'S',
		T_NSTRING = 'N',
		T_NUMERIC = '#',
		T_COLOR = 'C';

 var	Set;
 var	Colors;

 public init(iset) do var i;
	Colors := [ "black", "blue", "green", "cyan",
		"red", "magenta", "yellow", "grey" ];
	Set := iset;
	i := 0;
	while (Set[i]) do
		ie (Set[i][C_TYPE] = T_NUMERIC)
			Set[i][C_VAR][0] := Set[i][C_DEFAULT];
		else ie (Set[i][C_TYPE] = T_NSTRING)
			str.ncopy(Set[i][C_LIMIT], Set[i][C_VAR],
				Set[i][C_DEFAULT]);
		else ie (Set[i][C_TYPE] = T_STRING)
			str.copy(Set[i][C_VAR], Set[i][C_DEFAULT]);
		else ie (Set[i][C_TYPE] = T_COLOR)
			Set[i][C_VAR][0] := Set[i][C_DEFAULT];
		else
			return -1;
		i := i+1;
	end
 end

 colorname(value, buf) do
	str.format(buf, "%S%S/%S", [
		(value & ttyctl.F_BRIGHT-> "light-": ""),
		(Colors[value & 0x07]),
		(Colors[value & 0x70 >> 4])
		]);
	return buf;
 end

 xtoi(s) ie (s::0 = '0' /\ (s::1 = 'x' \/ s::1 = 'X'))
	return str.strtonum(@s::2, 16, 0);
 else
	return str.strtonum(s, 10, 0);

 colorvalue(s) do var c, i, s2;
	c := 0;
	if (\t.memcomp(s, "light-", 6)) do
		c := ttyctl.F_BRIGHT;
		s := @s::6;
	end
	i := str.scan(s, '/');
	if (i = -1) return 0x07;
	s2 := @s::(i+1);
	s::i := 0;
	for (i=0, 8) if (\str.comp(Colors[i], s)) leave;
	if (i = 8) return 0x07;
	c := c|i;
	for (i=0, 8) if (\str.comp(Colors[i], s2)) leave;
	if (i = -1) return 0x07;
	return i<<4 | c;
 end

 public load(file, errp) do
	object	cnf[iostream];
	var	cnfbuf::BUFSIZE, line::BUFSIZE;
	var	name, value;
	var	k, i, n, flags;

	flags := iostream.FREAD | iostream.FTRANS;
	if (cnf.open(file, cnfbuf, BUFSIZE, flags) = %1)
		return -2;
	n := 0;
	name := line;
	while (1) do
		k := cnf.reads(line, BUFSIZE);
		n := n+1;
		if (cnf.eof()) leave;
		if (line::(k-1) = '\n') line::(k-1) := 0;
		if (line::0 = 0 \/ line::0 = '#') loop;
		errp[0] := n;
		for (i=0, k) if (line::i = '\s' \/ line::i = '\t') leave;
		if (i = k) return -1;
		line::i := 0;
		for (i=i+1, k) if (line::i = '=') leave;
		if (i = k) return -1;
		for (i=i+1, k) if (line::i \= '\s' /\ line::i \= '\t') leave;
		if (i = k) return -1;
		value := @line::i;
		if (value::0 = '"') do
			value := value+1;
			value::(str.length(value)-1) := 0;
		end
		i := 0;
		while(Set[i]) do
			if (\str.comp(name, Set[i][C_NAME])) leave;
			i := i+1;
		end
		if (\Set[i]) return -1;
		ie (Set[i][C_TYPE] = T_NUMERIC)
			Set[i][C_VAR][0] := xtoi(value);
		else ie (Set[i][C_TYPE] = T_COLOR)
			Set[i][C_VAR][0] := colorvalue(value);
		else ie (Set[i][C_TYPE] = T_NSTRING)
			str.ncopy(Set[i][C_LIMIT], Set[i][C_VAR], value);
		else ie (Set[i][C_TYPE] = T_STRING)
			str.copy(Set[i][C_VAR], value);
		else
			return -1;
	end
	cnf.close();
 end

 public save_subset(file, head, subset) do
	object	cnf[iostream];
	var	cnfbuf::BUFSIZE, linebuf::BUFSIZE;
	var	buf::BUFSIZE;
	var	i, j;

	if (cnf.open(file, cnfbuf, BUFSIZE, iostream.FWRITE) = %1)
		return -2;
	if (head) do
		i := 0;
		while (head[i]) do
			cnf.write(head[i], str.length(head[i]));
			cnf.wrch('\n');
			i := i+1;
		end
	end
	i := 0;
	while (Set[i]) do
		if (subset) do
			j := 0;
			while (subset[j]) do
				if (str.find(Set[i][0], subset[j]) \= -1)
					leave;
				j := j+1;
			end
			if (\subset[j]) do
				i := i+1;
				loop;
			end
		end
		ie (Set[i][C_TYPE] = T_NUMERIC)
			str.format(linebuf, "%S = %D\n",
				[ (Set[i][C_NAME], Set[i][C_VAR][0]) ]);
		else ie (Set[i][C_TYPE] = T_STRING)
			str.format(linebuf, "%S = \q%S\q\n",
				[ (Set[i][C_NAME], Set[i][C_VAR]) ]);
		else ie (Set[i][C_TYPE] = T_NSTRING) do
			str.ncopy(Set[i][C_LIMIT], buf, Set[i][C_VAR]);
			str.format(linebuf, "%S = \q%S\q\n",
				[ (Set[i][C_NAME], buf) ]);
		end
		else
			str.format(linebuf, "%S = %S\n", [
				(Set[i][C_NAME]),
				(colorname(Set[i][C_VAR][0], buf)) ]);
		cnf.write(linebuf, str.length(linebuf));
		i := i+1;
	end
	return cnf.close() < 0-> -1: 0;
 end

 public save(file, head) return self.save_subset(file, head, 0);

end

do end
