! TIDE -- T3X Integrated Development Environment
! Nils M Holm, 2000,2001,2002,2003,2019,2020
! See the file LICENSE for conditions of use.
!
! *********************************************
! Do set at least FILES=20 in CONFIG.SYS on DOS
! *********************************************

#classpath "classes";

module tide(t3x, string, char, ttyctl, iostream, config, xeditor,
	xvedit, menu, prompt, system, filebox, form);

object	t[t3x], str[string], tty[ttyctl], sys[system], v[xvedit], mnu[menu],
	pmpt[prompt], fbox[filebox], frm[form], chr[char];

var	EditBufLen;

const	MAXBUFFERS	= 6;
const	STRLEN		= 128;
const	MAXWIDTH	= 128;
const	MAXPATH		= 256;
const	LINELEN		= 256;
const	MACLEN		= 1024;

struct	Bstate =
	B_name,
	B_title,
	B_modified,
	B_readonly,
	B_truncated,
	B_compiled,
	B_pointpos;

struct BUFLIST = ID_MAIN, ID_HELP, ID_ERRLOG, ID_INPUT, ID_OUTPUT, ID_LICENSE;

object	Cnf[config];
var	Xmax, Ymax, Lines, Columns;
var	Ctext, Csotext, Cmark, Cback;
var	Cmtext, Cmbar, Cmframe, Cstat, Cmsg, Chotkey;
var	Cprompt, Cpdflt;
var	Cftext, Cffield, Cfbar;
var	Cfbtext, Cfbhigh, Cfbback, Cfbframe, Cfbbar;
var	Etabsize, Easyncmode, Eshowctl, Evisblks, Eautoindent;
var	Eautosave, Eautocomp;
var	Expertmode;
var	Spaces::MAXWIDTH+1;
var	Menucset[menu.CSET];
var	Promptcset[prompt.CSET];
var	Fboxcset[filebox.CSET];
var	Formcset[form.CSET];
var	Frameset::menu.FSET+1;
var	Posted, Msgbuf::MAXWIDTH, Msgseen;
var	Macbuf[MACLEN], Macp;
var	Record, Play;
var	Quit;
var	Buffers[MAXBUFFERS];
var	State;
var	Filename::MAXPATH;
var	Searchtext::STRLEN, Revsrch, Repltext::STRLEN;
var	Delbuf::LINELEN;
var	Delln;
var	Keyword::STRLEN;
var	Errormsg::STRLEN, Errorln;
var	Arguments::STRLEN;
var	Homedir::MAXPATH;
var	TX_native,
	TX_basedir::MAXPATH,
	TX_tmp::MAXPATH,
	TX_bindir::MAXPATH,
	TX_classdir::MAXPATH,
	TX_libdir::MAXPATH,
	TX_tmpdir::MAXPATH,
	TX_target::STRLEN,
	TX_dpfilter,
	TX_optimize,
	TX_flags::STRLEN;
var	Tide_path::MAXPATH,
	Tide_dosflag,
	Tide_input::MAXPATH,
	Tide_output::MAXPATH,
	Tide_errlog::MAXPATH,
	Tide_buffer::MAXPATH,
	Tide_txrc::MAXPATH,
	Tide_helpfile::MAXPATH;

vsetup() do
	v.setup([
		[ xvedit.C_TABSIZE,	(Etabsize)	],
		[ xvedit.C_ASYNC,	(Easyncmode)	],
		[ xvedit.C_SHOWCTL,	(Eshowctl)	],
		[ xvedit.C_VBLOCKS,	(Evisblks)	],
		[ xvedit.C_TEXTCOLOR,	(Ctext)		],
		[ xvedit.C_HIGHCOLOR,	(Csotext)	],
		[ xvedit.C_MARKCOLOR,	(Cmark)		],
		[ xvedit.C_XBASE,	0		],
		[ xvedit.C_YBASE,	1		],
		[ xvedit.C_XMAX,	(Xmax)		],
		[ xvedit.C_YMAX,	(Ymax-2)	],
	0 ]);
	t.memfill(Spaces, '\s', Xmax);
	Spaces::Xmax := 0;
	Menucset[menu.C_BACKGROUND] := Cmtext;
	Menucset[menu.C_TEXT] := Cmtext;
	Menucset[menu.C_SOTEXT] := Cmtext;
	Menucset[menu.C_FRAME] := Cmtext;
	Menucset[menu.C_SELECT] := Cmbar;
	Promptcset[prompt.C_BACKGROUND] := Cprompt;
	Promptcset[prompt.C_TEXT] := Cprompt;
	Promptcset[prompt.C_SOTEXT] := Cprompt;
	Promptcset[prompt.C_FRAME] := Cprompt;
	Promptcset[prompt.C_SELECT] := Cpdflt;
	Fboxcset[filebox.C_BACKGROUND] := Cfbback;
	Fboxcset[filebox.C_TEXT] := Cfbtext;
	Fboxcset[filebox.C_SOTEXT] := Cfbhigh;
	Fboxcset[filebox.C_FRAME] := Cfbframe;
	Fboxcset[filebox.C_SELECT] := Cfbbar;
	Formcset[form.C_BACKGROUND] := Cftext;
	Formcset[form.C_TEXT] := Cftext;
	Formcset[form.C_SOTEXT] := Cffield;
	Formcset[form.C_FRAME] := Cftext;
	Formcset[form.C_SELECT] := Cfbar;
end

expand(s) do
	ie (\t.memcomp(s, "~/", 2)) do
		str.nformat(MAXPATH, TX_tmp, "%S/%S", [(Homedir), (@s::2)]);
	end
	else ie (\t.memcomp(s, "$TIDE/", 6)) do
		str.nformat(MAXPATH, TX_tmp, "%S/%S", [(Tide_path), (@s::6)]);
	end
	else ie (\t.memcomp(s, "$T3XDIR/", 8)) do
		str.nformat(MAXPATH, TX_tmp, "%S/%S", [(TX_basedir), (@s::8)]);
	end
	else do
		str.ncopy(MAXPATH, TX_tmp, s);
	end
	return TX_tmp;
end

unexpand(s) do var k;
	k := str.length(Homedir);
	if (\t.memcomp(s, Homedir, k) /\ s::k = '/') do
		str.nformat(MAXPATH, TX_tmp, "~%S", [(@s::k)]);
		str.ncopy(MAXPATH, s, TX_tmp);
		return;
	end
	k := str.length(Tide_path);
	if (\t.memcomp(s, Tide_path, k) /\ s::k = '/') do
		str.nformat(MAXPATH, TX_tmp, "$TIDE%S", [(@s::k)]);
		str.ncopy(MAXPATH, s, TX_tmp);
		return;
	end
	k := str.length(TX_basedir);
	if (\t.memcomp(s, TX_basedir, k) /\ s::k = '/') do
		str.nformat(MAXPATH, TX_tmp, "$T3XDIR%S", [(@s::k)]);
		str.ncopy(MAXPATH, s, TX_tmp);
		return;
	end
end

expandall() do
	str.ncopy(MAXPATH, Tide_path, expand(Tide_path));
	str.ncopy(MAXPATH, Tide_helpfile, expand(Tide_helpfile));
	str.ncopy(MAXPATH, Tide_input, expand(Tide_input));
	str.ncopy(MAXPATH, Tide_output, expand(Tide_output));
	str.ncopy(MAXPATH, Tide_errlog, expand(Tide_errlog));
	str.ncopy(MAXPATH, Tide_buffer, expand(Tide_buffer));
	str.ncopy(MAXPATH, Tide_txrc, expand(Tide_txrc));
end

unexpandall() do
	unexpand(Tide_helpfile);
	unexpand(Tide_input);
	unexpand(Tide_output);
	unexpand(Tide_errlog);
	unexpand(Tide_buffer);
	unexpand(Tide_txrc);
	unexpand(Tide_path);
end

init() do
	const	TXT = config.T_NSTRING,
		NUM = config.T_NUMERIC,
		COL = config.T_COLOR;
	const	textc = ttyctl.F_GREY   | ttyctl.B_BLUE,
		highc = ttyctl.F_YELLOW | ttyctl.B_BLUE     | ttyctl.F_BRIGHT,
		markc = ttyctl.F_BLUE   | ttyctl.B_GREY,
		backc = ttyctl.F_GREY   | ttyctl.B_BLACK,
		menuc = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		mbarc = ttyctl.F_CYAN   | ttyctl.B_RED      | ttyctl.F_BRIGHT,
		mfrmc = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		hotkc = ttyctl.F_CYAN   | ttyctl.B_RED      | ttyctl.F_BRIGHT,
		statc = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		msgc  = ttyctl.F_GREY   | ttyctl.B_RED      | ttyctl.F_BRIGHT,
		pmptc = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		pdflc = ttyctl.F_BLUE   | ttyctl.B_CYAN     | ttyctl.F_BRIGHT,
		fbbkc = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		fbfrm = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		fbtxc = ttyctl.F_BLACK  | ttyctl.B_GREY,
		fbsoc = ttyctl.F_RED    | ttyctl.B_GREY     | ttyctl.F_BRIGHT,
		fbbac = ttyctl.F_CYAN   | ttyctl.B_RED      | ttyctl.F_BRIGHT,
		formc = ttyctl.F_BLACK  | ttyctl.B_GREY,
		frmfc = ttyctl.F_BLUE   | ttyctl.B_GREY     | ttyctl.F_BRIGHT,
		frmbc = ttyctl.F_CYAN   | ttyctl.B_RED      | ttyctl.F_BRIGHT;

	sys.init();
	tty.init();
	if (tty.lines() < 18 \/ tty.columns() < 64) do var b::4;
		t.write(T3X.SYSERR, "TIDE: screen too small", 22);
		t.write(T3X.SYSERR, t.newline(b), str.length(b));
		sys.fini();
		tty.fini();
		halt 1;
	end
	chr.init();
	Cnf.init([
[ "tide-path",		TXT, "$T3XDIR/tide",	@Tide_path,	MAXPATH	],
[ "tide-input",		TXT, "~/.tide/input",	@Tide_input,	MAXPATH	],
[ "tide-output",	TXT, "~/.tide/output",	@Tide_output,	MAXPATH	],
[ "tide-errlog",	TXT, "~/.tide/errors",	@Tide_errlog,	MAXPATH	],
[ "tide-buffer",	TXT, "~/.tide/buffer",	@Tide_buffer,	MAXPATH	],
[ "tide-txrc",		TXT, "~/.tide/txrc",	@Tide_txrc,	MAXPATH	],
[ "tide-help",		TXT, "$TIDE/helpfile",	@Tide_helpfile,	MAXPATH	],
[ "dos-flag",		NUM, 0,			@Tide_dosflag	],
[ "expert-mode",	NUM, 0,			@Expertmode	],
[ "screen-columns",	NUM, 0,			@Columns	],
[ "screen-rows",	NUM, 0,			@Lines		],
[ "text-color",		COL, textc,		@Ctext		],
[ "standout-color",	COL, highc,		@Csotext	],
[ "background-color",	COL, backc,		@Cback		],
[ "hotkey-color",	COL, hotkc,		@Chotkey	],
[ "status-color",	COL, statc,		@Cstat		],
[ "message-color",	COL, msgc,		@Cmsg		],
[ "mark-color",		COL, markc,		@Cmark		],
[ "menu-color",		COL, menuc,		@Cmtext		],
[ "menu-barcolor",	COL, mbarc,		@Cmbar		],
[ "menu-framecolor",	COL, mfrmc,		@Cmframe	],
[ "prompt-color",	COL, pmptc,		@Cprompt	],
[ "prompt-dflcolor",	COL, pdflc,		@Cpdflt		],
[ "fbox-bgcolor",	COL, fbbkc,		@Cfbback	],
[ "fbox-framecolor",	COL, fbfrm,		@Cfbframe	],
[ "fbox-textcolor",	COL, fbtxc,		@Cfbtext	],
[ "fbox-highcolor",	COL, fbsoc,		@Cfbhigh	],
[ "fbox-barcolor",	COL, fbbac,		@Cfbbar		],
[ "form-color",		COL, formc,		@Cftext		],
[ "form-fieldcolor",	COL, frmfc,		@Cffield	],
[ "form-barcolor",	COL, frmbc,		@Cfbar		],
[ "frame-characters",	TXT, "\s\s\s\s\s\s",	@Frameset,	7	],
[ "edit-tabsize",	NUM, 8,			@Etabsize	],
[ "edit-asyncmode",	NUM, 1,			@Easyncmode	],
[ "edit-showctl",	NUM, 1,			@Eshowctl	],
[ "edit-visualblocks",	NUM, 1,			@Evisblks	],
[ "edit-autoindent",	NUM, 1,			@Eautoindent	],
[ "edit-autosave",	NUM, 1,			@Eautosave	],
[ "edit-autocompile",	NUM, 1,			@Eautocomp	],
[ "t3x-native",		NUM, 1,			@TX_native	],
[ "t3x-bindir",		TXT, "$T3XDIR/bin",	@TX_bindir,	MAXPATH	],
[ "t3x-classdir",	TXT, "$T3XDIR/classes",	@TX_classdir,	MAXPATH	],
[ "t3x-libdir",		TXT, "$T3XDIR/lib",	@TX_libdir,	MAXPATH	],
[ "t3x-tmpdir",		TXT, ".",		@TX_tmpdir,	MAXPATH	],
[ "t3x-target",		TXT, "Tcode",		@TX_target,	STRLEN	],
[ "t3x-dpfilter",	NUM, 1,			@TX_dpfilter	],
[ "t3x-optimize",	NUM, 1,			@TX_optimize	],
[ "t3x-flags",		TXT, "",		@TX_flags,	STRLEN	],
	0 ]);
	do var b::10;
		if (t.getenv("COMSPEC", b, 9) \= %1) do
			str.copy(Tide_input, "$TIDE/input");
			str.copy(Tide_output, "$TIDE/output");
			str.copy(Tide_errlog, "$TIDE/errors");
			str.copy(Tide_buffer, "$TIDE/buffer");
			str.copy(Tide_txrc, "$TIDE/txrc");
			Tide_dosflag := %1;
		end
	end
	Xmax := Columns = 0-> tty.columns(): Columns;
	Ymax := Lines = 0-> tty.lines(): Lines;
	EditBufLen := 32760;
	EditBufLen := 2 .* EditBufLen;
	if (v.init(0, 1, Xmax, Ymax-2, EditBufLen) = -1) do var b;
		b := "TIDE: xvedit initialization failed\n";
		t.write(T3X.SYSERR, b, str.length(b));
		halt 1;
	end
	vsetup();
	Posted := 0;
	Msgseen := 1;
	Filename::0 := 0;
	Searchtext::0 := 0;
	Repltext::0 := 0;
	Revsrch := 0;
	Delbuf::0 := 0;
	Delln := 0;
	Errormsg::0 := 0;
	Arguments::0 := 0;
	Record := 0;
	Play := 0;
	Macp := 0;
	t.getenv("T3XDIR", TX_basedir, MAXPATH);
	t.getenv("HOME", Homedir, MAXPATH);
	expandall();
	sys.mkdir(expand("~/.tide"));
	Buffers[ID_MAIN] := [ @Filename, "[Main] ", 0,0,0,0,0 ];
	Buffers[ID_HELP] := [ @Tide_helpfile, "\a[Help]\a ", 0,1,0,0,0 ];
	Buffers[ID_ERRLOG] := [ @Tide_errlog, "\a[Errors]\a ", 0,1,0,0,0 ];
	Buffers[ID_INPUT] := [ @Tide_input, "\a[Input]\a ", 0,0,0,0,0 ];
	Buffers[ID_OUTPUT] := [ @Tide_output, "\a[Output]\a ", 0,1,0,0,-1 ];
	Buffers[ID_LICENSE] := [ "", "\a[License]\a ", 0,1,0,0,0 ];
	State := 0;
end

fini() do
	v.fini();
	tty.color(Cback);
	tty.clear();
	tty.reset();
	tty.fini();
	sys.fini();
	t.remove(Tide_buffer);
	t.remove(Tide_input);
	t.remove(Tide_output);
	t.remove(Tide_errlog);
end

msgwrites(s) do var so, i, j, k;
	so := 0;
	k := str.length(s);
	for (i=0, k) do
		j := i;
		while (s::i /\ s::i \= '\a')
			i := i+1;
		if (\s::i) leave;
		s::i := 0;
		tty.writes(@s::j);
		s::i := '\a';
		ie (so) do
			tty.color(Cmtext);
			so := 0;
		end
		else do
			tty.color(Chotkey);
			so := 1;
		end
	end
	tty.writes(@s::j);
end

decl	message(2);

showname(prefix) do var k, name, buf::MAXWIDTH, max;
	if (\Msgseen) do
		Msgseen := 1;
		message(Posted, 0);
		return;
	end
	tty.move(0, Ymax-1);
	tty.color(Cstat);
	if (State[B_title] /\ \prefix) prefix := State[B_title];
	max := (prefix-> Xmax-str.length(prefix): Xmax) - 2;
	if (prefix) msgwrites(prefix);
	name := State[B_name]::0-> State[B_name]: "(anonymous)";
	k := str.length(name);
	ie (k > max)
		str.nformat(MAXWIDTH, buf, "...%S", [(@name::(k-max+4))]);
	else
		str.ncopy(MAXWIDTH, buf, name);
	tty.writes(buf);
	tty.clreol();
end

message(msg, args) do var x;
	if (\Msgseen /\ msg) do
		tty.move(Xmax-7, Ymax-1);
		tty.color(Cstat);
		tty.writes("[more]");
		tty.readc();
	end
	Posted := msg;
	tty.move(0, Ymax-1);
	tty.color(Cmsg);
	if (\Posted) do
		Msgseen := 1;
		showname(0);
		v.sync();
		return;
	end
	Msgseen := 0;
	tty.clreol();
	tty.writes(@Spaces::1);
	str.nformat(MAXWIDTH, Msgbuf, msg, args);
	Msgbuf::Xmax := 0;
	x := (Xmax - str.length(Msgbuf))/2;
	if (x < 0) x := 0;
	tty.move(x, Ymax-1);
	tty.writes(Msgbuf);
	Posted := Msgbuf;
end

exists(file) do var fd;
	fd := t.open(file, T3X.OREAD);
	if (fd < 0) return 0;
	t.close(fd);
	return 1;
end

getconffile(max, s, load) do
	ie (load /\ exists("tide.rc")) do
		str.ncopy(max, s, "tide.rc");
	end
	else ie (Tide_dosflag) do
		str.nformat(max, s, "%S/tide.rc", [(TX_basedir)]);
	end
	else do
		str.nformat(max, s, "%S/.tide/rc", [(Homedir)]);
		if (\load \/ exists(s)) return;
		str.nformat(max, s, "%S/tide.rc", [(TX_basedir)]);
	end
end

getkey(sync) do var ch;
	if (Play) do
		if (Play < Macp) do
			v.sync();
			Play := Play+1;
			return Macbuf[Play-1];
		end
		Play := 0;
	end
	ch := sync-> v.key(): tty.readc();
	if (Record) do
		ie (ch = Record) do
			Record := 0;
			message("macro recorded -- press [^_] to play", 0);
			return v.key();
		end
		else ie (Macp < MACLEN) do
			Macbuf[Macp] := ch;
			Macp := Macp+1;
		end
		else do
			Record := 0;
			message(
			"macro buffer overflow -- recording stopped", 0);
		end
	end
	return ch;
end

domenu(x, y, d, m) do var i, n, r;
	n := 0;
	while (m[n]) n := n+1;
	ie (Expertmode \/ Play \/ Record) do
		r := getkey(0);
		i := 0;
		for (i=0, n) if (m[i][0] = r) return i;
		return -1;
	end
	else do
		r := mnu.scrselect(x, y, d, Ymax<20-> 10: 0, Menucset,
			Frameset, m, 0);
		ie (Ymax < 20)
			v.refresh(0, 11);
		else
			v.refresh(y-1, y+n);
		return r;
	end
end

configure() do
	var	path::MAXPATH;
	var	r, e;

	if (t.getenv("T3XDIR", TX_basedir, MAXPATH) < 0) do
		TX_basedir::0 := 0;
		message("T3XDIR is undefined", 0);
	end
	getconffile(MAXPATH, path, 1);
	r := Cnf.load(path, @e);
	if (r < -1) return;
	Xmax := Columns = 0-> tty.columns(): Columns;
	Ymax := Lines = 0-> tty.lines(): Lines;
	expandall();
	vsetup();
	if (r < 0) message("error in config file, line %D", [(e)]);
end

frame() do var mbar;
	tty.color(Cmtext);
	tty.move(0, 0);
	tty.writes(Spaces);
	tty.move(0, 0);
	ie (State = Buffers[ID_HELP])
		mbar:= "         \a^K\a Block  \a^Q\a Quick  \a^O\a Other";
	else
		mbar:= "\a^J\a Help  \a^K\a Block  \a^Q\a Quick  \a^O\a Other";
	msgwrites(mbar);
	tty.color(Cstat);
	tty.move(0, Ymax-1);
	tty.writes(@Spaces::1);
end

redraw() do
	tty.color(Cback);
	tty.clear();
	frame();
	v.display();
	showname(0);
end

readstr(ps, buf, len) do var k;
	k := str.length(ps);
	tty.move(0, Ymax-1);
	tty.color(Cprompt);
	tty.writes(@Spaces::1);
	tty.move(0, Ymax-1);
	tty.writes(ps);
	Posted := "";
	return pmpt.input(k, Ymax-1, Xmax-k-1, len, Promptcset, buf, [0]);
end

getsavename(ps, buf) do var k, path::STRLEN, pmt::MAXWIDTH;
	ie (buf::0 = '/') do
		str.nformat(MAXWIDTH, pmt, "%S: ", [(ps)]);
	end
	else do
		k := sys.getdir(path, STRLEN);
		ie (k < 21)
			str.nformat(MAXPATH, pmt, "%S: %S/", [(ps), (path)]);
		else
			str.nformat(MAXPATH, pmt, "%S: ...%S/",
				[(ps), (@path::(k-17))]);
	end
	return readstr(pmt, buf, MAXPATH);
end

rdonly() if (State[B_readonly]) do
		message("buffer is read-only", 0);
		return -1;
	end

confirm(x, y, yes, no) do var r;
	r := domenu(x, y, 0, [ ['y', (yes)], ['n', (no)], 0 ]);
	return r;
end

overwrite(name) do var fd;
	fd := t.open(name, T3X.OREAD);
	if (fd < 0) return 1;
	if (confirm(22, 4, "[Y] overwrite existing file",
		"[N] do not overwrite") = 0)
			return 1;
end

newfile() do
	if (State[B_modified] /\ confirm(24, 2, "[Y] lose changes",
		"[N] keep changes") \= 0)
			return;
	v.nv_goto(0);
	v.mark();
	v.nv_goto(v._length());
	v.setblock();
	v.deleteblock();
	vsetup();
	State[B_name]::0 := 0;
	State[B_modified] := 0;
	State[B_compiled] := 0;
	State[B_readonly] := 0;
	State[B_truncated] := 0;
	v.display();
	showname(0);
end

savefile(newname) do
	if (rdonly()) return;
	if (\State[B_name]::0 \/ newname) do
		if (getsavename("Save file", State[B_name]) = ttyctl.K_ESC)
			return;
		newname := 1;
	end
	if (\State[B_name]::0) return;
	if (newname /\ \overwrite(State[B_name])) do
		State[B_name]::0 := 0;
		return;
	end
	if (v.failed(v.save(State[B_name]))) do
		message("could not save file", 0);
		return;
	end
	State[B_modified] := 0;
end

saveas() do var name::MAXPATH+1;
	if (rdonly()) return;
	name::0 := 0;
	if (getsavename("Save as", name) = ttyctl.K_ESC)
		return;
	if (\name::0 \/ \overwrite(name)) return;
	if (v.failed(v.save(name)))
		message("could not save file", 0);
end

savetmp() do
	if (v.failed(v.save(Tide_buffer))) do
		message("failed to save buffer to temporary storage", 0);
		return -1;
	end
end

loadtmp() do
	if (v.failed(v.load(Tide_buffer))) do
		message("failed to load buffer from temporary storage", 0);
		return -1;
	end
end

loadfile(name, msg) do var buf::256, path::MAXPATH+1, p, i, k, r;
	ie (name) do
		str.ncopy(256, buf, name);
	end
	else do
		if (State[B_modified] /\ confirm(24, 2,
			"[Y] lose changes",
			"[N] keep changes") \= 0)
				return;
		str.copy(buf, " Type file name or press [ENTER] for a list ");
		if (readstr("Load: ", buf, 256) = ttyctl.K_ESC)
			return;
		showname(0);
		if (buf::0 = '\s') do
			sys.getdir(path, MAXPATH+1);
			p := fbox.select(10, 2, 40, 12, Fboxcset, Frameset,
				"[Load File]", path, 0, buf);
			v.refresh(1, 14);
			if (\p) return;
		end
		State[B_compiled] := 0;
		State[B_readonly] := 0;
	end
	if (Tide_dosflag) do
		i := 0; while (buf::i) do
			buf::i := chr.lcase(buf::i);
			i := i+1;
		end
	end
	State[B_modified] := 0;
	State[B_truncated] := 0;
	Errormsg::0 := 0;
	r := v.load(buf);
	if (r = xeditor.E_IO) do
		if (msg) message("could not load file", 0);
		return;
	end
	if (Tide_dosflag) do
		message("stripping CR characters ...", 0);
		Msgseen := 1;
		k := v._length();
		for (i=0, k) do
			v.nv_goto(i);
			if (v._char() = '\r') do
				v.nv_delete(1);
				k := v._length();
			end
		end
		v.nv_goto(0);
	end
	if (r = xeditor.E_SPACE) do
		message("text in buffer is truncated", 0);
		State[B_readonly] := 1;
		State[B_truncated] := 1;
		State[B_name]::0 := 0;
		return;
	end
	str.ncopy(MAXPATH, State[B_name], buf);
	showname(0);
end

doquit() do
	if (State[B_modified] /\ confirm(24, 2, "[Y] quit (lose changes)",
		"[N] don't quit") \= 0)
			return;
	Quit := 1;
end

doclose() do
	if (State[B_modified] /\ confirm(24, 2, "[Y] close (lose changes)",
		"[N] don't close") \= 0)
			return;
	Quit := 1;
end

gotobuf(id) do
	if (State) do
		if (State = Buffers[id]) return;
		ie (State = Buffers[ID_MAIN]) do
			if (savetmp()) return;
		end
		else if (State[B_modified]) do
			savefile(0);
			if (State[B_modified]) do
				message("could not save buffer", 0);
				return -1;
			end
		end
		if (State[B_pointpos] \= -1) State[B_pointpos] := v._point();
	end
	State := Buffers[id];
	ie (State = Buffers[ID_MAIN])
		loadtmp();
	else if (State[B_name]::0)
		loadfile(State[B_name], 0);
	ie (State[B_pointpos] = -1)
		v.bottom();
	else if (State[B_pointpos] <= v._length())
		v.goto(State[B_pointpos]);
	frame();
	showname(0);
end

loadinternal(vtext) do var i, k;
	i := 0;
	while (vtext[i]) i := i+1;
	for (i=i-1, -1, -1) do
		v.sync();
		v.insert('\n');
		v.up();
		v.nv_insert(vtext[i]);
	end
	v.refresh(0, 0);
end

license() do
	if (gotobuf(ID_LICENSE)) do
		redraw();
		message("Please type [^K][S] to name your current buffer", 0);
		return;
	end
	State[B_readonly] := 1;
	v.reset();
	vsetup();
	loadinternal([
"[======= Press CONTROL + K, then Q to close this buffer =======]",
"",
"TIDE -- T3X Integrated Development Environment.",
"Nils M Holm, 2000,2001,2002,2003,2019,2020",
"",
"Redistribution and use in source and binary forms, with or without",
"modification, are permitted provided that the following conditions",
"are met:",
"",
"1. Redistributions of source code must retain the above copyright",
"   notice, this list of conditions and the following disclaimer.",
"2. Redistributions in binary form must reproduce the above copyright",
"   notice, this list of conditions and the following disclaimer in the",
"   documentation and/or other materials provided with the distribution.",
"",
"THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND",
"ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE",
"IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE",
"ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE",
"FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL",
"DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS",
"OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)",
"HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT",
"LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY",
"OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF",
"SUCH DAMAGE.",
	0 ]);
end

banner() do var i, x, y, h, w;
	h := 6;
	w := 50;
	x := (Xmax-w)/2;
	y := (Ymax-h)/2;
	tty.color(Cmtext);
	for (i=y, y+h) do
		tty.move(x, i);
		tty.writes(@Spaces::(Xmax-w));
	end
	tty.move(x+2, y+1); msgwrites("\aT\a3X");
	tty.move(x+2, y+2); msgwrites("\aI\antegrated");
	tty.move(x+2, y+3); msgwrites("\aD\aevelopment");
	tty.move(x+2, y+4); msgwrites("\aE\anvironment");
	tty.move(x+18, y+1); tty.writes("Nils M Holm, 2000-2020");
	tty.move(x+18, y+2); tty.writes("Version 2.0.3");
	tty.move(x+18, y+3); msgwrites("Press \aL\a to load the License");
	tty.move(x+18, y+4); msgwrites("or any other key to start");
	tty.move(0, 0);
	i := getkey(0);
	ie (i = 'L' \/ i = 'l')
		license();
	else
		v.refresh(y-1, y+h);
end

unavailable() if (State \= Buffers[ID_MAIN]) do
		message("function not avaliable in this buffer", 0);
		return -1;
	end

doexit() do
	if (State[B_modified]) savefile(0);
	if (\State[B_modified])
		ie (State = Buffers[ID_MAIN])
			doquit();
		else
			doclose();
end

readblock() do var buf::MAXPATH, path::MAXPATH, p, r;
	if (rdonly()) return;
	str.copy(path, " Type file name or press [ENTER] for a list ");
	if (readstr("Read block: ", path, MAXPATH) = ttyctl.K_ESC)
		return;
	p := path;
	if (path::0 = '\s') do
		sys.getdir(path, MAXPATH);
		p := fbox.select(10, 2, 40, 12, Fboxcset, Frameset,
			"[Read Block]", path, 0, buf);
	end
	v.refresh(1, 14);
	if (\p) return;
	r := v.readblock(p);
	if (v.failed(r)) do
		message("could not read/insert block", 0);
		return;
	end
	State[B_modified] := 1;
end

writeblock() do var name::MAXPATH;
	name::0 := 0;
	if (getsavename("Write block", name) = ttyctl.K_ESC)
		return;
	if (\overwrite(name)) return;
	if (v.failed(v.writeblock(name)))
		message("could not write block", 0);
end

explain(r) do
	if (r = xeditor.E_SPACE) 
		return message("out of space", 0);
	if (r = xeditor.E_SRCDST)
		return message("source/destination conflict", 0);
	State[B_modified] := 1;
end

blockmenu() do var r, i, k;
	r := domenu(9, 1, 0, [
		[ 'n', "New      [^KN]" ],
		[ 'd', "Open     [^KD]" ],
		[ 's', "Save     [^KS]" ],
		[ 'a', "Save as  [^KA]" ],
		[ 't', "Rename   [^KT]" ],
		[ 'b', "Begin    [^KB]" ],
		[ 'k', "End      [^KK]" ],
		[ 'h', "Hide     [^KH]" ],
		[ 'c', "Copy     [^KC]" ],
		[ 'v', "Move     [^KV]" ],
		[ 'y', "Delete   [^KY]" ],
		[ 'r', "Read     [^KR]" ],
		[ 'w', "Write    [^KW]" ],
		[ '0', "About         " ],
		(State = Buffers[ID_MAIN]->
			[ 'q', "Quit     [^KQ]" ]:
			[ 'q', "Close    [^KQ]" ]
		),
		[ 'x', "Exit     [^KX]" ],
	0 ]);
	ie (r = 0) if (\unavailable()) newfile();
	else ie (r = 1) if (\unavailable()) loadfile(0, 1);
	else ie (r = 2) savefile(0);
	else ie (r = 3) saveas();
	else ie (r = 4) savefile(1);
	else ie (r = 5) v.mark();
	else ie (r = 6) v.setblock();
	else ie (r = 7) v.clearblock();
	else ie (r = 8) if (\rdonly()) explain(v.copyblock());
	else ie (r = 9) if (\rdonly()) explain(v.moveblock());
	else ie (r = 10) if (\rdonly())
				State[B_modified] :=
					State[B_modified] | \v.deleteblock();
	else ie (r = 11) if (\rdonly()) readblock();
	else ie (r = 12) writeblock();
	else ie (r = 13) banner();
	else ie (r = 14) ie (State = Buffers[ID_MAIN])
				doquit();
			else
				doclose();
	else if (r = 15) doexit();
end

gotoline() do var buf::20, k;
	buf::0 := 0;
	if (readstr("Goto line: ", buf, 20) = ttyctl.K_ESC)
		return;
	k := str.strtonum(buf, 10, 0);
	if (\k) return;
	if (v.gotoln(k-1) < 0)
		message("no such line", 0);
end

findtext(again, rev) do var r;
	if (\again) do
		r := rev-> "Reverse find: ": "Find: ";
		if (readstr(r, Searchtext, STRLEN) = ttyctl.K_ESC)
			return;
	end
	if (\Searchtext::0) return;
	ie (rev)
		r := v.rfind(Searchtext, 0, v._point()-1);
	else
		r := v.find(Searchtext, v._point()+1, v._length());
	if (v.failed(r)) message("no match", 0);
	Revsrch := rev;
end

findreplace() do
	var	r0, rn, p, pp, k, ls, lr;
	var	auto, dorepl;
	var	nrepl;

	r0 := v._block();
	rn := v._eblock();
	pp := v._point();
	auto := 1;
	if (r0 .>= rn) do
		r0 := 0;
		rn := v._length();
		auto := 0;
	end
	if (readstr(auto-> "Find in block: ": "Find: ", Searchtext,
		STRLEN) = ttyctl.K_ESC \/ \Searchtext::0)
			return;
	if (readstr("Replace with: ", Repltext, STRLEN) = ttyctl.K_ESC)
		return;
	p := v.nv_find(Searchtext, r0, rn);
	if (p = xeditor.E_NOMATCH) do
		message("no match", 0);
		return;
	end
	ls := str.length(Searchtext);
	lr := str.length(Repltext);
	nrepl := 0;
	while (\v.failed(p)) do
		pp := p;
		if (\auto) do
			v.goto(p);
			while (1) do
				message("Replace? (Yes/No/All/Quit)", 0);
				Msgseen := 1;
				k := getkey(1);
				dorepl := 0;
				ie (k = 'y' \/ k = 'Y') dorepl := 1;
				else ie (k = 'q' \/ k = 'Q') ;
				else ie (k = 'n' \/ k = 'N') ;
				else ie (k = 'a' \/ k = 'A') auto := 1;
				else loop;
				leave;
			end
			if (k = 'q' \/ k = 'Q') leave;
		end
		if (auto \/ dorepl) do
			v.nv_goto(p);
			v.nv_delete(ls);
			State[B_modified] := 1;
			if (	Repltext::0 /\
				v.failed(v.nv_insert(Repltext)))
			do
				k := 'q';
				message("Substitution failed", 0);
				leave;
			end
			nrepl := nrepl+1;
			rn := rn+lr-ls;
			if (\auto) v.display();
		end
		p := v.nv_find(Searchtext, p+lr, rn);
	end
	ie (\auto /\ k \= 'q' /\ k \= 'Q')
		message("no more matches", 0);
	else
		message("%D substitutions done", [(nrepl)]);
	if (auto) v.goto(pp);
	v.display();
end

gotoerror() do
	if (\Errormsg::0) do
		message("no errors", 0);
		return;
	end
	if (State[B_modified]) do
		message("error message has expired", 0);
		return 0;
	end
	ie (Errorln < 0) do
		message(Errormsg, 0);
	end
	else do
		v.gotoln(Errorln-1);
		message("T3X: %S", [(Errormsg)]);
	end
end

deltoeol() do var pt, k;
	pt := v._point();
	k := v.copy(pt, Delbuf, LINELEN);
	Delln := 0;
	if (k > 0 /\ Delbuf::(k-1) = '\n') do
		Delbuf::(k-1) := 0;
		if (k > 1 /\ Delbuf::(k-2) = '\r') do
			Delbuf::(k-2) := 0;
		end
	end
	v.mark();
	v.eol();
	v.setblock();
	v.goto(pt);
	v.deleteblock();
	State[B_modified] := 1;
end

recordmac() do
	if (Play \/ Record) do
		message("macro recording terminated", 0);
		Macp := 0;
		Record := 0;
		Play := 0;
		return;
	end
	message("press termination key: ", 0);
	Msgseen := 1;
	Record := tty.readc();
	Macp := 1;
	message("recording, press key [0x%2:0X] to terminate", [(Record)]);
end

playmac() do
	if (rdonly()) return;
	if (Play \/ Record) do
		message("macro replay terminated", 0);
		Macp := 0;
		Play := 0;
		Record := 0;
		return;
	end
	Play := 1;
end

quickmenu() do var r, i, k;
	r := domenu(19, 1, 0, [
		[ 't', "Top of file     [^QT]" ],
		[ 'v', "Bottom of file  [^QV]" ],
		[ 'y', "Delete to EOL   [^QY]" ],
		[ 'e', "Goto error      [^QE]" ],
		[ 'i', "Goto line       [^QI]" ],
		[ 'b', "Goto block      [^QB]" ],
		[ 'k', "Goto EOB        [^QK]" ],
		[ 'f', "Find            [^QF]" ],
		[ 'r', "Reverse find    [^QR]" ],
		[ 'a', "Find/replace    [^QA]" ],
		[ '=', "Record macro    [^Q=]" ],
	0 ]);
	ie (r = 0) v.top();
	else ie (r = 1) v.bottom();
	else ie (r = 2) if (\rdonly()) deltoeol();
	else ie (r = 3) if (\unavailable()) gotoerror();
	else ie (r = 4) gotoline();
	else ie (r = 5) if (v._eblock() > v._block()) v.goto(v._block());
	else ie (r = 6) if (v._eblock() > v._block()) v.goto(v._eblock());
	else ie (r = 7) findtext(0,0);
	else ie (r = 8) findtext(0,1);
	else ie (r = 9) if (\rdonly()) findreplace();
	else if (r = 10) recordmac();
end

geterrmsg() do
	var	buf::STRLEN;
	var	errfd;
	var	k;
	var	i, j;
	var	cantparse;
	var	msg::STRLEN, errln;

	cantparse := 0;
	errfd := t.open(Tide_errlog, T3X.OREAD);
	if (errfd < 0) do
		message("no errors", 0);
		return;
	end
	k := t.read(errfd, buf, STRLEN-1);
	t.close(errfd);
	if (k < 10) do
		if (k > 0) message("bad error file", 0);
		return;
	end
	j := str.scan(buf, '\n');
	if (j \= -1) k := j+1;
	if (buf::(k-1) = '\n') do
		buf::(k-1) := 0;
		if (buf::(k-2) = '\r') buf::(k-2) := 0;
	end
	if (t.memcomp(buf, "TXTRN: ", 7)) cantparse := 1;
	if (\cantparse) do
		i := 7;
		if (buf::i = '*') do
			for (i=i+1, k) if (buf::i = '*') leave;
			if (buf::i \= '*' \/ buf::(i+1) \= ':')
				cantparse := 1;
			i := i+3;
		end
		if (str.parse(@buf::i, "%D: %S", [ (@errln), (msg) ]) \= 2)
			cantparse := 1;
	end
	buf::(Xmax-5) := 0;
	str.ncopy(STRLEN, Errormsg, cantparse-> buf: msg);
	Errorln := cantparse-> -1: errln;
	gotoerror();
end

wrcnf(fd, s) do var k;
	k := str.length(s);
	return t.write(fd, s, k) \= k;
end

mkt3xconf() do
	const	LEN = MAXPATH*2;
	var	buf::LEN;
	var	tcf;
	var	r;

	tcf := t.open(Tide_txrc, T3X.OWRITE);
	if (tcf < 0) do
		message("failed to create T3X config", 0);
		return -1;
	end
	r := 0;
	r := r|wrcnf(tcf, str.nformat(LEN, buf, "bindir = %S\n",
		[(expand(TX_bindir))]));
	r := r|wrcnf(tcf, str.nformat(LEN, buf, "classdir = %S\n",
		[(expand(TX_classdir))]));
	r := r|wrcnf(tcf, str.nformat(LEN, buf, "libdir = %S\n",
		[(expand(TX_libdir))]));
	r := r|wrcnf(tcf, str.nformat(LEN, buf, "tmpdir = %S\n",
		[(expand(TX_tmpdir))]));
	if (str.comp(TX_target, "Tcode"))
		r := r|wrcnf(tcf, str.nformat(LEN, buf, "target = %S\n",
			[(TX_target)]));
	r := r|wrcnf(tcf, str.nformat(LEN, buf, "dpfilter = %D\n",
		[(TX_dpfilter)]));
	r := r|wrcnf(tcf, str.nformat(LEN, buf, "optimize = %D\n",
		[(TX_optimize)]));
	t.close(tcf);
	if (r) message("failed to write T3X config", 0);
	return r-> -1: 0;
end

splitargs(argb, args, p, max) do var i;
	i := 0;
	while (1) do
		while (argb::i = '\s' \/ argb::i = '\t')
			i := i+1;
		if (\argb::i) leave;
		if (p >= max-1) do
			message("too many aruments", 0);
			return -1;
		end
		args[p] := @argb::i; p := p+1;
		while (argb::i /\ argb::i \= '\s' /\ argb::i \= '\t')
			i := i+1;
		if (argb::i) do
			argb::i := 0;
			i := i+1;
		end
	end
	args[p] := 0;
	return p;
end

spawn(prog, args) do var txxpath::MAXPATH, xargs[18];
	if (\TX_native) do
		t.memcopy(@xargs[1], args, 17*t.bpw());
		xargs[0] := prog;
		args := xargs;
		prog := txxpath;
		str.nformat(MAXPATH, txxpath, "%s/txx", [(expand(TX_bindir))]);
	end
	return sys.spawn(prog, args, SYSTEM.SPAWN_WAIT);
end

compile() do
	var	old2;
	var	errfd;
	var	r;
	var	driver::MAXPATH*2;
	var	args[17], p, abuf::STRLEN;
	var	txrc::MAXPATH;

	if (State[B_modified] \/ \State[B_compiled]) do
		ie (Eautosave) do
			savefile(0);
			if (State[B_modified]) return;
		end
		else do
			message("changes must be saved first", 0);
			return;
		end
	end
	str.ncopy(STRLEN, abuf, TX_flags);
	p := splitargs(abuf, args, 1, 15);
	if (p < 0) do
		message("too many compiler flags", 0);
		return;
	end
	args[0] := str.nformat(MAXPATH, txrc, "-U%S", [(Tide_txrc)]);
	args[p] := State[B_name];
	args[p+1] := 0;
	State[B_compiled] := 0;
	if (\State[B_name]::0) do
		message("buffer has no name", 0);
		return;
	end
	Errormsg::0 := 0;
	if (mkt3xconf() < 0) return;
	errfd := t.open(Tide_errlog, T3X.OWRITE);
	if (errfd < 0) do
		message("cannot create error file", 0);
		return;
	end
	old2 := sys.dup(2);
	if (old2 < 0 \/ sys.dup2(errfd, 2) < 0) do
		if (old2 >= 0) t.close(old2);
		t.close(errfd);
		message("INTERNAL: failed to redirect stderr", 0);
		return;
	end
	message("Compiling...", 0);
	Msgseen := 1;
	ie (Tide_dosflag)
		str.nformat(MAXPATH*2, driver, "%S/tx.exe", [(TX_bindir)]);
	else
		str.nformat(MAXPATH*2, driver, "%S/tx", [(TX_bindir)]);
	str.ncopy(MAXPATH*2, driver, expand(driver));
	r := spawn(driver, args);
	if (sys.dup2(old2, 2) < 0) do
		t.close(old2);
		t.close(errfd);
		message("INTERNAL: redirection cleanup failed", 0);
		return;
	end
	t.close(old2);
	t.close(errfd);
	ie (r = 0) do
		State[B_compiled] := 1;
		message(0, 0);
	end
	else ie (r = 1) do
		geterrmsg();
	end
	else if (r > 1) do
		message("failed to run the compiler -- check setup", 0);
	end
end

runprog(redirect) do
	const	NARG = 18;
	var	args[NARG], argb::STRLEN;
	var	txxpath::MAXPATH, progpath::MAXPATH, exepath::MAXPATH;
	var	p, i, k, r;
	var	sysin, sysout, oldin, oldout;
	var	noredir;

	noredir := "I/O rdirection failed";
	k := str.length(State[B_name]);
	if (k < 3 \/ str.comp(@State[B_name]::(k-2), ".t")) do
		message("bad file name (missing .t suffix)", 0);
		return;
	end
	i := str.rscan(State[B_name], '/');
	str.ncopy(MAXPATH, progpath, @State[B_name]::(i+1));
	ie (i < 0)
		progpath::(k-2) := 0;
	else
		progpath::(k-i-3) := 0;
	if (\State[B_compiled] \/ State[B_modified]) do
		ie (Eautocomp) do
			compile();
			if (\State[B_compiled]) return;
		end
		else do
			message("program must be (re)compiled first", 0);
			return;
		end
	end
	ie (\str.comp(TX_target, "Tcode")) do
		ie (Tide_dosflag)
			args[0] := str.nformat(MAXPATH, txxpath, "%S/txx.exe",
				[(expand(TX_bindir))]);
		else
			args[0] := str.nformat(MAXPATH, txxpath, "%S/txx",
				[(expand(TX_bindir))]);
		p := 1;
	end
	else do
		p := 0;
	end
	ie (Tide_dosflag /\ str.comp(TX_target, "Tcode"))
		args[p] := str.nformat(MAXPATH, exepath, "%S.exe",
				[(progpath)]);
	else
		args[p] := progpath;
	p := p+1;
	str.ncopy(STRLEN, argb, Arguments);
	p := splitargs(argb, args, p, NARG);
	if (p < 0) return;
	tty.color(Cback);
	tty.clear();
	if (redirect) do
		sysin := t.open(Tide_input, T3X.OREAD);
		if (sysin < 0) do
			redraw();
			message("redirection failed -- no input", 0);
			return;
		end
		sysout := t.open(Tide_output, T3X.OWRITE);
		if (sysout < 0) do
			t.close(sysin);
			redraw();
			message("output file creation failed", 0);
			return;
		end
		oldin := sys.dup(0);
		if (\oldin) do
			t.close(sysin);
			t.close(sysout);
			redraw();
			message(noredir, 0);
			return;
		end
		oldout := sys.dup(1);
		if (\oldout) do
			t.close(sysin);
			t.close(sysout);
			t.close(oldin);
			redraw();
			message(noredir, 0);
			return;
		end
		if (sys.dup2(sysin, 0) < 0 \/ sys.dup2(sysout, 1) < 0) do
			t.close(sysin);
			t.close(sysout);
			t.close(oldin);
			t.close(oldout);
			redraw();
			message(noredir, 0);
			return;
		end
	end
	if (\redirect) v.ttymode(0);
	r := sys.spawn(args[0], @args[1], SYSTEM.SPAWN_WAIT);
	if (\redirect) v.ttymode(1);
	if (redirect) do
		t.close(sysin);
		t.close(sysout);
		if (sys.dup2(oldin, 0) < 0 \/  sys.dup2(oldout, 1) < 0) do
			tty.writes("Redirection cleanup failed ");
			tty.writes("(should not happen) -- press ENTER");
			getkey(0);
		end
		t.close(oldin);
		t.close(oldout);
	end
	tty.color(Cmtext);
	tty.move(0, tty.lines()-1);
	tty.writes(str.nformat(STRLEN, argb,
		"Run complete -- return code = %D", [(r)]));
	getkey(0);
	redraw();
end

manual() do var rc, txhelppath::MAXPATH*2, b::16;
	ie (Tide_dosflag)
		str.nformat(MAXPATH*2, txhelppath, "%S/txhelp.exe",
			[(expand(TX_bindir))]);
	else
		str.nformat(MAXPATH*2, txhelppath, "%S/txhelp",
			[(expand(TX_bindir))]);
	rc := spawn(txhelppath, [
		"-c", (str.format(b, "%2:0X", [(Ctext)])),
		(Frameset::0 > 128 \/ Tide_dosflag -> "-l": "-"),
		(Xmax<80-> "-p": "-"),
		0 ]);
	redraw();
	if (rc > 1) message("Failed to run TXHELP -- check configuration", 0);
end

editconf() do var rc, confpath::MAXPATH, b::16;
	ie (Tide_dosflag)
		str.nformat(MAXPATH, confpath, "%S/tideconf.exe",
			[(expand(TX_bindir))]);
	else
		str.nformat(MAXPATH, confpath, "%S/tideconf",
			[(expand(TX_bindir))]);
	rc := spawn(confpath, [0]);
	ie (rc > 1) do
		redraw();
		message("Failed to run TIDECONF -- check configuration", 0);
	end
	else do
		configure();
		redraw();
	end
end

compmenu() do var k, i, r;
	ie (State = Buffers[ID_MAIN])
		r := domenu(29, 1, 0, [
			[ 'm', "Main buffer    [^OM]" ],
			[ 'i', "Input buffer   [^OI]" ],
			[ 'o', "Output buffer  [^OO]" ],
			[ 'e', "View errors    [^OE]" ],
			[ 'c', "Compile        [^OC]" ],
			[ 'r', "Run            [^OR]" ],
			[ 's', "Run w/stdio    [^OS]" ],
			[ 'a', "Arguments      [^OA]" ],
			[ 'd', "T3X documents  [^OD]" ],
			[ 'n', "Configure      [^ON]" ],
		0 ]);
	else
		r := domenu(29, 1, 0, [
			[ 'm', "Main buffer    [^OM]" ],
			[ 'i', "Input buffer   [^OI]" ],
			[ 'o', "Output buffer  [^OO]" ],
			[ 'l', "View Errors    [^OE]" ],
		0 ]);
	ie (r = 0) gotobuf(ID_MAIN);
	else ie (r = 1) gotobuf(ID_INPUT);
	else ie (r = 2) gotobuf(ID_OUTPUT);
	else ie (r = 3) gotobuf(ID_ERRLOG);
	else ie (r = 4) compile();
	else ie (r = 5) runprog(0);
	else ie (r = 6) runprog(1);
	else ie (r = 7) readstr("Arguments: ", Arguments, STRLEN);
	else ie (r = 8) manual();
	else if (r = 9) editconf();
end

status() do var buf::50, sv[xvedit.SIZES], k;
	k := v._size(sv);
	str.nformat(50, buf, "Size %UD Line %D(%D) Col %D %S%S%S-- ", [
		(k,
		 sv[xvedit.S_LINE]+1,
		 sv[xvedit.S_TOTAL],
		 v._point()-v._line()+1,
		 State[B_modified]->"modified ":"",
		 State[B_truncated]->"truncated ":"",
		 State[B_readonly]->"readonly ":"")
	]);
	showname(buf);
	Posted := "";
end

clear(file) do var fd;
	t.remove(file);
	fd := t.open(file, T3X.OWRITE);
	if (fd >= 0) t.close(fd);
end

nextw() do
	while (\chr.space(v._char()) /\ \v.right())
		;
	while (chr.space(v._char()) /\ \v.right())
		;
end

prevw() do
	v.left();
	while (chr.space(v._char()) /\ \v.left())
		;
	while (\chr.space(v._char()) /\ \v.left())
		;
	if (v._point() \= 0) v.right();
end

undelete() do
	if (Delln) v.home();
	v.insertblock(Delbuf);
end

edit(id) do var k;
	clear(Tide_buffer);
	gotobuf(id);
	redraw();
	if (Filename::0) loadfile(Filename, 1);
	if (Msgseen /\ \State[B_name]::0) banner();
	Quit := 0;
	while (\Quit) do
		if (State[B_modified]) State[B_compiled] := 0;
		k := getkey(1);
		if (Posted) message(0, 0);
		ie (k = ttyctl.K_UP \/ k = 'P'-'@') v.up();
		else ie (k = ttyctl.K_DOWN \/ k = 'N'-'@') v.down();
		else ie (k = ttyctl.K_LEFT \/ k = 'B'-'@') v.left();
		else ie (k = ttyctl.K_RIGHT \/ k = 'F'-'@') v.right();
		else ie (k = ttyctl.K_HOME \/ k = 'A'-'@') v.home();
		else ie (k = ttyctl.K_END \/ k = 'E'-'@') v.eol();
		else ie (k = ttyctl.K_PREV \/ k = 'T'-'@') v.prev();
		else ie (k = ttyctl.K_NEXT \/ k = 'V'-'@') v.next();
		else ie (k = 'W'-'@') nextw();
		else ie (k = 'S'-'@') prevw();
		else ie (k = 'K'-'@') blockmenu();
		else ie (k = 'Q'-'@') quickmenu();
		else ie (k = 'O'-'@' \/ k = '\\'-'@') compmenu();
		else ie (k = 'R'-'@') redraw();
		else ie (k = 'L'-'@') findtext(1, Revsrch);
		else ie (k = 'G'-'@') status();
		else ie (k = ']'-'@') v.findmatch();
		else ie (k = 'J'-'@') do gotobuf(ID_HELP); loop; end
		else ie (k = ttyctl.K_F1) gotobuf(ID_HELP);
		else ie (k = ttyctl.K_F2) gotobuf(ID_MAIN);
		else ie (k = ttyctl.K_F3) gotobuf(ID_ERRLOG);
		else ie (k = ttyctl.K_F4) gotobuf(ID_INPUT);
		else ie (k = ttyctl.K_F5) gotobuf(ID_OUTPUT);
		else if (State[B_readonly]) do
			rdonly();
			loop;
		end
		ie (k = ttyctl.K_DEL \/ k = 'D'-'@') do
			v.delete(); State[B_modified] := 1;
		end
		else ie (k = '\b' \/ k = TTYCTL.K_BKSP \/ k = 127) do
			v.backspace();
			State[B_modified] := 1;
		end
		else ie (k = 'Y'-'@') do
			v.copyln(Delbuf, LINELEN);
			Delln := 1;
			v.deleteln();
			State[B_modified] := 1;
		end
		else ie (k = 'U'-'@') do
			undelete();
		end
		else ie (k = '_'-'@') do
			playmac();
		end
		else if (k >= '\s' /\ k < 256 \/ k = '\t' \/ k = '\r') do
			ie (v.insert(k='\r'->'\n':k) < 0)
				message("out of space", 0);
			else do
				if (Eautoindent /\ k = '\r') v.indent();
				State[B_modified] := 1;
			end
		end
                if (Quit /\ State \= Buffers[ID_MAIN]) do
			gotobuf(ID_MAIN);
			Quit := 0;
		end
	end
end

do
	if (t.getarg(2, Filename, MAXPATH) >= 0) do var b::3;
		t.write(T3X.SYSERR, "Usage: tide [file]", 19);
		t.write(T3X.SYSERR, t.newline(b), str.length(b));
		halt 1;
	end
	init();
	configure();
	t.getarg(1, Filename, MAXPATH);
	edit(ID_MAIN);
	fini();
end
