! TIDECONF -- T3X Configuration program
! Nils M Holm, 2000-2003,2019-2020
! See the file LICENSE for conditions of use.

module tideconf(t3x, string, ttyctl, menu, prompt, filebox, form, config);

object	t[t3x], str[string], tty[ttyctl];
object	mnu[menu], frm[form], fbox[filebox], pmt[prompt];

const	STRLEN		= 128;
const	MAXWIDTH	= 128;
const	MAXPATH	= 256;

const	MTEXT	=	TTYCTL.F_GREY | TTYCTL.B_BLACK;

object	Cnf[config];
var	Conftbl;
var	Xmax, Ymax, Lines, Columns;
var	Ctext, Csotext, Cmark, Cback;
var	Cmtext, Cmbar, Cmframe, Cstat, Cmsg, Chotkey;
var	Cprompt, Cpdflt;
var	Cftext, Cffield, Cfbar;
var	Cfbtext, Cfbhigh, Cfbback, Cfbframe, Cfbbar;
var	Etabsize, Easyncmode, Eshowctl, Evisblks, Eautoindent;
var	Eautosave, Eautocomp;
var	Expertmode;
var	Spaces::MAXWIDTH+1;
var	Cset[menu.CSET];
var	Fset::menu.FSET+1;
var	Frameset::menu.FSET+1;
var	Homedir::MAXPATH;
var	TX_native,
	TX_basedir::MAXPATH,
	TX_tmp::MAXPATH*2,
	TX_bindir::MAXPATH*2,
	TX_classdir::MAXPATH*2,
	TX_libdir::MAXPATH*2,
	TX_tmpdir::MAXPATH*2,
	TX_target::STRLEN,
	TX_dpfilter,
	TX_optimize,
	TX_flags::STRLEN;
var	Tide_path::MAXPATH+1,
	Tide_dosflag,
	Tide_input::MAXPATH+1,
	Tide_output::MAXPATH+1,
	Tide_errlog::MAXPATH+1,
	Tide_buffer::MAXPATH+1,
	Tide_txrc::MAXPATH+1,
	Tide_helpfile::MAXPATH+1;

msgwrites(s) do var so, i, j, k;
	so := 0;
	k := str.length(s);
	for (i=0, k) do
		j := i;
		while (s::i /\ s::i \= '\a')
			i := i+1;
		if (\s::i) leave;
		s::i := 0;
		tty.writes(@s::j);
		s::i := '\a';
		ie (so) do
			tty.color(Cmtext);
			so := 0;
		end
		else do
			tty.color(Chotkey);
			so := 1;
		end
	end
	tty.writes(@s::j);
end

box(x, y, dx, dy, fs) do var i;
	for (i=y, dy) do
		tty.move(x, i);
		tty.writes(@Spaces::(Xmax-dx+x));
		if (fs) do
			tty.move(x, i);
			tty.writec(Frameset::menu.F_VL);
			tty.move(dx-1, i);
			tty.writec(Frameset::menu.F_VL);
		end
	end
	if (fs) do
		tty.move(x+1, y);
		for (i=x+1, dx) tty.writec(Frameset::menu.F_HL);
		tty.move(x+1, dy-1);
		for (i=x+1, dx) tty.writec(Frameset::menu.F_HL);
	end
	if (fs) do
		tty.move(x, y);
		tty.writec(Frameset::menu.F_UL);
		tty.move(dx-1, y);
		tty.writec(Frameset::menu.F_UR);
		tty.move(x, dy-1);
		tty.writec(Frameset::menu.F_LL);
		tty.move(dx-1, dy-1);
		tty.writec(Frameset::menu.F_LR);
	end
end

refresh() do
	tty.color(Ctext);
	box(0, 2, Xmax, Ymax-1, 0);
	tty.color(MTEXT);
	tty.move(0, 0);
	tty.writes(Spaces);
	tty.move(0, 0);
	tty.writes("TIDE Configuration");
	tty.color(Cmtext);
	tty.move(0, 1);
	tty.writes(Spaces);
	tty.move(0, 1);
	msgwrites(
	"\a^J\a Help  \a^K\a Block  \a^Q\a Quick  \a^O\a Program");
	tty.color(Cstat);
	tty.move(0, Ymax-1);
	tty.writes(Spaces);
	tty.move(1, Ymax-1);
	tty.writes("Status line");
	tty.color(Cmsg);
	tty.move(Xmax/3, Ymax-1);
	tty.writes(@Spaces::(Xmax/3));
	tty.move(Xmax/3+1, Ymax-1);
	tty.writes("Message line");
	tty.color(Cprompt);
	tty.move(2*Xmax/3, Ymax-1);
	tty.writes(@Spaces::(2*Xmax/3));
	tty.move(2*Xmax/3+1, Ymax-1);
	tty.writes("Prompt");
	tty.color(Cpdflt);
	tty.move(2*Xmax/3+8, Ymax-1);
	tty.writes("Default");
	tty.color(Ctext);
	tty.move(0, 3);
	tty.writes("Normal text");
	tty.color(Cmark);
	tty.move(0, 4);
	tty.writes("Marked block text");
	tty.color(Csotext);
	tty.move(0, 5);
	tty.writes("Stand-out text");
	tty.color(Cmframe);
	box(19, 2, 32, 6, 1);
	tty.color(Cmtext);
	tty.move(21, 3);
	tty.writes("Menu item");
	tty.color(Cmbar);
	tty.move(21, 4);
	tty.writes("Menu bar ");
	tty.color(Cftext);
	box(2, 7, 16, 12, 1);
	tty.move(4, 8);
	tty.writes("Form text");
	tty.color(Cffield);
	tty.move(4, 9);
	tty.writes("Form field");
	tty.color(Cfbar);
	tty.move(4, 10);
	tty.writes("Form bar");
	tty.color(Cfbframe);
	box(19, 7, 41, 14, 1);
	tty.move(21, 7);
	tty.writes("Filebox frame");
	tty.color(Cfbback);
	box(20, 8, 40, 13, 1);
	tty.move(21, 12);
	tty.writes("Filebox background");
	tty.color(Cfbtext);
	box(21, 9, 39, 12, 0);
	tty.move(21, 9);
	tty.writes("Filename text");
	tty.color(Cfbhigh);
	tty.move(21, 10);
	tty.writes("Highlighted text");
	tty.color(Cfbbar);
	tty.move(21, 11);
	tty.writes("Selection bar     ");
end

message(s, a) do var b::100;
	tty.move(0, 0);
	tty.writes(Spaces);
	tty.move(0, 0);
	tty.writes(str.nformat(Xmax, b, s, a));
	tty.readc();
end

exists(file) do var fd;
	fd := t.open(file, T3X.OREAD);
	if (fd < 0) return 0;
	t.close(fd);
	return 1;
end

pickcolor(x, y, title, cp) do
	var	i, j, k;

	tty.move(x, y);
	tty.color(Cback);
	tty.writes(@Spaces::(Xmax-48));
	tty.move(x, y);
	tty.writes("Pick color of ");
	for (i=0, 8) do
		tty.move(x, y+i+1);
		for (j=0, 16) do
			tty.color(i<<4|j);
			tty.writes("xyz");
		end
	end
	i := cp[0]>>4;
	j := cp[0]&15;
	while (1) do
		tty.move(x+j*3, y+i+1);
		tty.color(ttyctl.F_GREY | ttyctl.B_BLACK | ttyctl.F_BRIGHT);
		tty.writec('[');
		tty.color(i<<4|j);
		tty.writec('y');
		tty.color(ttyctl.F_GREY | ttyctl.B_BLACK | ttyctl.F_BRIGHT);
		tty.writec(']');
		tty.move(x+15, y);
		tty.color(i<<4|j);
		tty.writes(title);
		k := tty.readc();
		tty.move(x+j*3, y+i+1);
		tty.color(i<<4|j);
		tty.writes("xyz");
		ie (k = ttyctl.K_RIGHT \/ k = 'F'-'@') do
			j := j+1;
			if (j > 15) j := 0;
		end
		else ie (k = ttyctl.K_LEFT \/ k = 'B'-'@') do
			j := j-1;
			if (j < 0) j := 15;
		end
		else ie (k = ttyctl.K_DOWN \/ k = 'N'-'@') do
			i := i+1;
			if (i > 7) i := 0;
		end
		else ie (k = ttyctl.K_UP \/ k = 'P'-'@') do
			i := i-1;
			if (i < 0) i := 7;
		end
		else ie (k = '\r') do
			cp[0] := i<<4 | j;
			leave;
		end
		else if (k = ttyctl.K_ESC) do
			leave;
		end
	end
	refresh();
end

domenu(d, m) do var i, n, r, k, j;
	n := 0;
	k := 0;
	while (m[n]) do
		j := str.length(m[n][1]);
		if (j > k) k := j;
		n := n+1;
	end
	r := mnu.scrselect(Xmax-k-6, 3, d, Ymax<20-> 10: 0, Cset, Fset, m, 0);
	refresh();
	return r;
end

editorcolors() do var r;
	r := 0;
	while (1) do
		r := domenu(r, [
			[ 't', "Text color"		],
			[ 's', "Standout color"		],
			[ 'b', "Marked Block color"	],
			[ 'h', "Hotkey color"		],
			[ 'l', "Status line color"	],
			[ 'm', "Message color"		],
			[ 'g', "Background color"	],
			[ 'x', "Exit"			],
		0 ]);
		ie (r = 0) pickcolor(8, 4, "editor text", @Ctext);
		else ie (r = 1) pickcolor(8, 4, "standout text", @Csotext);
		else ie (r = 2) pickcolor(8, 4, "marked block", @Cmark);
		else ie (r = 3) pickcolor(8, 4, "hotkey", @Chotkey);
		else ie (r = 4) pickcolor(8, 4, "status line", @Cstat);
		else ie (r = 5) pickcolor(8, 4, "message", @Cmsg);
		else ie (r = 6) pickcolor(8, 4, "background", @Cback);
		else leave;
	end
end

menucolors() do var r;
	r := 0;
	while (1) do
		r := domenu(r, [
			[ 't', "Menu text color"	],
			[ 'f', "Menu frame color"	],
			[ 'b', "Menu bar color"		],
			[ 'x', "Exit"			],
		0 ]);
		ie (r = 0) pickcolor(8, 4, "menu text", @Cmtext);
		else ie (r = 1) pickcolor(8, 4, "menu frame", @Cmframe);
		else ie (r = 2) pickcolor(8, 4, "menu bar", @Cmbar);
		else leave;
	end
end

promptcolors() do var r;
	r := 0;
	while (1) do
		r := domenu(r, [
			[ 't', "Prompt text color"		],
			[ 's', "Prompt smart default color"	],
			[ 'x', "Exit"				],
		0 ]);
		ie (r = 0) pickcolor(8, 4, "prompt text", @Cprompt);
		else ie (r = 1) pickcolor(8, 4, "smart default", @Cpdflt);
		else leave;
	end
end

fileboxcolors() do var r;
	r := 0;
	while (1) do
		r := domenu(r, [
			[ 'f', "Filebox frame color"		],
			[ 'b', "Filebox background color"	],
			[ 't', "Filebox text color"		],
			[ 's', "Filebox standout text color"	],
			[ 'e', "Filebox selection bar color"	],
			[ 'x', "Exit"				],
		0 ]);
		ie (r = 0) pickcolor(8, 4, "fbox frame", @Cfbframe);
		else ie (r = 1) pickcolor(8, 4, "fbox background", @Cfbback);
		else ie (r = 2) pickcolor(8, 4, "fbox text", @Cfbtext);
		else ie (r = 3) pickcolor(8, 4, "fbox standout text",
			@Cfbhigh);
		else ie (r = 4) pickcolor(8, 4, "fbox selection bar",
			@Cfbbar);
		else leave;
	end
end

formcolors() do var r;
	r := 0;
	while (1) do
		r := domenu(r, [
			[ 'b', "Form text color"		],
			[ 'f', "Form field color"		],
			[ 's', "Form selection bar color"	],
			[ 'x', "Exit"				],
		0 ]);
		ie (r = 0) pickcolor(8, 4, "form text", @Cftext);
		else ie (r = 1) pickcolor(8, 4, "form fields", @Cffield);
		else ie (r = 2) pickcolor(8, 4, "form selection bar", @Cfbar);
		else leave;
	end
end

framechars() do var r;
	r := Frameset::0 = '\s'-> 0: Frameset::0 = '+'-> 1: 2;
	r := domenu(r, [
		[ 'b', "Blank frame set"	],
		[ 'a', "ASCII frame set"	],
		[ 'i', "IBM frame set"		],
		[ 'x', "Exit"			],
	0 ]);
	ie (r = 0) str.copy(Frameset, "\s\s\s\s\s\s");
	else ie (r = 1) str.copy(Frameset, "++++-|");
	else if (r = 2) str.copy(Frameset, packed[218,191,192,217,196,179,0]);
	refresh();
end

expand(s) do
	ie (\t.memcomp(s, "~/", 2)) do
		str.nformat(MAXPATH, TX_tmp, "%S/%S", [(Homedir), (@s::2)]);
	end
	else ie (\t.memcomp(s, "$TIDE/", 6)) do
		str.nformat(MAXPATH, TX_tmp, "%S/%S", [(Tide_path), (@s::6)]);
	end
	else ie (\t.memcomp(s, "$T3XDIR/", 8)) do
		str.nformat(MAXPATH, TX_tmp, "%S/%S", [(TX_basedir), (@s::8)]);
	end
	else do
		str.ncopy(MAXPATH, TX_tmp, s);
	end
	return TX_tmp;
end

loadcset() do var p, buf::MAXPATH+1, r, e, sfx;
	sfx := Tide_dosflag-> ".CS": ".cs";
	p := fbox.select(10, 2, 40, 12, Cset, Fset,
		"[Select Color Set]", expand(Tide_path), sfx, buf);
	refresh();
	if (\p) return;
	r := Cnf.load(p, @e);
	if (r < -1) do
		message("failed to open color set file", 0);
		return;
	end
	if (r < 0) do
		message("invalid color setfile (line %D)", [(e)]);
		return;
	end
	refresh();
end

savecset() do var path::MAXPATH, r;
	path::0 := 0;
	while (1) do
		refresh();
		tty.move(0, 0);
		tty.color(MTEXT);
		tty.writes(Spaces);
		tty.move(0, 0);
		tty.writes("Color set name: ");
		if (pmt.input(16, 0, Xmax-16, MAXPATH, Cset, path, [0])
			= TTYCTL.K_ESC
		)
			return;
		ie (exists(path)) do
			r := domenu(0, [[ 0, "Choose different name" ],
				        [ 0, "Overwrite existing color set" ],
				        0]);
			if (r = 1) leave;
			if (r < 0) return;
		end
		else leave;
	end
	if (Cnf.save_subset(path, [ "# TIDE color set", "", 0 ],
			[ "screen", "color", "frame", 0 ]) < 0)
	do
		message("could not save file: %S", [(path)]);
	end
end

colorsetup() do var r;
	r := 0;
	while (1) do
		r := domenu(r, [
			[ 'e', "Editor colors..."	],
			[ 'm', "Menu colors..."		],
			[ 'p', "Prompt colors..."	],
			[ 'f', "Filebox colors..."	],
			[ 'o', "Form colors..."		],
			[ 'r', "Frame characters"	],
			[ 'l', "Load color set"		],
			[ 's', "Save color set"		],
			[ 'x', "Exit"			],
		0 ]);
		ie (r = 0) editorcolors();
		else ie (r = 1) menucolors();
		else ie (r = 2) promptcolors();
		else ie (r = 3) fileboxcolors();
		else ie (r = 4) formcolors();
		else ie (r = 5) framechars();
		else ie (r = 6) loadcset();
		else ie (r = 7) savecset();
		else leave;
	end
	refresh();
end

globalsetup() do
	var	x, y, w, h;
	var	setupform, boolmenu, nativemenu;
	var	zoom;

	w := 54;
	h := 10;
	x := (Xmax-w)/2;
	y := (Ymax-h)/2;
	zoom := Xmax < 80-> 1: 0;
	boolmenu := [
		[ 'd', "Disabled" ],
		[ 'e', "Enabled"  ],
	0 ];
	nativemenu := [
		[ 'b', "Tcode"		],
		[ 'n', "Native code"	],
	0 ];
	setupform := [
	[ form.T_TEXT,   5,  1, 0,	"Global Options"		],
	[ form.T_TEXT,   5,  3, 0,	"TIDE base directory"		],
	[ form.T_INPUT, 28,  3, @Tide_path,[20,MAXPATH]		],
	[ form.T_TEXT,   5,  4, 0,	"T3X Installation"		],
	[ form.T_MENU,  28,  4, @TX_native,(nativemenu)			],
	[ form.T_TEXT,   5,  5, 0,	"Expert mode"			],
	[ form.T_MENU,  28,  5, (@Expertmode),(boolmenu)		],
	[ form.T_TEXT,   5,  6, 0,	"Palmtop zoom mode"		],
	[ form.T_MENU,  28,  6, (@zoom),(boolmenu)			],
	[ form.T_DONE,  24,  8, 0,	"[Done]"			],
	0 ];
	frm.complete(x, y, w, h, Cset, Cset, Fset,
		[ ttyctl.K_DOWN, ttyctl.K_UP, ttyctl.K_ESC ], setupform);
	Xmax := zoom-> 64: Columns = 0-> tty.columns(): Columns;
	Ymax := zoom-> 18: Lines = 0-> tty.lines(): Lines;
	refresh();
end

editorsetup() do
	var	x, y, w, h;
	var	setupform, boolmenu;
	var	tabsz::3, nt;

	w := 40;
	h := 13;
	x := (Xmax-w)/2;
	y := (Ymax-h)/2;
	boolmenu := [
		[ 'd', "Disabled" ],
		[ 'e', "Enabled"  ],
	0 ];
	str.numtostr(tabsz, Etabsize&15, 10);
	setupform := [
	[ form.T_TEXT,   5,  1, 0,		"Editor Options"	],
	[ form.T_TEXT,   5,  3, 0,		"Asynchronous mode"	],
	[ form.T_MENU,  28,  3, @Easyncmode,	(boolmenu)		],
	[ form.T_TEXT,   5,  4, 0,		"Allow control chars"	],
	[ form.T_MENU,  28,  4, @Eshowctl,	(boolmenu)		],
	[ form.T_TEXT,   5,  5, 0,		"Visual blocks"		],
	[ form.T_MENU,  28,  5, @Evisblks,	(boolmenu)		],
	[ form.T_TEXT,   5,  6, 0,		"Autoindent"		],
	[ form.T_MENU,  28,  6, @Eautoindent,	(boolmenu)		],
	[ form.T_TEXT,   5,  7, 0,		"Autosave on Compile"	],
	[ form.T_MENU,  28,  7, @Eautosave,	(boolmenu)		],
	[ form.T_TEXT,   5,  8, 0,		"Autocompile on Run"	],
	[ form.T_MENU,  28,  8, @Eautocomp,	(boolmenu)		],
	[ form.T_TEXT,   5,  9, 0,		"Tab Size (chars)"	],
	[ form.T_INPUT, 28,  9, (tabsz),	[2,2]			],
	[ form.T_DONE,  17, 11, 0,		"[Done]"		],
	0 ];
	frm.complete(x, y, w, h, Cset, Cset, Fset,
		[ ttyctl.K_DOWN, ttyctl.K_UP, ttyctl.K_ESC ], setupform);
	nt := str.strtonum(tabsz, 10, 0);
	if (nt & 15) Etabsize := nt&15;
	refresh();
end

t3xsetup() do
	var	x, y, w, h;
	var	setupform, boolmenu;
	var	targetmenu, targetid;

	w := 54;
	h := 14;
	x := (Xmax-w)/2;
	y := (Ymax-h)/2;
	boolmenu := [
		[ 'd', "Disabled"	],
		[ 'e', "Enabled"	],
	0 ];
	targetmenu := [
		[ 't', "Tcode"		],
		[ 'n', "Native code"	],
	0 ];
	targetid := \str.comp(TX_target, "Tcode")-> 0: 1;
	setupform := [
	[ form.T_TEXT,   5,  1, 0,		"T3X Compiler Options"	],
	[ form.T_TEXT,   5,  3, 0,		"Binary directory"	],
	[ form.T_INPUT, 28,  3, @TX_bindir,	[20,STRLEN]		],
	[ form.T_TEXT,   5,  4, 0,		"Class directory"	],
	[ form.T_INPUT, 28,  4, @TX_classdir,	[20,STRLEN]		],
	[ form.T_TEXT,   5,  5, 0,		"Library directory"	],
	[ form.T_INPUT, 28,  5, @TX_libdir,	[20,STRLEN]		],
	[ form.T_TEXT,   5,  6, 0,		"Temporary directory"	],
	[ form.T_INPUT, 28,  6, @TX_tmpdir,	[20,STRLEN]		],
	[ form.T_TEXT,   5,  7, 0,		"Target"		],
	[ form.T_MENU,  28,  7, (@targetid),	(targetmenu)		],
	[ form.T_TEXT,   5,  8, 0,		"Tcode optimization"	],
	[ form.T_MENU,  28,  8, @TX_optimize,	(boolmenu)		],
	[ form.T_TEXT,   5,  9, 0,		"Dead procedure filter"	],
	[ form.T_MENU,  28,  9, @TX_dpfilter,	(boolmenu)		],
	[ form.T_TEXT,   5, 10, 0,		"Additional flags"	],
	[ form.T_INPUT, 28, 10, @TX_flags,	[20,STRLEN]		],
	[ form.T_DONE,  22, 12, 0,		"[Done]"		],
	0 ];
	frm.complete(x, y, w, h, Cset, Cset, Fset,
		[ ttyctl.K_DOWN, ttyctl.K_UP, ttyctl.K_ESC ], setupform);
	str.ncopy(STRLEN, TX_target, targetid=1-> "l": "Tcode");
	refresh();
end

getconffile(max, s, load) do
	ie (load /\ exists("tide.rc")) do
		str.ncopy(max, s, "tide.rc");
	end
	else ie (Tide_dosflag) do
		str.nformat(max, s, "%S/tide.rc", [(TX_basedir)]);
	end
	else do
		str.nformat(max, s, "%S/.tide/rc", [(Homedir)]);
		if (\load \/ exists(s)) return;
		str.nformat(max, s, "%S/tide.rc", [(TX_basedir)]);
	end
end

saveconf() do
	var	path::MAXPATH;
	var	head, r;

	head := [
	"# This is a configuration file for TIDE,",
	"# the T3X Integrated Development Environment",
	"",
	0
	];
	r := domenu(0, [
		[ 'r', "Replace default configuration"	],
		[ 's', "Save as local configuration"	],
	0 ]);
	ie (r = -1) return;
	else ie (r = 0) getconffile(MAXPATH, path, 0);
	else str.copy(path, "tide.rc");
	if (Cnf.save(path, head) < 0) message("error saving config file", 0);
end

abandon() do
	return domenu(0, [['y', "[Y] Quit (lose changes)"],
	                  ['n', "[N] Do not quit"], 0]) = 0;
end

setup() do var r;
	r := 0;
	while (1) do
		r := domenu(r, [
			[ 'g', "Global setup..."	],
			[ 'e', "Editor setup..."	],
			[ 'c', "Compiler setup..."	],
			[ 'o', "Color setup..."		],
			[ 's', "Save and exit"		],
			[ 'x', "Exit without saving"	],
		0 ]);
		ie (r = 0) globalsetup();
		else ie (r = 1) editorsetup();
		else ie (r = 2) t3xsetup();
		else ie (r = 3) colorsetup();
		else ie (r = 4) do saveconf(); leave; end
		else ie (abandon()) leave;
		else r := 5;
	end
end

fini() do
	tty.color(Cback);
	tty.mode(0);
	tty.reset();
	tty.clear();
	tty.move(0, 0);
	tty.clreol();
	tty.fini();
end

configure() do
	var	path::MAXPATH;
	var	r, e;

	getconffile(MAXPATH, path, 1);
	r := Cnf.load(path, @e);
	if (r < -1) return;
	if (r < 0) do
		tty.clear();
		message("error in config file, line %D", [(e)]);
		return %1;
	end
end

init() do
	const	TXT = config.T_NSTRING,
		NUM = config.T_NUMERIC,
		COL = config.T_COLOR;
	const	textc = ttyctl.F_GREY   | ttyctl.B_BLUE,
		highc = ttyctl.F_YELLOW | ttyctl.B_BLUE     | ttyctl.F_BRIGHT,
		markc = ttyctl.F_BLUE   | ttyctl.B_GREY,
		backc = ttyctl.F_GREY   | ttyctl.B_BLACK,
		menuc = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		mbarc = ttyctl.F_CYAN   | ttyctl.B_RED      | ttyctl.F_BRIGHT,
		mfrmc = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		hotkc = ttyctl.F_CYAN   | ttyctl.B_RED      | ttyctl.F_BRIGHT,
		statc = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		msgc  = ttyctl.F_GREY   | ttyctl.B_RED      | ttyctl.F_BRIGHT,
		pmptc = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		pdflc = ttyctl.F_BLUE   | ttyctl.B_CYAN     | ttyctl.F_BRIGHT,
		fbbkc = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		fbfrm = ttyctl.F_BLACK  | ttyctl.B_CYAN,
		fbtxc = ttyctl.F_BLACK  | ttyctl.B_GREY,
		fbsoc = ttyctl.F_RED    | ttyctl.B_GREY     | ttyctl.F_BRIGHT,
		fbbac = ttyctl.F_CYAN   | ttyctl.B_RED      | ttyctl.F_BRIGHT,
		formc = ttyctl.F_BLACK  | ttyctl.B_GREY,
		frmfc = ttyctl.F_BLUE   | ttyctl.B_GREY     | ttyctl.F_BRIGHT,
		frmbc = ttyctl.F_CYAN   | ttyctl.B_RED      | ttyctl.F_BRIGHT;
	t.getenv("T3XDIR", TX_basedir, MAXPATH);
	t.getenv("HOME", Homedir, MAXPATH);
	Conftbl := [
[ "tide-path",		TXT, "$T3XDIR/tide",	@Tide_path,	MAXPATH	],
[ "tide-input",		TXT, "~/.tide/input",	@Tide_input,	MAXPATH	],
[ "tide-output",	TXT, "~/.tide/output",	@Tide_output,	MAXPATH	],
[ "tide-errlog",	TXT, "~/.tide/errors",	@Tide_errlog,	MAXPATH	],
[ "tide-buffer",	TXT, "~/.tide/buffer",	@Tide_buffer,	MAXPATH	],
[ "tide-txrc",		TXT, "~/.tide/txrc",	@Tide_txrc,	MAXPATH	],
[ "tide-help",		TXT, "$TIDE/helpfile",	@Tide_helpfile,	MAXPATH	],
[ "dos-flag",		NUM, 0,			@Tide_dosflag	],
[ "expert-mode",	NUM, 0,			@Expertmode	],
[ "screen-columns",	NUM, 0,			@Columns	],
[ "screen-rows",	NUM, 0,			@Lines		],
[ "text-color",		COL, textc,		@Ctext		],
[ "standout-color",	COL, highc,		@Csotext	],
[ "background-color",	COL, backc,		@Cback		],
[ "hotkey-color",	COL, hotkc,		@Chotkey	],
[ "status-color",	COL, statc,		@Cstat		],
[ "message-color",	COL, msgc,		@Cmsg		],
[ "mark-color",		COL, markc,		@Cmark		],
[ "menu-color",		COL, menuc,		@Cmtext		],
[ "menu-barcolor",	COL, mbarc,		@Cmbar		],
[ "menu-framecolor",	COL, mfrmc,		@Cmframe	],
[ "prompt-color",	COL, pmptc,		@Cprompt	],
[ "prompt-dflcolor",	COL, pdflc,		@Cpdflt		],
[ "fbox-bgcolor",	COL, fbbkc,		@Cfbback	],
[ "fbox-framecolor",	COL, fbfrm,		@Cfbframe	],
[ "fbox-textcolor",	COL, fbtxc,		@Cfbtext	],
[ "fbox-highcolor",	COL, fbsoc,		@Cfbhigh	],
[ "fbox-barcolor",	COL, fbbac,		@Cfbbar		],
[ "form-color",		COL, formc,		@Cftext		],
[ "form-fieldcolor",	COL, frmfc,		@Cffield	],
[ "form-barcolor",	COL, frmbc,		@Cfbar		],
[ "frame-characters",	TXT, "\s\s\s\s\s\s",	@Frameset,	7	],
[ "edit-tabsize",	NUM, 8,			@Etabsize	],
[ "edit-asyncmode",	NUM, 1,			@Easyncmode	],
[ "edit-showctl",	NUM, 1,			@Eshowctl	],
[ "edit-visualblocks",	NUM, 1,			@Evisblks	],
[ "edit-autoindent",	NUM, 1,			@Eautoindent	],
[ "edit-autosave",	NUM, 1,			@Eautosave	],
[ "edit-autocompile",	NUM, 1,			@Eautocomp	],
[ "t3x-native",		NUM, 1,			@TX_native	],
[ "t3x-bindir",		TXT, "$T3XDIR/bin",	@TX_bindir,	MAXPATH	],
[ "t3x-classdir",	TXT, "$T3XDIR/classes",	@TX_classdir,	MAXPATH	],
[ "t3x-libdir",		TXT, "$T3XDIR/lib",	@TX_libdir,	MAXPATH	],
[ "t3x-tmpdir",		TXT, ".",		@TX_tmpdir,	MAXPATH	],
[ "t3x-target",		TXT, "Tcode",		@TX_target,	STRLEN	],
[ "t3x-dpfilter",	NUM, 1,			@TX_dpfilter	],
[ "t3x-optimize",	NUM, 1,			@TX_optimize	],
[ "t3x-flags",		TXT, "",		@TX_flags,	STRLEN	],
	0 ];
	Cnf.init(Conftbl);
	do var b::10;
		if (t.getenv("COMSPEC", b, 9) \= %1) do
			str.copy(Tide_input, "$TIDE/input");
			str.copy(Tide_output, "$TIDE/output");
			str.copy(Tide_errlog, "$TIDE/errors");
			str.copy(Tide_buffer, "$TIDE/buffer");
			str.copy(Tide_txrc, "$TIDE/txrc");
			Tide_dosflag := %1;
		end
	end
	tty.init();
	tty.mode(1);
	if (configure()) do
		fini();
		halt;
	end
	Xmax := Columns = 0-> tty.columns(): Columns;
	Ymax := Lines = 0-> tty.lines(): Lines;
	if (Xmax > MAXWIDTH) do var b;
		message("screen too wide", 0);
		fini();
		halt 0;
	end
	t.memfill(Spaces, '\s', Xmax);
	Spaces::Xmax := 0;
	Cset[MENU.C_BACKGROUND] := TTYCTL.F_GREY | TTYCTL.B_BLACK;
	Cset[MENU.C_FRAME] := TTYCTL.F_GREY | TTYCTL.B_BLACK;
	Cset[MENU.C_TEXT] := TTYCTL.F_GREY | TTYCTL.B_BLACK;
	Cset[MENU.C_SOTEXT] := TTYCTL.F_GREY | TTYCTL.B_BLACK
		| TTYCTL.F_BRIGHT;
	Cset[MENU.C_SELECT] := TTYCTL.F_BLACK | TTYCTL.B_GREY;
	Fset::MENU.F_UL := '+';
	Fset::MENU.F_UR := '+';
	Fset::MENU.F_LL := '+';
	Fset::MENU.F_LR := '+';
	Fset::MENU.F_HL := '-';
	Fset::MENU.F_VL := '|';
end

do
	init();
	refresh();
	setup();
	fini();
end
