! TIDE -- Extended editor class
! Nils M Holm, 2000,2002,2019,2020
! See the file LICENSE for conditions of use.

module xeditor();

public class xeditor(t3x, string, xmem, util)

 object	t[t3x], str[string], xm[xmem], u[util];

 var	BID, Size;
 var	Gap, Egap, Point;
 var	Line;
 var	Block, Eblock;

 public struct	SIZES = S_LENGTH, S_TOTAL, S_LINE;
 public const	E_SPACE = -1, E_NOMATCH = -2, E_OPEN = -3, E_SIZE = -4,
		E_NOBLK = -5, E_SRCDST = -6, E_IO = -7;

 public failed (x) return x <= E_SPACE /\ x .>= E_IO;

 public reset() do
	Gap := 0;
	Egap := Size;
	Point := 0;
	Line := 0;
	Block := 0;
	Eblock := 0;
 end

 public init(n) do
	if (n .>= E_IO) return %1;
	xm.init();
	BID := xm.alloc(n);
	if (BID < 0) return %1;
	Size := n;
	self.reset();
 end

 public fini() do
	xm.free(BID);
	xm.fini();
 end

 movegap(p) do
	ie (Gap .< p) do
		xm.copy(BID, Gap, Egap, p-Gap);
		Egap := Egap + p - Gap;
	end
	else if (Gap .> p) do
		Egap := Egap + p - Gap;
		xm.copy(BID, Egap, p, Gap-p);
	end
	Gap := p;
 end

 normalize() movegap(Size-Egap+Gap);

 pos(n) return n.<Gap-> n: n+Egap-Gap;

 public _point() return Point;

 public _line() return Line;

 public _offset() return Point-Line;

 public _block() return Block;

 public _eblock() return Eblock;

 public _length() return Size-Egap+Gap;

 public _char() return xm.get(BID, pos(Point));

 public _size(sv) do var i, n, m, lim;
	normalize();
	n := 0;
	m := 1;
	lim := Size-Egap+Gap;
	i := 0; while (i .< lim) do
		if (xm.get(BID, i) = '\n') do
			m := m+1;
			if (i .< Point) n := n+1;
		end
		i := i+1;
	end
	sv[S_LENGTH] := lim;
	sv[S_TOTAL] := m;
	sv[S_LINE] := n;
	return lim;
 end

 public goto(p) do
	Point := p;
	Line := xm.get(BID, pos(point)) = '\n'-> Point-1: Point;
	while (Line \= %1) do
		if (xm.get(BID, pos(Line)) = '\n') leave;
		Line := Line-1;
	end
	Line := Line+1;
 end

 public gotoln(n) do var i, lim;
	normalize();
	lim := self._length();
	i := 0; while (i .< lim) do
		if (\n) leave;
		if (xm.get(BID, i) = '\n') n := n-1;
		i := i+1;
	end
	if (n) return E_SIZE;
	Point := i;
	Line := i;
 end

 public gotocol(n) do var lim;
	lim := self._length();
	Point := Line;
	while (Point .< lim) do
		if (\n) leave;
		if (xm.get(BID, pos(Point)) = '\n') leave;
		n := n-1;
		Point := Point+1;
	end
	if (n) return E_SIZE;
 end

 public advance(n) do var m, lim;
	if (Point + n .> Size-Egap+Gap) return E_SIZE;
	m := 0;
	lim := Point+n;
	while (Point .< lim) do
		if (xm.get(BID, pos(Point)) = '\n') m := m+1;
		Point := Point+1;
	end
	self.goto(Point);
	return m;
 end

 public goback(n) do var m, i, lim;
	if (Point .< n) return E_SIZE;
	Point := Point - n;
	m := 0;
	i := Point;
	while (i .< Point + n) do
		if (xm.get(BID, pos(Point)) = '\n') m := m+1;
		i := i+1;
	end
	self.goto(Point);
	return m;
 end

 public next(n) do var i, lim;
	normalize();
	lim := Size-Egap+Gap;
	i := Point; while (i .< lim) do
		if (\n) leave;
		if (xm.get(BID, i) = '\n') n := n-1;
		i := i+1;
	end
	if (n) return E_SIZE;
	self.goto(i);
 end

 public prev(n) do var i;
	normalize();
	if (\Line) return E_SIZE;
	i := Line-1; while (i \= %1) do
		if (xm.get(BID, i) = '\n') n := n-1;
		if (\n) leave;
		i := i-1;
	end
	if (n) return E_SIZE;
	while (i .> 0 /\ xm.get(BID, i-1) \= '\n') i := i-1;
	self.goto(i);
 end

 public top() do
	Point := 0;
	Line := 0;
 end

 public bottom() do
	Point := Size-Egap+Gap;
	self.goto(Point);
 end

 public home() Point := Line;

 public eol() do var lim;
	lim := Size-Egap+Gap;
	while (Point .< lim) do
		if (xm.get(BID, pos(Point)) = '\n') leave;
		Point := Point+1;
	end
	return Point-Line;
 end

 public insertc(k) do
	Eblock := 0;
	if (Gap .>= Egap) return E_SPACE;
	movegap(Point);
	xm.put(BID, Gap, k);
	Gap := Gap+1;
 end

 public inserts(s) do var k;
	Eblock := 0;
	k := str.length(s);
	if (Gap+k .>= Egap) return E_SPACE;
	movegap(Point);
	xm.write(BID, Gap, s, k);
	Gap := Gap+k;
 end

 public replacec(k) do
	Eblock := 0;
	movegap(Point);
	if (Egap .>= Size) return E_SPACE;
	xm.put(BID, Egap, k);
 end

 public replaces(s) do var k;
	Eblock := 0;
	movegap(Point);
	k := str.length(s);
	if (Egap+k .>= Size) return E_SPACE;
	xm.write(BID, Egap, s, k);
 end

 public delete(n) do var i, nl;
	Eblock := 0;
	movegap(Point);
	if (Egap+n .> Size) return E_SIZE;
	nl := 0;
	i := Egap; while (i .< Egap+n) do
		if (xm.get(BID, i) = '\n') nl := nl+1;
		i := i+1;
	end
	Egap := Egap+n;
	return nl;
 end

 public copy(offset, buf, len) do var i, c, lim;
	lim := Size-Egap+Gap;
	i := 0; while (i .< len-1 /\ offset+i .< lim) do
		c := xm.get(BID, pos(offset+i));
		buf::i := c;
		i := i+1;
		if (c = '\n') leave;
	end
	buf::i := 0;
	return i;
 end

 public copyln(buf, len) return self.copy(Line, buf, len);

 public find(s, start, lim) do var k, i, c, buf::256;
	k := str.length(s);
	if (k .> 256) return E_SIZE;
	normalize();
	lim := (lim-> lim: Gap) - k;
	c := s::0;
	i := start; while (i .< lim) do
		if (xm.get(BID, i) = c) do
			xm.read(BID, i, buf, k);
			if (\t.memcomp(buf, s, k)) return i;
		end
		i := i+1;
	end
	return E_NOMATCH;
 end

 public rfind(s, start, lim) do var k, i, c, buf::256;
	k := str.length(s);
	if (k .> 256) return E_SIZE;
	normalize();
	lim := (lim-> lim: Gap) - k;
	c := s::0;
	start := start-1;
	i := lim-1; while (i \= %1) do
		if (xm.get(BID, i) = c) do
			xm.read(BID, i, buf, k);
			if (\t.memcomp(buf, s, k)) return i;
		end
		i := i-1;
	end
	return E_NOMATCH;
 end

 loadfile(name, plen) do var fd, k, n, buf::512;
	fd := t.open(name, T3X.OREAD);
	if (fd < 0) return E_OPEN;
	n := Gap;
	while (1) do
		k := t.read(fd, buf, 512);
		if (n + k .>= Egap \/ n + k .< n) do
			t.close(fd);
			if (plen) plen[0] := Gap;
			Gap := n;
			return E_SPACE;
		end
		xm.write(BID, n, buf, k);
		if (k > 0) n := n + k;
		if (k < 512) leave;
	end
	t.close(fd);
	k := Gap;
	Gap := n;
	return n-k;
 end

 public load(name) do
	self.reset();
	return loadfile(name, 0);
 end

 public load_dos(name) do
	var b::512, fd, k, len, j1, j2, c;

	self.reset();
	fd := t.open(name, T3X.OREAD);
	if (fd < 0) return E_OPEN;
	len := 0;
	while (len .< Size) do
		k := t.read(fd, b, 512);
		j1 := 0;
		j2 := 0;
		while (j1 < k /\ len+j2 .< Size) do
			c := b::j1;
			j1 := j1+1;
			if (c \= '\r') do
				xm.put(BID, len+j2, c);
				j2 := j2+1;
			end
		end
		len := len + j2;
		if (k < 512) leave;
	end
	Gap := len;
	t.close(fd);
	return Gap;
 end

 savefile(name, from, to, app) do var fd, k, n, buf::512;
	normalize();
	ie (app) do
		fd := t.open(name, T3X.ORDWR);
		t.seek(fd, 0, T3X.SEEK_END);
	end
	else do
		fd := t.open(name, T3X.OWRITE);
	end
	if (fd < 0) return E_OPEN;
	while (from .< to) do
		ie (to .< 512)
			n := to;
		else ie (from .< to-512)
			n := 512;
		else
			n := to-from;
		xm.read(BID, from, buf, n);
		k := t.write(fd, buf, n);
		if (k \= n) do
			t.close(fd);
			return E_IO;
		end
		from := from + n;
	end
	t.close(fd);
	return from = to-> 0: E_IO;
 end

 public save(name) do
	normalize();
	return savefile(name, 0, Gap, 0);
 end

 public save_dos(name) do
	var b::512, fd, k, len, lim, i, j1, j2, c;

	normalize();
	fd := t.open(name, T3X.OWRITE);
	if (fd < 0) return E_OPEN;
	len := Gap;
	i := 0; while (i .< Gap) do
		j1 := 0;
		j2 := 0;
		lim := len<256-> len: 256;
		while (j1 .< lim) do
			c := xm.get(BID, i+j1);
			j1 := j1+1;
			ie (c = '\n') do
				b::j2 := '\r';
				b::(j2+1) := '\n';
				j2 := j2+2;
			end
			else do
				b::j2 := c;
				j2 := j2+1;
			end
		end
		k := t.write(fd, b, j2);
		if (k \= j2) do
			t.close(fd);
			return E_IO;
		end
		len := len - lim;
		i := i+256;
	end
	t.close(fd);
 end

 public mark(here) do
	Block := here;
	Eblock := 0;
 end

 public setblock(here) Eblock := here;

 public clearblock() Eblock := 0;

 public readblock(name) do var fd, k, oldgap;
	movegap(Point);
	k := loadfile(name, @oldgap);
	if (k = E_SPACE) do
		Gap := oldgap;
		return E_SPACE;
	end
	Block := self._point();
	Eblock := Block+k;
	return k;
 end

 public appendblock(name) do var fd, k;
	if (Eblock .<= Block) return E_NOBLK;
	normalize();
	return savefile(name, Block, Eblock, 1);
 end

 public writeblock(name) do var fd, k;
	if (Eblock .<= Block) return E_NOBLK;
	normalize();
	return savefile(name, Block, Eblock, 0);
 end

 public deleteblock() do
	if (Eblock .<= Block) return E_NOBLK;
	self.goto(Block);
	movegap(Block);
	Egap := Egap+Eblock-Block;
	Eblock := 0;
 end

 public copyblock() do var src, z;
	if (Eblock .<= Block) return E_NOBLK;
	if (Point .> Block /\ Point .< Eblock) return E_SRCDST;
	z := Eblock-Block;
	if (z .> Egap - Gap) return E_SPACE;
	movegap(Point);
	src := pos(Block);
	xm.copy(BID, Gap, src, z);
	Gap := Gap+z;
	Block := Gap-z;
	Eblock := Gap;
 end

 public moveblock() do var z, b0, bn, r, n0, nn;
	if (Eblock .<= Block) return E_NOBLK;
	b0 := Block;
	bn := Eblock;
	z := Eblock-Block;
	if (Point >= Block /\ Point < Eblock) return E_SRCDST;
	r := self.copyblock();
	if (r \= 0) return r;
	n0 := Block;
	nn := Eblock;
	ie (Block .< b0) do
		Block := b0+z;
		Eblock := bn+z;
	end
	else do
		Block := b0;
		Eblock := bn;
		n0 := n0-z;
		nn := nn-z;
	end
	self.deleteblock();
	Block := n0;
	Eblock := nn;
 end

 public deleteln() do
	Eblock := 0;
	self.home();
	self.mark(Point);
	if (self.next(1) = E_SIZE) self.eol();
	self.setblock(Point);
	self.home();
	return self.deleteblock();
 end

 public findmatch() do var i, n;
	normalize();
	n := 0;
	ie (xm.get(BID, Point) = '(') do
		i := Point; while (i .< Gap) do
			ie (xm.get(BID, i) = '(') n := n+1;
			else if (xm.get(BID, i) = ')') n := n-1;
			if (\n) leave;
			i := i+1;
		end
	end
	else ie (xm.get(BID, Point) = ')') do
		i := Point; while (i \= 0xffff) do
			ie (xm.get(BID, i) = ')') n := n+1;
			else if (xm.get(BID, i) = '(') n := n-1;
			if (\n) leave;
			i := i-1;
		end
	end
	else do
		return E_NOMATCH;
	end
	return n-> E_NOMATCH: i;
 end

 public marklist() do var i, j, n;
	normalize();
	Block := 0;
	Eblock := 0;
	n := xm.get(BID, Point) = ')'-> 0: 1;
	i := Point; while (i \= 0xffff) do
		ie (xm.get(BID, i) = ')') n := n+1;
		else if (xm.get(BID, i) = '(') n := n-1;
		if (\n) leave;
		i := i-1;
	end
	if (n) return E_NOMATCH;
	n := xm.get(BID, Point) = '('-> 0: 1;
	j := Point; while (j .< Size) do
		ie (xm.get(BID, j) = '(') n := n+1;
		else if (xm.get(BID, j) = ')') n := n-1;
		if (\n) leave;
		i := i+1;
	end
	if (n) return E_NOMATCH;
	Block := i;
	Eblock := j+1;
 end

end

do end
