! TXCM -- T3X Class Manager
! Nils M Holm, 2000,2002,2003,2019
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txcm(t3x, string, iostream);

object	t[t3x], str[string];

struct	MODI =
	M_NONE,
	M_MAKE,		! make class dir
	M_LIST,		! list class dir
	M_SPLIT,	! split class dir (extract all)
	M_EXTR,		! extract class dir
	M_INDEX,	! create index file
	M_INST,		! install class
	M_UNINST;	! uninstall class

const	BUFLEN =	512;
const	STRLEN =	256;
const	MAXMEMBERS =	1024;	! max. # of members in class definition
const	MAXNAMEPOOL =	8192;	! space reserved for symbol names
const	CACHESPACE =	4096;	! symbol table (class dir cache) size

!--- Copied from TXTRN ------------------------------------------------!

struct	SYMREC =	! Symbol table record
	SNAME,
	SVAL,
	SSIZE,
	SFLAGS;	

struct SR2 =		! XSYMREC
	SNL, SNH,
	SVL, SVH,
	SZL, SZH,
	SFL, SFH;

const	CNST=	1,	! Symbol Flags
	GLOBL=	2,
	PROC=	4,
	PROTO=	8,
	SYS=	16,
	IPROC=	32,
	PUBLC=	64,
	EXTRN=	128,
	CLSS=	256,
	OBJCT=	512,
	IVAR=	1024,
	REDIR=	2048,
	LINKD=	4096;

!----------------------------------------------------------------------!

struct MEMBER = MNAME, MSIZE, MVAL, MFLAGS;

var	Members[MAXMEMBERS*MEMBER], Mt;	! members of new class dir
var	Names::MAXNAMEPOOL+1, Nt;	! name pool
var	Symbols[CACHESPACE];		! symbol table
var	K64k;				! 65536 constant
var	Flaglist;			! flag names and bitmasks
var	Line, Arg;			! input line and arg #
object	In[iostream], Out[iostream];
var	Inbuf::BUFLEN, Outbuf::BUFLEN;
var	Classname::STRLEN;		! class to manipulate
var	Outfile::STRLEN;		! where the output goes (if not SYSOUT)
var	Basedir::STRLEN;		! $T3XDIR
var	Quiet, Force;			! verbosity flags
var	Itable;				! installable class file suffices

init() do
	K64k := 16384;		! 32-bit hack
	K64k := K64k << 2;
	Mt := 0;
	Nt := 0;
	Line := 0;
	Classname::0 := 0;
	Outfile::0 := 0;
	Basedir::0 := 0;
	Quiet := 0;
	Force := 0;
	Flaglist := [
		[ 5, "CNST",	CNST	],
		[ 6, "GLOBL",	GLOBL	],
		[ 5, "PROC",	PROC	],
		[ 6, "PROTO",	PROTO	],
		[ 4, "SYS",	SYS	],
		[ 6, "IPROC",	IPROC	],
		[ 6, "PUBLC",	PUBLC	],
		[ 6, "EXTRN",	EXTRN	],
		[ 5, "CLSS",	CLSS	],
		[ 6, "OBJCT",	OBJCT	],
		[ 5, "IVAR",	IVAR	],
		[ 6, "REDIR",	REDIR	],
		[ 6, "LINKD",	LINKD	],
		0
	];
	Itable := [
		! suffix  directory   required
		[ ".cls", "/classes",	1 ],
		[ ".dir", "/classes",	1 ],
		[ ".o",   "/lib",	0 ],
		[ ".x",   "/lib",	0 ],
		[ 0,      0,            0 ]
	];
	In.create(T3X.SYSIN, Inbuf, BUFLEN, iostream.FREAD);
	Out.create(T3X.SYSOUT, Outbuf, BUFLEN, iostream.FWRITE);
end

error(m, n) do var b::10;
	t.write(T3X.SYSERR, "TXCM: ", 6);
	t.write(T3X.SYSERR, m, str.length(m));
	if (n) do
		t.write(T3X.SYSERR, ": ", 2);
		t.write(T3X.SYSERR, n, str.length(n));
	end
	t.write(T3X.SYSERR, t.newline(b), str.length(b));
	halt 1;
end

fini() if (Out.flush()) error("file write error", 0);

writep(s) Out.writes(s);

exists(file) do var fd;
	fd := t.open(file, T3X.OREAD);
	if (fd < 0) return 0;
	t.close(fd);
	return %1;
end

var	ntoa_buf::290;
ntoa(n, w) do var k;
	t.memfill(ntoa_buf, '\s', 290);
	str.numtostr(@ntoa_buf::256, n, -10);
	k := str.length(@ntoa_buf::256);
	return k > w -> @ntoa_buf::256: @ntoa_buf::(256-w+k);
end

mk_error(msg) do var b;
	t.write(T3X.SYSERR, "TXCM(make):", 11);
	b := ntoa(Line, 0);
	t.write(T3X.SYSERR, b, str.length(b));
	t.write(T3X.SYSERR, ":", 1);
	b := ntoa(Arg, 0);
	t.write(T3X.SYSERR, b, str.length(b));
	t.write(T3X.SYSERR, ": ", 2);
	t.write(T3X.SYSERR, msg, str.length(msg));
	t.write(T3X.SYSERR, t.newline(b), str.length(b));
	halt 1;
end

rdch() return In.rdch();

skip() do var c;
	while (1) do
		c := rdch();
		if (c \= '\s' /\ c \= '\t' /\ c \= '\r') return c;
	end
end

! extract next entry from symbol description into buffer B
extract(b, n) do var i, c;
	i := 0;
	c := skip();
	while (	i < n-1 /\
		c \= '\s' /\ c \= '\t' /\ c \= '\r' /\ c \= '\n' /\
		c \= %1
	) do
		b::i := c;
		i := i+1;
		c := rdch();
	end
	b::i := 0;
	Arg := Arg+1;
	if (Arg = 4) do
		while (c \= '\n') do
			if (c \= '\s' /\ c \= '\t' /\ c \= '\r')
				mk_error("missing end of line");
			c := rdch();
		end
	end
	return i;
end

! extract numeric argument
number() do var	b::20;
	extract(b, 20);
	if (Arg = 1 /\ In.eof()) return -1;
	if (('0' > b::0 \/ b::0 > '9') /\ b::0 \= '-')
		mk_error("bad numeric argument");
	return str.strtonum(b, 10, 0);
end

! compute flag value
flagval(s) do var i;
	i := 0;
	while (Flaglist[i]) do
		if (\t.memcomp(Flaglist[i][1], s, Flaglist[i][0]))
			return Flaglist[i][2];
		i := i+1;
	end
	mk_error("unknown flag value");
end

! extract list of flag values: flag1|...|flagN
flags() do
	var	b::128;
	var	f, i, j, k, c;

	k := extract(b, 128)+1;
	if (In.eof()) mk_error("symbol flags expected");
	f := 0;
	i := 0;
	j := 0;
	while (1) do
		if (b::i = '|' \/ \b::i) do
			c := b::i;
			b::i := 0;
			f := f | flagval(@b::j);
			j := i+1;
			b::i := c;
		end
		if (\b::i) leave;
		i := i+1;
	end
	return f;
end

! accept symbol name
name() do
	var	b::128;
	var	k, s;

	k := extract(b, 128)+1;
	if (In.eof()) mk_error("symbol name expected");
	if (k + Nt >= MAXNAMEPOOL) mk_error("too many symbol names");
	s := @Names::Nt;
	Nt := Nt+k;
	t.memcopy(s, b, k);
	return s;
end

! generate class dir from definition file
readclassdef() do var id, c;
	id := 0;
	while (1) do
		Line := Line+1;
		Arg := 0;
		if (Mt + MEMBER >= MAXMEMBERS)
			mk_error("too many members");
		Members[Mt+MSIZE] := number();
		if (In.eof()) leave;
		Members[Mt+MVAL] := number();
		Members[Mt+MNAME] := name();
		Members[Mt+MFLAGS] := flags();
		Mt := Mt+MEMBER;
	end
end

mkclassdir() do
	var	i, j, m, f;
	var	nb::2, r::SR2, buf::128;

	Out.write("CDIR", 4);	! magic number
	m := Mt/MEMBER+1;
	nb::0 := m & 255;
	nb::1 := m >> 8;
	Out.write(nb, 2);
	for (i=0, Mt, MEMBER) do
		r::SNL := -1;
		r::SNH := -1;
		r::SVL := Members[i+MVAL];
		r::SVH := Members[i+MVAL] >> 8;
		r::SZL := Members[i+MSIZE];
		r::SZH := Members[i+MSIZE] >> 8;
		ie (Members[i+MFLAGS] & CLSS)
			f := Members[i+MFLAGS] | PUBLC;
		else
			f := Members[i+MFLAGS] | PUBLC | IVAR;
		r::SFL := f;
		r::SFH := f >> 8;
		Out.write(r, SR2);
	end
	r::SNL := -1;	! make module entry
	r::SNH := -1;
	r::SVL := 0;
	r::SVH := 0;
	r::SZL := 0;
	r::SZH := 0;
	r::SFL := 0;
	r::SFH := 0;
	Out.write(r, SR2);
	Out.write(Names, Nt);
	Out.write("*", 1);
	i := 0;
	while (Names::i) i := i+1;
	Names::i := '*';
	Names::(i+1) := 0;
	Out.write(Names, i+2);
end

makecldir() do
	readclassdef();
	mkclassdir();
end

showflag(f, m, nm) do var lm;
	if (f & m) do
		! is there any bit with a lower significance than F?
		lm := ~(~m - (m-1))-m;
		if (lm & f) Out.writes("|");
		Out.write(nm, str.length(nm));
	end
end

! sign-extend h*256+l on >16-bit machines
extend(h, l) do var v;
	v := h<<8 | l;
	if (v > 32767) v := v-K64k;
	return v;
end

dump(nm, xsym) do var f, y[SYMREC], b::10;
	y[SVAL] := extend(xsym::SVH, xsym::SVL);
	y[SSIZE] := extend(xsym::SZH, xsym::SZL);
	y[SFLAGS] := extend(xsym::SFH, xsym::SFL);
	Out.writes(ntoa(y[SSIZE], 6));
	Out.writes(ntoa(y[SVAL], 6));
	Out.write("\s", 1);
	Out.writes(nm);
	Out.write("\t", 1);
	f := y[SFLAGS];
	showflag(f, CNST, "CNST");
	showflag(f, GLOBL, "GLOBL");
	showflag(f, PROC, "PROC");
	showflag(f, PROTO, "PROTO");
	showflag(f, SYS, "SYS");
	showflag(f, IPROC, "IPROC");
	showflag(f, PUBLC, "PUBLC");
	showflag(f, EXTRN, "EXTRN");
	showflag(f, CLSS, "CLSS");
	showflag(f, OBJCT, "OBJCT");
	showflag(f, IVAR, "IVAR");
	showflag(f, REDIR, "REDIR");
	showflag(f, LINKD, "LINKD");
	Out.write(t.newline(b), str.length(b));
end

readcldir() do var hd::4, nb::2, n;
	if (In.read(hd, 4) \= 4 \/ t.memcomp(hd, "CDIR", 4))
		error("magic match failed", 0);
	if (In.read(nb, 2) \= 2) error("short file", 0);
	n := nb::1 << 8 | nb::0;
	if (In.read(Symbols, n*2*SYMREC) \= n*2*SYMREC)
 		error("short file", 0);
	if (In.read(Names, MAXNAMEPOOL+1) = MAXNAMEPOOL+1)
		error("file too long", 0);
	return n;
end

listcldir() do var i, n, p;
	n := readcldir();
	p := 0;
	for (i=0, n*SR2, SR2) do
		dump(@Names::p, @Symbols::i);
		p := p+str.length(@Names::p)+1;
	end
end

writedir(tbl, n) do
	var	k, i, dirname::256, s, b::10;
	object	dir[iostream];
	var	dirbuf::BUFLEN;
	var	fwerr;
	var	nb::2;

	fwerr := "file write error";
	s := @Names::(tbl::SNH << 8 | tbl::SNL);
	str.ncopy(256, dirname, s);
	k := str.length(dirname);
	t.memcopy(@dirname::k, ".dir", 5);
	if (\Quiet) do
		Out.writes(dirname);
		Out.writes(": ");
	end
	if (\Force /\ exists(dirname)) do
		if (\Quiet) do
			Out.writes("exists -- not replaced");
			Out.writes(t.newline(b));
		end
		return;
	end
	if (dir.open(dirname, dirbuf, BUFLEN, iostream.FWRITE) = %1)
		error("file creation failed", dirname);
	if (dir.write("CDIR", 4) \= 4) error(fwerr, 0);
	nb::0 := n;
	nb::1 := n >> 8;
	if (dir.write(nb, 2) \= 2) error(fwerr, 0);
	if (dir.write(tbl, n*SR2) \= n*SR2) error(fwerr, 0);
	for (i=0, n*SR2, SR2) do
		s := @Names::(tbl::(i+SNH) << 8 | tbl::(i+SNL));
		k := str.length(s) + 1;
		if (dir.write(s, k) \= k) error(fwerr, 0);
	end
	if (dir.close()) error(fwerr, 0);
	if (\Quiet) do
		Out.writes("extracted");
		Out.writes(t.newline(b));
	end
end

! Extract class directory of class 'cname' from input file.
! If cname=0, extract all class dirs.
splitdir(n, cname) do
	var	i, j, f, cn;
	var	fnd;

	fnd := 0;
	for (i=0, n*SR2, SR2) do
		f := Symbols::(i+SFH) << 8 | Symbols::(i+SFL);
		if (f & CLSS) do
			j := i;
			cn := @Names::(Symbols::(i+SNH) << 8 |
				Symbols::(i+SNL));
			while (1) do
				if (	\(Symbols::(j+SFH) << 8 |
					  Symbols::(j+SFL))
				)
					leave;
				if (j >= n*SR2)
					error("missing end of class", 0);
				j := j+SR2;
			end
			if (\cname \/ \str.comp(cname, cn)) do
				writedir(@Symbols::i, (j+SR2-i)/SR2);
				fnd := 1;
			end
		end
	end
	if (\fnd /\ \Quiet) do
		ie (cname)
			error("directory not found", cname);
		else
			error("no directories found", 0);
	end
end

splitcldir(cname) do var i, n, p;
	n := readcldir();
	p := 0;
	! fix name pool references
	! (not required any longer, but anyway...)
	for (i=0, n*SR2, SR2) do
		Symbols::(i+SNL) := p;
		Symbols::(i+SNH) := p >> 8;
		p := p+str.length(@Names::p)+1;
	end
	splitdir(n, cname);
end

! make CLASSNDX entry for module 'bname'
! by writing out all module names in CLASSDIR, i.e.
! bname: module1, ..., moduleN
mkindex(bname) do
	var	n, i, f, cn, p, k;
	var	first;

	n := readcldir();
	p := 0;
	first := 1;
	Out.writes(bname);
	Out.wrch(':');
	for (i=0, n*SR2, SR2) do
		f := Symbols::(i+SFH) << 8 | Symbols::(i+SFL);
		k := str.length(@Names::p);
		if (\f) do
			Names::(p+k-1) := 0;
			if (str.comp(@Names::(p+1), bname)) do
				if (\first) Out.wrch(',');
				Out.writes(@Names::(p+1));
				first := 0;
			end
		end
		p := p + k+1;
	end
	Out.wrch('\n');
end

! copy file with error checking
copy(srcname, destname) do
	var	src, dest;
	var	copybuf::BUFLEN;
	var	k, w;

	src := t.open(srcname, T3X.OREAD);
	if (src < 0) error("cannot open file", srcname);
	dest := t.open(destname, T3X.OWRITE);
	if (dest < 0) error("cannot create file", destname);
	while (1) do
		k := t.read(src, copybuf, BUFLEN);
		if (k < 0) error("file read error", srcname);
		if (k = 0) leave;
		w := t.write(dest, copybuf, k);
		if (w \= k) error("file write error", destname);
		if (k < BUFLEN) leave;
	end
	t.close(src);
	t.close(dest);
end

getbasedir() do
	if (Basedir::0) return;
	if (t.getenv("T3XDIR", basedir, STRLEN) < 0)
		error("cannot find T3X base dir. Use -b or set $T3XDIR", 0);
end

! insert an entry into CLASSNDX
insertndx(cls, ndx) do
	var	ndxpath::STRLEN, tmp::STRLEN, ndxent::STRLEN;
	var	srcbuf::BUFLEN, destbuf::BUFLEN;
	object	src[iostream], dest[iostream];
	var	k, l, replaced, fd;

	str.nformat(STRLEN, ndxpath, "%S/classes/CLASSNDX", [(Basedir)]);
	if (\exists(ndxpath)) error("cannot open class index", 0);
	str.ncopy(STRLEN, tmp, ndxpath);
	k := str.length(ndxpath);
	tmp::(k-1) := '-';
	tmp::k := 0;
	copy(ndxpath, tmp);	! make backup copy (CLASSND-)
	! copy in new index entry
	fd := t.open(ndx, T3X.OREAD);
	if (fd < 0) error("cannot open index file", ndx);
	k := t.read(fd, ndxent, STRLEN);
	if (k >= STRLEN) error("index entry too long", ndx);
	ndxent::k := 0;
	t.close(fd);
	if (src.open(tmp, srcbuf, BUFLEN, iostream.FREAD) = %1)
		error("cannot reopen index copy", 0);
	if (dest.open(ndxpath, destbuf, BUFLEN, iostream.FWRITE) = %1) do
		t.remove(ndxpath);
		t.rename(tmp, ndxpath);
		error("cannot create new index file", 0);
	end
	replaced := 0;
	l := str.length(cls);
	while (1) do
		k := src.reads(tmp, STRLEN-1);
		if (src.eof()) leave;
		if (\t.memcomp(tmp, cls, l) /\ tmp::l = ':') do
			str.ncopy(STRLEN, tmp, ndxent);
			replaced := 1;
		end
		dest.writes(tmp);
	end
	src.close();
	! if no entry was replaced, append the new one
	if (\replaced) dest.writes(ndxent);
	if (dest.close()) error("error updating index", 0);
end

wo(s) do var b::3;
	Out.writes(s);
	Out.writes(t.newline(b));
end

install(clsname) do
	var	file::STRLEN, path::STRLEN;
	var	i;

	getbasedir();
	! install the files
	i := 0;
	while (Itable[i][0]) do
		str.nformat(STRLEN, file, "%S%S", [(clsname, Itable[i][0])]);
		ie (exists(file)) do
			str.nformat(STRLEN, path, "%S%S/%S",
				[(Basedir, Itable[i][1], file)]);
			copy(file, path);
		end
		else do
			if (Itable[i][2])
				error("cannot open required file", file);
		end
		i := i+1;
	end
	! update CLASSNDX
	str.nformat(STRLEN, file, "%S.ndx", [(clsname)]);
	ie (exists(file)) do
		insertndx(clsname, file);
	end
	else do
		if (\Quiet) do
			wo("*** No index file has been found for the class");
			wo(str.nformat(STRLEN, file, "*** '%S'", [(clsname)]));
			wo("*** If this class depends on other classes, it");
			wo("*** will not link correctly.");
		end
	end
end

deletendx(cls) do
	var	ndxpath::STRLEN, tmp::STRLEN;
	var	srcbuf::BUFLEN, destbuf::BUFLEN;
	object	src[iostream], dest[iostream];
	var	k, l;

	str.nformat(STRLEN, ndxpath, "%S/classes/CLASSNDX", [(Basedir)]);
	if (\exists(ndxpath)) error("cannot open class index", 0);
	str.ncopy(STRLEN, tmp, ndxpath);
	k := str.length(ndxpath);
	tmp::(k-1) := '-';
	tmp::k := 0;
	copy(ndxpath, tmp);	! create backup copy (CLASSND-)
	if (src.open(tmp, srcbuf, BUFLEN, iostream.FREAD) = %1)
		error("cannot reopen index copy", 0);
	if (dest.open(ndxpath, destbuf, BUFLEN, iostream.FWRITE) = %1) do
		t.remove(ndxpath);
		t.rename(tmp, ndxpath);
		error("cannot create new index file", 0);
	end
	l := str.length(cls);
	while (1) do
		k := src.reads(tmp, STRLEN-1);
		if (src.eof()) leave;
		ie (\t.memcomp(tmp, cls, l) \/ tmp::l \= ':')
			;
		else
			dest.writes(tmp);
	end
	src.close();
	if (dest.close()) error("error updating index", 0);
end

uninstall(clsname) do
	var	path::STRLEN;
	var	i;

	getbasedir();
	i := 0;
	while (Itable[i][0]) do
		str.nformat(STRLEN, path, "%S%S/%S%S",
			[(Basedir, Itable[i][1], clsname, Itable[i][0])]);
		ie (exists(path) /\ t.remove(path) < 0)
			error("cannot remove file", path);
		else
			if (Itable[i][2]) error("file not found", path);
		i := i+1;
	end
	deletendx(clsname);
end

ws(fd, s) do var b::3;
	t.write(fd, s, str.length(s));
	t.write(fd, t.newline(b), str.length(b));
end

usage(fd, q) do
	ws(fd, "Usage:\ttxcm -m [-o file] [file]");
	ws(fd, "\ttxcm -l [-o file] [file]");
	ws(fd, "\ttxcm -n baseclass [-o file] [file]");
	ws(fd, "\ttxcm {-x class | -s} [-qf] [file]");
	ws(fd, "\ttxcm {-i |-u} class [-q] [-b dir]");
	ws(fd, "\ttxcm [-?h]");
	if (q) halt 1;
end

longusage() do
	ws(1, "");
	usage(1, 0);
	ws(1, "");
	ws(1, "-m      make class directory");
	ws(1, "-l      list class directory");
	ws(1, "-n base create class index entry for 'base'");
	ws(1, "-s      extract all directories (split)");
	ws(1, "-x clss extract directory of class 'clss'");
	ws(1, "-i clss install class");
	ws(1, "-u clss uninstall class");
	ws(1, "-b dir  define T3X base directory for -i and -u");
	ws(1, "-o file store output in 'file'");
	ws(1, "-f      overwrite existing directories");
	ws(1, "-q      quiet mode (don't report errors)");
	ws(1, "");
	halt 0;
end

! extract argument of command line argument
argarg(abuf, pap, pi, plen) do var i;
	pi[0] := pi[0]+1;
	if (\abuf::(pi[0])) do
		pap[0] := pap[0]+1;
		plen[0] := t.getarg(pap[0], abuf, STRLEN);
		if (plen[0] < 0) usage(2, 1);
		pi[0] := 0;
	end
	i := pi[0];
	pi[0] := plen[0]-1;
	return @abuf::i;
end

do
	var	abuf::STRLEN, ap, i, k, c;
	var	mode;

	init();
	mode := M_NONE;
	ap := 1;
	while (1) do
		k := t.getarg(ap, abuf, STRLEN);
		if (k < 0 \/ abuf::0 \= '-') leave;
		for (i=1, k) do
			c := abuf::i;
			ie (c = '?' \/ c = 'h') longusage();
			else ie (c = 'b') do
				str.ncopy(STRLEN, Basedir,
					argarg(abuf, @ap, @i, @k));
			end
			else ie (c = 'f') Force := 1;
			else ie (c = 'i') do
				mode := M_INST;
				str.ncopy(STRLEN, Classname,
					argarg(abuf, @ap, @i, @k));
			end
			else ie (c = 'l') mode := M_LIST;
			else ie (c = 'm') mode := M_MAKE;
			else ie (c = 'n') do
				mode := M_INDEX;
				str.ncopy(STRLEN, Classname,
					argarg(abuf, @ap, @i, @k));
			end
			else ie (c = 'o') do
				str.ncopy(STRLEN, Outfile,
					argarg(abuf, @ap, @i, @k));
			end
			else ie (c = 'q') Quiet := 1;
			else ie (c = 's') mode := M_SPLIT;
			else ie (c = 'u') do
				mode := M_UNINST;
				str.ncopy(STRLEN, Classname,
					argarg(abuf, @ap, @i, @k));
			end
			else ie (c = 'x') do
				mode := M_EXTR;
				str.ncopy(STRLEN, Classname,
					argarg(abuf, @ap, @i, @k));
			end
			else usage(2, 1);
		end
		ap := ap+1;
	end
	if (mode = M_NONE) usage(2, 1);
	if (k > 0) do
		if (mode = M_INST \/ mode = M_UNINST) usage(2, 1);
		if (t.getarg(ap+1, abuf, STRLEN) > 0) usage(2, 1);
		t.getarg(ap, abuf, STRLEN);
		if (In.open(abuf, Inbuf, BUFLEN, iostream.FREAD) = %1)
			error("cannot open", abuf);
	end
	if (Outfile::0) do
		! extract and split mode have no redirectable output
		if (mode = M_EXTR \/ mode = M_SPLIT) usage(2, 1);
		if (Out.open(Outfile, Outbuf, BUFLEN, iostream.FWRITE) = %1)
			error("cannot create", Outfile);
	end
	ie (mode = M_MAKE) makecldir();
	else ie (mode = M_LIST) listcldir();
	else ie (mode = M_INDEX) mkindex(Classname);
	else ie (mode = M_SPLIT) splitcldir(0);
	else ie (mode = M_EXTR) splitcldir(Classname);
	else ie (mode = M_INST) install(Classname);
	else ie (mode = M_UNINST) uninstall(Classname);
	else usage(2, 1);
	fini();
end
