! TXDUMP -- Examine Tcode programs
! Nils M Holm, 2000,2001,2002,2019
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txdump(t3x, tcode, iostream, string);

object	t[t3x], str[string];

const	BUFSIZE = 1024;

const	M_INFO=0,	! info mode (print size, etc)
	M_UX=1,		! UX (disassembler) mode
	M_BANG=2;	! print #! paths

var	Addr;			! input address
var	K64;			! 65536 constant
var	Opcodes, Lastop,	! opcode tables, largest opcode
	Decls, Ndcl;		! declaration tables, # of decl codes
object	In[iostream], Out[iostream];
var	Inbuf::BUFSIZE, Outbuf::BUFSIZE;
var	Nlstr::4;		! newline
var	O_mode,			! mode
	O_hex, O_one;		! hex output flag, single-line output flag
var	Name;			! module name
var	Header::256;		! #! header

tc_table() return [			! Tcode7 instruction set
	[ tcode.IGLUE,	"GLUE",	""	],
	[ tcode.IHINT,	"HINT",	"N"	],
	[ tcode.ICLAB,	"CLAB",	"N"	],
	[ tcode.IDLAB,	"DLAB",	"N"	],
	[ tcode.IDATA,	"DATA",	"N"	],
	[ tcode.ICREF,	"CREF",	"N"	],
	[ tcode.IDREF,	"DREF",	"N"	],
	[ tcode.IVEC,	"VEC",	"N"	],
	[ tcode.ISTR,	"STR",	"NB"	],
	[ tcode.IHDR,	"HDR",	""	],
	[ tcode.IEND,	"END",	""	],
	[ tcode.IMHDR,	"MHDR",	""	],
	[ tcode.IENDM,	"ENDM",	""	],
	[ tcode.IPOP,	"POP",	""	],
	[ tcode.IDUP,	"DUP",	""	],
	[ tcode.ISWAP,	"SWAP",	""	],
	[ tcode.ISTACK,	"STACK","N"	],
	[ tcode.ICLEAN,	"CLEAN","N"	],
	[ tcode.INEG,	"NEG",	""	],
	[ tcode.ILNOT,	"LNOT",	""	],
	[ tcode.IBNOT,	"BNOT",	""	],
	[ tcode.IMUL,	"MUL",	""	],
	[ tcode.IDIV,	"DIV",	""	],
	[ tcode.IUMUL,	"UMUL",	""	],
	[ tcode.IUDIV,	"UDIV",	""	],
	[ tcode.IMOD,	"MOD",	""	],
	[ tcode.IADD,	"ADD",	""	],
	[ tcode.ISUB,	"SUB",	""	],
	[ tcode.IBAND,	"BAND",	""	],
	[ tcode.IBOR,	"BOR",	""	],
	[ tcode.IBXOR,	"BXOR",	""	],
	[ tcode.IBSHL,	"BSHL",	""	],
	[ tcode.IBSHR,	"BSHR",	""	],
	[ tcode.IEQU,	"EQU",	""	],
	[ tcode.INEQU,	"NEQU",	""	],
	[ tcode.ILESS,	"LESS",	""	],
	[ tcode.IGRTR,	"GRTR",	""	],
	[ tcode.ILTEQ,	"LTEQ",	""	],
	[ tcode.IGTEQ,	"GTEQ",	""	],
	[ tcode.IULESS,	"ULESS",""	],
	[ tcode.IUGRTR,	"UGRTR",""	],
	[ tcode.IULTEQ,	"ULTEQ",""	],
	[ tcode.IUGTEQ,	"UGTEQ",""	],
	[ tcode.ILDG,	"LDG",	"N"	],
	[ tcode.ILDGV,	"LDGV",	"N"	],
	[ tcode.ILDL,	"LDL",	"N"	],
	[ tcode.ILDLV,	"LDLV",	"N"	],
	[ tcode.ILDI,	"LDI",	"N"	],
	[ tcode.ILDIV,	"LDIV",	"N"	],
	[ tcode.ILDLAB,	"LDLAB","N"	],
	[ tcode.INUM,	"NUM",	"N"	],
	[ tcode.ISELF,	"SELF",	""	],
	[ tcode.IDEREF,	"DEREF",""	],
	[ tcode.IDREFB,	"DREFB",""	],
	[ tcode.INORM,	"NORM",""	],
	[ tcode.INORMB,	"NORMB",""	],
	[ tcode.ISAVG,	"SAVG",	"N"	],
	[ tcode.ISAVL,	"SAVL",	"N"	],
	[ tcode.ISAVI,	"SAVI",	"N"	],
	[ tcode.ISTORE,	"STORE",""	],
	[ tcode.ISTORB,	"STORB",""	],
	[ tcode.IBRF,	"BRF",	"N"	],
	[ tcode.IBRT,	"BRT",	"N"	],
	[ tcode.INBRF,	"NBRF",	"N"	],
	[ tcode.INBRT,	"NBRT",	"N"	],
	[ tcode.IJUMP,	"JUMP",	"N"	],
	[ tcode.IUNEXT,	"UNEXT","N"	],
	[ tcode.IDNEXT,	"DNEXT","N"	],
	[ tcode.IHALT,	"HALT",	"N"	],
	[ tcode.ICALL,	"CALL",	"N"	],
	[ tcode.ICALR,	"CALR",	""	],
	[ tcode.ICALX,	"CALX",	"N"	],
	[ tcode.ISYS,	"SYS",	"N"	],
	[ tcode.ILIB,	"LIB",	"NB"	],
	[ tcode.IICALL,	"ICALL","N"	],
	[ tcode.IICALX,	"ICALX","N"	],
	[ tcode.ILINE,	"LINE",	"N"	],
	[ tcode.IINIT,	"INIT",	"NN"	],
	[ tcode.IINCG,	"INCG",	"NN"	],
	[ tcode.IINCI,	"INCI",	"NN"	],
	[ tcode.IINCL,	"INCL",	"NN"	],
	[ tcode.IPUB,	"PUB",	"NNB"	],
	[ tcode.IEXT,	"EXT",	"NNB"	],
	[ tcode.IIPROC,	"IPROC","NNB"	],
	[ tcode.IIREF,	"IREF",	"NNB"	],
	[ tcode.ICMAP,	"CMAP",	"NN"	],
	[ tcode.IGSYM,	"GSYM",	"NNB"	],
	[ tcode.ILSYM,	"LSYM",	"NNB"	],
	[ tcode.IISYM,	"ISYM",	"NNB"	],
0 ];

tc_decls() return [		! declaration opcodes
	tcode.ICLAB,
	tcode.ICMAP,
	tcode.ICREF,
	tcode.IDATA,
	tcode.IDLAB,
	tcode.IDREF,
	tcode.IEXT,
	tcode.IGSYM,
	tcode.IINIT,
	tcode.IIPROC,
	tcode.IISYM,
	tcode.ILIB,
	tcode.ILINE,
	tcode.ILSYM,
	tcode.IPUB,
	tcode.ISTR,
	tcode.IVEC,
0 ];

init() do
	K64 := 16384;	! avoid 16-bit constant folding
	K64 := K64 << 2;
	Addr := 0;
	t.newline(Nlstr);
	Opcodes := tc_table();
	Decls := tc_decls();
end

ws(s) Out.write(s, str.length(s));

wf(f, a) do var b::1024;
	ws(str.nformat(1024, b, f, a));
	Out.flush();
end

wm(s) t.write(2, s, str.length(s));

wl(s) do
	t.write(2, s, str.length(s));
	t.write(2, Nlstr, str.length(Nlstr));
end

fail(m, n) do var b::256;
	wm(str.nformat(256, b, "TXDUMP: %D: ", [(Addr)]));
	wm(str.nformat(256, b, m, [(n)]));
	wm(Nlstr);
	halt 1;
end

rdc() do
	Addr := Addr+1;
	return In.rdch();
end

rdop() do var n;
	n := rdc();
	n := rdc() << 8 | n;
	if (n > 32767) n := n-K64;	! 32-bit hack
	return n;
end

! print visual string (with non-printable characters escaped)
vstr(s, k) do var b::1024, i, j, c;
	i := 0;
	j := 0;
	for (i=0, k) do
		c := s::i;
		ie (c < '\s' \/ c > '~') do
			str.nformat(b-j, @b::j, "\\%2:0X", [(c)]);
			j := j+3;
		end
		else ie (c = '\\') do
			b::j := '\\';
			b::(j+1) := '\\';
			j := j+2;
		end
		else do
			b::j := c;
			j := j+1;
		end
	end
	b::j := 0;
	wf("  \q%S\q", [(b)]);
end

getversion() do var op, a1, version, hdlen;
	op := rdc();
	if (op < 0) return 0;
	Header::0 := 0;
	hdlen := 0;
	if (op = '#') do
		Header::0 := op;
		for (hdlen=1, 256) do
			a1 := rdc();
			Header::hdlen := a1;
			if (a1 = '\n') leave;
		end
		Header::hdlen := 0;
		op := rdc();
	end
	if (op \= tcode.IINIT)
		fail("not a Tcode program (magic match failed)", 0);
	a1 := rdop();
	version := a1;
	if (version \= 7)
		fail("unsupported Tcode version: %D", version);
	! move to beginning of Tcode (skip #! header, if any)
	In.move(hdlen, iostream.SEEK_SET);
	! compute instruction set size and declaration set size
	for (Lastop=0, 256)
		if (\Opcodes[Lastop])
			leave;
	for (Ndcl=0, 256)
		if (\Decls[Ndcl])
			leave;
	return a1;
end

disasm() do
	var	op, op2, a1;
	var	sbuf::128;
	var	type, n, i;

	Addr := 0;
	while (1) do
		op := rdc();
		if (In.eof()) leave;
		op2 := op & 127;
		if (op2 > Lastop)
			fail("invalid instruction (opcode=%D)", op);
		wf(O_hex-> "%4:0UX  %5LS": "%5UD  %5LS",
			[(Addr-1, Opcodes[op2][1])]);
		type := Opcodes[op2][2];
		for (n=0, 3) do
			if (\type::n) leave;
			ie (type::n = 'N') do
				a1 := rdop();
				wf(O_hex-> "  %4:0UX": "  %6D", [(a1)]);
			end
			else ie (type::n = 'B') do
				if (a1 > 255)
					fail("suspicious string size", a1);
				for (i=0, a1) do
					sbuf::i := rdc();
				end
				vstr(sbuf, a1);
			end
			else ie (type::n = 'V') do
				for (i=0, a1) do
					rdc();
					sbuf::i := rdc();
				end
				vstr(sbuf, a1);
			end
			else do
				fail("bad argument type (INTERNAL ERROR)", 0);
			end
		end
		ws(Nlstr);
	end
end

info(v) do
	var	op, op2, a1, a2;
	var	sbuf::128;
	var	type, n, i;
	var	dsize, fsize, fp;
	var	isize, dclsize;

	dsize := 0;	! size of data declared by program
	fsize := 0;	! max stack frame size used by program
	fp := 0;	! used for frame pointer tracking
	dclsize := 0;	! size of declaration part of program
	while (1) do
		op := rdc();
		if (In.eof()) leave;
		op2 := op & 127;
		if (op2 > Lastop \/ op2 < 1)
			fail("invalid instruction (opcode=0x%2:0X)", op);
		type := Opcodes[op2][2];
		isize := 1;	! instruction size
		for (n=0, 3) do
			if (\type::n) leave;
			ie (type::n = 'N') do
				a1 := rdop();
				isize := isize+2;
			end
			else ie (type::n = 'B') do
				for (i=0, a1) sbuf::i := rdc();
				isize := isize+a1;
			end
			else ie (type::n = 'V') do
				for (i=0, a1) do
					rdc();
					sbuf::i := rdc();
				end
				isize := isize + a1*2;
			end
			else do
				fail("bad argument type (INTERNAL ERROR)", 0);
			end
		end
		ie (op = tcode.IDATA \/ op = tcode.ICREF \/ op = tcode.IDREF)
			dsize := dsize + 2;
		else ie (op = tcode.IVEC)
			dsize := dsize + (a1-> a1*2: 2);
		else ie (op = tcode.ISTR)
			dsize := dsize + a1 + (a1 mod 2 -> 1 : 2);
		else ie (op = tcode.ISTACK)
			fp := fp + a1*2;
		else if (op = tcode.IHDR \/ op = tcode.IMHDR)
			fp := 0;
		if (fp > fsize) fsize := fp;
		for (i=0, Ndcl)
			if (Decls[i] = op) do
				dclsize := dclsize + isize;
				leave;
			end
	end
	Addr := Addr-1;
	ie (O_one) do
		 wf(O_hex-> "v=%D i=%4:0X t=%4:0X d=%4:0X s=%4:0X %S%S":
			"v=%D i=%5:0UD t=%5:0UD d=%5:0UD s=%5:0UD %S%S",
			[(v, Addr, Addr-dclsize, dsize, fsize, Name, Nlstr)]);
	end
	else do
		ie (O_hex) do
		 ie (Header::0)
		  wf("Tcode version:    %4D (%S)%S", [(v, Header, Nlstr)]);
		 else
		  wf("Tcode version:    %4D%S", [ (v, Nlstr) ]);
		end
		else do
		 ie (Header::0)
		  wf("Tcode version:    %5D (%S)%S", [(v, Header, Nlstr)]);
		 else
		  wf("Tcode version:     %4D%S", [ (v, Nlstr) ]);
		 wf("load image size:  %5UD%S", [(Addr, Nlstr)]);
		 wf("text length:      %5UD%S", [(Addr-dclsize, Nlstr)]);
		 wf("data length:      %5UD%S", [(dsize, Nlstr)]);
		 wf("min stack size:   %5UD + 256%S", [(fsize, Nlstr)]);
		end
	end
end

tm_path() do var i;
	ie (O_one) do
		if (\Header::0) return;
		i := str.scan(Header, '/');
		if (i<0) i := 0;
		ws(@Header::i);
	end
	else do
		wf("%S:\t%S", [(Name, Header::0-> Header: "(none)")]);
	end
	ws(Nlstr);
end

process() do var v;
	Addr := 0;
	v := getversion();
	ie (v = 0) wf("%S: empty%S", [(Name, Nlstr)]);
	else ie (O_mode = M_UX) disasm();
	else ie (O_mode = M_INFO) info(v);
	else ie (O_mode = M_BANG) tm_path();
	else ;
end

usage() do
	wm(Nlstr);
	wl("Usage: txdump [-imux1] [file ...]");
	wm(Nlstr);
	wl("-i - summarize Tcode sizes (TXINFO mode)");
	wl("-m - print the #!header of the module");
	wl("-u - unassemble (UX mode)");
	wl("-x - use hex numbers (default = decimal)");
	wl("-1 - (with -i) print one-line summaries");
	wl("     (with -m) print only the interpreter path");
	wm(Nlstr);
	halt 1;
end

defaults() do
	O_mode := M_INFO;
	O_hex := 0;
	O_one := 0;
end

do
	var	arg::128;
	var	i, j, c, k;
	var	mode;

	init();
	defaults();
	i := 1; while (1) do
		k := t.getarg(i, arg, 127);
		if (k < 0 \/ arg::0 \= '-') leave;
		for (j=1, k) do
			c := arg::j;
			ie (c = 'i') O_mode := M_INFO;
			else ie (c = 'm') O_mode := M_BANG;
			else ie (c = 'u') O_mode := M_UX;
			else ie (c = 'x') O_hex := 1;
			else ie (c = '1') O_one := 1;
			else usage();
		end
		i := i+1;
	end
	Out.create(1, Outbuf, BUFSIZE, iostream.FWRITE);
	ie (k < 0) do
		Name := "(stdin)";
		In.create(0, Inbuf, BUFSIZE, iostream.FREAD);
		Out.create(1, Outbuf, BUFSIZE, iostream.FWRITE);
		process();
	end
	else while (k >= 0) do
		Name := arg;
		ie (In.open(arg, Inbuf, BUFSIZE, iostream.FREAD) = %1) do
			fail("%S: cannot open", arg);
		end
		else do
			process();
			In.close();
		end
		i := i+1;
		k := t.getarg(i, arg, 127);
	end
	Out.flush();
end
