! TXSDFMT - A Structured Document (Re-)Formatter
! Nils M Holm, 2002,2003,2004,2019
! See the file LICENSE for conditions of use.

#classpath "../classes";

module txsdfmt(t3x, string, char, iostream);

object	t[t3x], str[string], chr[char];

const	BUFLEN		= 1024;
const	HDBUFLEN	= 8192;
const	STYLEBUFLEN	= 1024;
const	MAXHDRS		= 128;
const	MAXWORD		= 256;
const	STRLEN		= 128;

object	In[iostream], Out[iostream];
var	Inbuf::BUFLEN, Outbuf::BUFLEN;
var	Hdbuf::HDBUFLEN, Ht;
var	Headers[MAXHDRS], Hx;
var	Title::MAXWORD;
var	Tmp::MAXWORD;
var	Hn, Sn, Ssn;
var	Intable, Inlist, Inpara, Inquote;
var	Firstrow;
var	Firsttopic;
var	DidNL;
var	ToFile, FileName::STRLEN;
var	Linebuf::MAXWORD+1, Len;
var	LineLength;
var	KeywordFd;
var	DoBanner, DoTitle, DoShortIDs, DoColor, Do1omit,
	DoTOC, DoFrame, DoStandout, DoTblHeads, DoScatter,
	DoHTML, DoIndent, DoRefmt;
var	Cbg::MAXWORD, Ctxt::MAXWORD, Clnk::MAXWORD;
var	Ctd::MAXWORD, Cth::MAXWORD, Ctb::MAXWORD;
var	Sfont::MAXWORD;
var	Styles::STYLEBUFLEN, Stp;

decl	error(1);

htmlize(s, s2) do
	var	i, j;

	i := 0;
	j := 0;
	while (s::i) do
		if (j >= MAXWORD-10) error("word too long (htmlize)");
		ie (s::i = '<') do
			t.memcopy(@s2::j, "&lt;", 4);
			j := j+4;
		end
		else ie (s::i = '>') do
			t.memcopy(@s2::j, "&gt;", 4);
			j := j+4;
		end
		else ie (s::i = '&') do
			t.memcopy(@s2::j, "&amp;", 5);
			j := j+5;
		end
		else do
			s2::j := s::i; j := j+1;
		end
		i := i+1;
	end
	s2::j := 0;
	return s2;
end

html_writes(s) do var s2::MAXWORD*6+1;
	htmlize(s, s2);
	Out.write(s2, str.length(s2));
end

writes(s) do
	if (DoIndent /\ DidNL /\ \DoHTML) Out.write("\t", 1);
	Out.write(s, str.length(s));
	DidNL := 0;
end

dowrite(s) ie (DoHTML)
		html_writes(s);
	else
		writes(s);

newline() do var b::4;
	Out.writes(t.newline(b));
	DidNL := 1;
end

mwrites(s) do
	t.write(2, s, str.length(s));
end

mnewline() do var b::5;
	mwrites(t.newline(b));
end

error(m) do
	t.remove("txsdfmt.tmp");
	mwrites("TXSDFMT: ");
	mwrites(m);
	mnewline();
	halt 1;
end

init() do
	chr.init();
	In.create(0, Inbuf, BUFLEN, iostream.FREAD);
	Out.create(1, Outbuf, BUFLEN, iostream.FWRITE);
	ToFile := 0;
	Hn := 1;
	Sn := 0;
	Ssn := 0;
	Ht := 0;
	Hx := 0;
	Intable := 0;
	DoBanner := 0;
	DoTOC := 1;
	DoTitle := 0;
	DoShortIDs := 0;
	DoColor := 0;
	Do1omit := 0;
	DoFrame := 0;
	DoStandout := 0;
	DoTblHeads := 0;
	DoScatter := 0;
	DoHTML := 1;
	DoIndent := 0;
	DoRefmt := 0;
	DidNL := 1;
	LineLength := 64;
	Len := 0;
	str.copy(Ctd, "#E0E0E0");
	str.copy(Cth, "#C0C0C0");
	str.copy(Ctb, "#000000");
	Sfont::0 := 0;
	Stp := 0;
	KeywordFd := -1;
end

saveheader(s) do var k, new;
	k := str.length(s);
	if (k + Ht + 1 >= HDBUFLEN) error("out of space for index");
	if (Hx >= MAXHDRS) error("too many headers");
	new := @Hdbuf::Ht;
	Ht := Ht+k+1;
	t.memcopy(new, s, k+1);
	Headers[Hx] := new;
	Hx := Hx+1;
end

trim(s, k) do var i;
	if (k>0 /\ s::(k-1) = '\n') do
		k := k-1;
		s::k := 0;
		if (k > 0 /\ s::(k-1) = '\r') do
			k := k-1;
			s::k := 0;
		end
	end
	i := 0;
	while (s::i = '\s' \/ s::i = '\t') do
		i := i+1;
		k := k-1;
	end
	if (i) t.memcopy(s, @s::i, k+1);
	return k;
end

writekeywords() do
	var	kwbuf::STRLEN, line::STRLEN, k;
	object	kw[iostream];

	writes("<META name=\"description\" content=\"");
	dowrite(Title);
	writes("\">");
	newline();
	kw.create(KeywordFd, kwbuf, STRLEN, IOSTREAM.FREAD);
	while (1) do
		k := kw.reads(line, STRLEN);
		if (kw.eof()) leave;
		k := trim(line, k);
		writes("<META name=\"keywords\" content=\"");
		dowrite(line);
		writes("\">");
		newline();
	end
	kw.close();
	KeywordFd := -1;
end

writestyles() if (Stp) do
	writes("<STYLE>");
	newline();
	writes(Styles);
	newline();
	writes("</STYLE>");
	newline();
end

html_printhead(dokw) do var buf::128;
	writes("<HTML>");
	newline();
	writes("<HEAD>");
	newline();
	writes("<TITLE>");
	dowrite(Title);
	writes("</TITLE>");
	newline();
	if (dokw) writekeywords();
	writestyles();
	writes("</HEAD>");
	newline();
	ie (DoColor) do
		writes(str.nformat(128, buf,
			"<BODY bgcolor=\q%S\q text=\q%S\q", [(Cbg, Ctxt)]));
		writes(str.nformat(128, buf,
			" link=\q%S\q vlink=\q%S\q>", [(Clnk, Clnk)]));
	end
	else do
		writes("<BODY>");
	end
	newline();
end

printhead(dokw) if (DoHTML) html_printhead(dokw);

html_printtail() do
	writes("</BODY>");
	newline();
	writes("</HTML>");
	newline();
end

printtail() if (DoHTML)
		html_printtail();

html_banner() do var i, abuf::256;
	newline();
	newline();
	writes("<P><I>");
	writes("This document has been automatically generated ");
	newline();
	writes("from a </I><B>structured document</B><I> using</I><BR>");
	newline();
	writes("<TT>TXSDFMT");
	i := 1;
	while (1) do
		if (t.getarg(i, abuf, 256) < 0) leave;
		writes("\s");
		writes(abuf);
		i := i+1;
	end
	writes("</TT></P>");
end

banner() if (DoHTML)
		html_banner();

html_printtitle() do var buf::256;
	newline();
	writes("<H1>");
	dowrite(Title);
	writes("</H1>");
	if (DoBanner) banner();
end

sd_printtitle() do
	newline();
	dowrite(Title);
	newline();
end

printtitle() ie (DoHTML)
		html_printtitle();
	else
		sd_printtitle();

fini() do
	In.flush();
	Out.flush();
	if (ToFile) do
		if (DoFrame) printtail();
		Out.close();
	end
end

nextfile(name, dokw) do
	if (ToFile) do
		if (DoFrame) printtail();
		Out.close();
	end
	ie (name)
		str.ncopy(STRLEN, FileName, name);
	else
		str.nformat(STRLEN, FileName, "%2:0D%2:0D%2:0D.html",
			[(Hn, Sn, Ssn)]);
	if (Out.open(FileName, Outbuf, BUFLEN,iostream.FWRITE) = -1)
		error("file creation error");
	ToFile := Filename;
	if (DoFrame) printhead(dokw);
end

shortid(n) do var d;
	d := "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	return n >= 36-> error("too many sections or topics"): d::n;
end

html_writehead(s) do
	var	hd::MAXWORD;
	var	buf::MAXWORD*6, tbuf::MAXWORD*6;
	var	i, j;
	var	start;
	var	newsect;
	var	noanchor;

	i := 0; j := 0;
	start := -1;
	while (s::i) do
		if (s::i = '\s') do
			while (s::i = '\s') i := i+1;
			if (start < 0) start := j+1;
			i := i-1;
		end
		hd::j := s::i; j := j+1; i := i+1;
	end
	hd::j := 0;
	if (start < 0) start := 0;
	noanchor := 0;
	ie (start /\ hd::(start-2) = ')') do
		Sn := Sn+1;
		Ssn := 0;
		if (DoScatter) nextfile(0, 0);
		newline();
		Firsttopic := 1;
		writes("<H2>");
		ie (DoScatter) do
			noanchor := 1;
		end
		else ie (DoShortIDs)
			writes(str.nformat(MAXWORD*6, buf, "<A name=\"%C%C\">",
				[(shortid(Hn), shortid(Sn))]));
		else
			writes(str.nformat(MAXWORD*6, buf, "<A name=\"%D.%D\">",
				[(Hn, Sn)]));
		writes(str.nformat(MAXWORD*6, buf, "%D.%D\s%S", [(Hn, Sn,
			htmlize(@hd::start, tbuf))]));
		newsect := 1;
	end
	else do
		if (Do1omit /\ Firsttopic) do
			Firsttopic := 0;
			return;
		end
		Ssn := Ssn+1;
		if (DoScatter > 1) nextfile(0, 0);
		newline();
		writes("<H3>");
		ie (DoScatter) do
			ie (DoScatter > 1)
				noanchor := 1;
			else
				writes(str.nformat(MAXWORD*6, buf,
					"<A name=\"%D\">", [(Ssn)]));
		end
		else ie (DoShortIDs)
			writes(str.nformat(MAXWORD*6, buf,
				"<A name=\"%C%C%C\">",
				[(shortid(Hn), shortid(Sn),
				 shortid(Ssn))]));
		else
			writes(str.nformat(MAXWORD*6, buf,
				"<A name=\"%D.%D.%D\">", [(Hn, Sn, Ssn)]));
		writes(str.nformat(MAXWORD*6, buf,
			"%D.%D.%D\s%S", [(Hn, Sn, Ssn,
			htmlize(@hd::start, tbuf))]));
		newsect := 0;
	end
	writes(noanchor->
		newsect-> "</H2>": "</H3>":
		newsect-> "</A></H2>": "</A></H3>");
	newline();
	saveheader(buf);
end

sd_writehead(s) do
	var	hd::MAXWORD;
	var	buf::MAXWORD*6, tbuf::MAXWORD*6;
	var	i, j;
	var	start;
	var	newsect;

	i := 0; j := 0;
	start := -1;
	while (s::i) do
		if (s::i = '\s') do
			while (s::i = '\s') i := i+1;
			if (start < 0) start := j+1;
			i := i-1;
		end
		hd::j := s::i; j := j+1; i := i+1;
	end
	hd::j := 0;
	if (start < 0) start := 0;
	newline();
	newline();
	ie (start /\ hd::(start-2) = ')') do
		Sn := Sn+1;
		Ssn := 0;
		Firsttopic := 1;
		writes(str.nformat(MAXWORD*6, buf, "%D)\s%S",
			[(Sn, @hd::start)]));
		newsect := 1;
	end
	else do
		Ssn := Ssn+1;
		writes(str.nformat(MAXWORD*6, buf, "%D.%D\s%S", [(Sn, Ssn,
			@hd::start)]));
		newsect := 0;
	end
	newline();
	saveheader(buf);
end

writehead(s) ie (DoHTML)
		html_writehead(s);
	else
		sd_writehead(s);

copy(s) do
	var	b::MAXWORD*2+1;
	var	c;
	var	i, j;

	if (\s::1) return newline();
	i := 2;
	j := 0;
	c := 10;
	while (s::i) do
		if (j >= MAXWORD*2-8)
			error("preformatted line too long");
		ie (s::i = '\t') do
			while (1) do
				b::j := '\s'; j := j+1;
				c := c+1;
				if (c mod 8 = 0) leave;
			end
		end
		else do
			b::j := s::i; j := j+1;
			c := c+1;
		end
		i := i+1;
	end
	b::j := 0;
	dowrite(b);
	newline();
end

mainheader() do
	var	hd::MAXWORD, k, ch;

	ch := In.rdch();
	while (chr.space(ch) /\ \In.eof()) ch := In.rdch();
	if (In.eof()) return;
	hd::0 := ch;
	k := In.reads(@hd::1, MAXWORD)+1;
	k := trim(hd, k);
	str.ncopy(MAXWORD, Title, hd);
	if (DoTOC \/ DoScatter) return;
	if (DoFrame) printhead(KeywordFd >= 0);
	if (DoTitle) printtitle();
end

tabledata() ie (DoColor) do var buf::MAXWORD;
		ie (Firstrow /\ DoTblHeads)
			writes(str.nformat(MAXWORD, buf,
				" <TH bgcolor=\"%S\">", [(Cth)]));
		else
			writes(str.nformat(MAXWORD, buf,
				" <TD bgcolor=\"%S\">", [(Ctd)]));
	end
	else
		writes(" <TD>");

html_tablerow(data, k) do var i, s, inb::2, outb::10;
	inb::1 := 0;
	writes("<TR>"); newline();
	s := 0;
	i := 0;
	while (i < k) do
		ie (s = 0) do
			tabledata();
			s := 1;
		end
		else ie (s = 1) do
			ie (data::i = '\t') do
				ie (Firstrow /\ DoColor)
					writes(DoTblHeads-> "</TH>": "</TD>");
				else
					writes("</TD>");
				newline();
				s := 2;
			end
			else do
				ie (	(data::(i-1) = '\t' \/ i = 0) /\
					(data::(i+1) = '\t' \/ i = k-1) /\
					data::i = '.'
				) do
					writes("&nbsp;");
				end
				else do
					inb::0 := data::i;
					htmlize(inb, outb);
					writes(outb);
				end
				i := i+1;
			end
		end
		else if (s = 2) do
			ie (data::i = '\t') do
				i := i+1;
			end
			else do
				tabledata();
				s := 1;
			end
		end
	end
	if (s = 1)
		ie (Firstrow /\ DoColor)
			writes(DoTblHeads-> "</TH>": "</TD>");
		else
			writes("</TD>");
	newline();
	writes("</TR>");
	Firstrow := 0;
end

sd_tablerow(data) do
	writes(data);
	newline();
end

tablerow(data, k) ie (DoHTML)
		html_tablerow(data, k);
	else
		sd_tablerow(data);

writebullet() ie (DoHTML)
		writes(" <LI>");
	else
		writes("- ");

indent() if (\DoHTML)
		writes("\s\s");

begintable() do var buf::80;
	ie (DoHTML) do
		newline();
		ie (DoColor)
			writes(str.nformat(80, buf, 
			"<TABLE bgcolor=\"#%s\" border=0 cellspacing=1 cellpadding=2>",
			[(Ctb)]));
		else
			writes("<TABLE border=1 cellspacing=0>");
		newline();
	end
	else do
		newline();
	end
	Intable := 1;
end

endtable() do
	if (DoHTML) do
		writes("</TABLE>");
		newline();
	end
	Intable := 0;
end

beginlist() do
	if (DoHTML) do
		newline();
		writes("<UL>");
	end
	Inlist := 1;
end

decl flush(1);

endlist() do
	ie (DoHTML) do
		writes("</UL>");
		newline();
	end
	else do
		flush(1);
	end
	Inlist := 0;
end

beginpara() do
	newline();
	if (DoHTML) writes("<P>");
	Inpara := 1;
end

endpara() do
	ie (DoHTML) do
		writes("</P>");
		newline();
	end
	else do
		flush(1);
	end
	Inpara := 0;
end

beginquote() do
	if (DoHTML) do
		newline();
		writes("<PRE>");
	end
	Inquote := 1;
end

endquote() do
	if (DoHTML) do
		writes("</PRE>");
		newline();
	end
	Inquote := 0;
end

standout(s) do var i, buf::MAXWORD, so, cbuf::2;
	if (s::0 = ''') return 0;
	i := 0;
	so := 0;
	cbuf::1 := 0;
	while (s::i \/ so) do
		ie (chr.upper(s::i)) do
			if (\so /\ chr.upper(s::(i+1))) do
				ie (Sfont::0)
					writes(str.nformat(MAXWORD, buf,
						"<FONT %s>", [(Sfont)]));
				else
					writes("<B>");
				so := 1;
			end
		end
		else do
			if (so) do
				writes(Sfont::0-> "</FONT>": "</B>");
				so := 0;
			end
		end
		if (s::i) do
			cbuf::0 := so>1-> chr.lcase(s::i): s::i;
			html_writes(cbuf);
			i := i+1;
		end
		if (so) so := so+1;
	end
	return 1;
end

termall() do
	ie (Intable) endtable();
	else ie (Inlist) endlist();
	else ie (Inpara) endpara();
	else if (Inquote) endquote();
end

html_print(s, k) do var i, j;
	j := 0;
	for (i=0, k+1) do
		if ((s::i = '\s' \/ \s::i) /\ j \= i) do
			if (j) writes("\s");
			s::i := 0;
			ie (DoStandout /\ DoHTML)
				if (\standout(@s::j)) dowrite(@s::j);
			else
				dowrite(@s::j);
			i := i+1;
			while (i<k /\ s::i = '\s') i := i+1;
			j := i;
		end
	end
	newline();
end

insert(s, p, k, c) do
	t.memcopy(@s::(p+1), @s::p, k-p+1);
	s::p := c;
	return k+1;
end

adjust(s, k, size) do var l, r, f;
	l := 0;
	r := k-1;
	f := -1;
	while (k < size) do
		ie (f) do
			while (r > 0 /\ s::r \= '\s') r := r-1;
			while (r > 0 /\ s::r = '\s') r := r-1;
			ie (r < 1)
				r := k-1;
			else
				k := insert(s, r+1, k, '\s');
		end
		else do
			while (l < k /\ s::l \= '\s') l := l+1;
			while (l < k /\ s::l = '\s') l := l+1;
			ie (l >= k) do
				l := 0;
			end
			else do
				k := insert(s, l, k, '\s');
				l := l+1;
			end
		end
		f := \f;
	end
end

flush(break) do
	if (\Len) return;
	if (\break /\ DoRefmt > 1) adjust(Linebuf, Len, LineLength);
	writes(Linebuf);
	newline();
	Len := 0;
end

reformat(s, k) do var n;
	n := Len-> k+1: k;
	if (Len + n >= LineLength) flush(0);
	if (Len) do
		Linebuf::Len := '\s';
		Len := Len+1;
	end
	t.memcopy(@Linebuf::Len, s, k);
	Len := Len+k;
	Linebuf::Len := 0;
end

sd_print(s, k) do var i, j;
	if (\DoRefmt \/ Inlist) do
		writes(s);
		newline();
		return;
	end
	i := 0;
	while (i < k) do
		while (s::i = '\s') i := i+1;
		if (i >= k) return;
		j := i;
		while (i<k /\ s::i \= '\s') i := i+1;
		reformat(@s::j, i-j);
	end
end

printtext(s, k) ie (DoHTML)
		html_print(s, k);
	else
		sd_print(s, k);

writequote(s) ie (DoHTML) do
		dowrite(s);
	end
	else do
		ie (s::0) do
			writes("|\s");
			dowrite(s);
		end
		else do
			dowrite("|");
		end
	end

sdfmt() do var nl, brk, k, lbuf::MAXWORD;
	mainheader();
	if (In.eof()) return;
	Intable := 0;
	Inlist := 0;
	Inpara := 0;
	Inquote := 0;
	nl := 0;
	brk := 0;
	while (1) do
		k := In.reads(lbuf, MAXWORD);
		if (In.eof()) leave;
		k := trim(lbuf, k);
		if (brk) do
			termall();
			writehead(lbuf);
			brk := 0;
			nl := 0;
			loop;
		end
		ie (k = 0) do
			termall();
		end
		else ie (str.scan(lbuf, '\t') \= -1) do
			if (\Intable) do
				termall();
				begintable();
				Firstrow := 1;
			end
		end
		else ie (lbuf::0 = '|' /\ (lbuf::1 = '\s' \/ \lbuf::1)) do
			if (\Inquote) do
				termall();
				beginquote();
				newline();
			end
		end
		else ie (lbuf::0 = '-' /\ lbuf::1 = '\s') do
			if (\Inlist) do
				termall();
				beginlist();
				newline();
			end
		end
		else if (\Inlist) do
			if (\Inpara) do
				termall();
				beginpara();
			end
			Inpara := 1;
		end
		ie (Inpara) do
			printtext(lbuf, k);
		end
		else ie (Intable) do
			tablerow(lbuf, k);
			Firstrow := 0;
		end
		else ie (Inlist) do
			ie (lbuf::0 = '-') do
				writebullet();
				printtext(@lbuf::2, k-2);
			end
			else do
				indent();
				printtext(lbuf, k);
			end
		end
		else if (Inquote) do
			writequote(@lbuf::2);
			newline();
		end
		nl := k=0-> nl+1: 0;
		if (nl > 1) brk := 1;
	end
	if (\DoHTML) do
		flush(1);
		newline();
	end
end

writeindex() if (Hx) do var i, j, n, h, s, u, buf::100;
	newline();
	writes("<!-- BEGIN TOC -->"); newline();
	writes("<P>"); newline();
	h := Hn;
	s := 0;
	u := 0;
	for (i=0, Hx) do
		writes(DoScatter-> "<A href=\"": "<A href=\"#");
		j := 0; n := 0;
		while (Headers[i]::j \= '\s') do
			if (Headers[i]::j = '.') n := n+1;
			if (\DoScatter) do
				ie (DoShortIDs)
					if (Headers[i]::j \= '.')
						Out.wrch(Headers[i]::j);
				else
					Out.wrch(Headers[i]::j);
			end
			j := j+1;
		end
		if (DoScatter) do
			ie (n < 2) do
				s := s + 1;
				u := 0;
			end
			else do
				u := u + 1;
			end
			ie (DoScatter = 1)
				ie (u)
					writes(str.nformat(100, buf,
						"%2:0D%2:0D00.html#%D",
							[(Hn, s, u)]));
				else
					writes(str.nformat(100, buf,
						"%2:0D%2:0D00.html",
							[(Hn, s)]));
			else
				writes(str.nformat(100, buf,
					"%2:0D%2:0D%2:0D.html",
					[(Hn, s, u)]));
		end
		writes("\">");
		if (n < 2) writes("<B>");
		dowrite(Headers[i]);
		if (n < 2) writes("</B>");
		writes("</A><BR>");
		newline();
	end
	writes("</P>"); newline();
	writes("<!-- END TOC -->");
	newline();
end

usage() do
	mnewline();
	mwrites("Usage: ");
	mwrites("txsdfmt [-befior[r[r]]stx[x]HS] [-h#] [-l#] [-k file]");
	mnewline();
	mwrites("              [-C b t l] [-F attr] [-T h d b] [-Y style]");
	mnewline();
	mnewline();
	mwrites("-b        print banner (generator message)");
	mnewline();
	mwrites("-e        generate embeddable document (without TOC)");
	mnewline();
	mwrites("-f        generate HTML frame (HTML,HEAD,BODY,etc)");
	mnewline();
	mwrites("-h N      start H2 numbering at N");
	mnewline();
	mwrites("-i        indent document (with -r)");
	mnewline();
	mwrites("-k file   generate HTML keywords from 'file'");
	mnewline();
	mwrites("-l len    set line length to len (with -r, default=64)");
	mnewline();
	mwrites("-o        omit the first subheading of each section");
	mnewline();
	mwrites("-r        re-format SD, do not generate HTML");
	mnewline();
	mwrites("-rr[r]    more -r's re-format harder");
	mnewline();
	mwrites("-s        generate short section anchor names");
	mnewline();
	mwrites("-t        generate title (H1 heading)");
	mnewline();
	mwrites("-x        explode the document into separate HTML files");
	mnewline();
	mwrites("-xx       explode harder");
	mnewline();
	mwrites("-C B T L  use these colors for Background, Text, Links");
	mnewline();
	mwrites("-F attr   use this font Attribute for standout text (-S)");
	mnewline();
	mwrites("-H        generate table headings");
	mnewline();
	mwrites("-S        convert all-upper-case words to standout");
	mnewline();
	mwrites("-T H D B  use these colors for table Headings, Data, Borders");
	mnewline();
	mwrites("-Y style  add this CSS Style element to HTML header");
	mnewline();
	mnewline();
	halt 1;
end

getarg(buf, pp, pi) do var arg;
	pi[0] := pi[0] + 1;
	if (\buf::pi[0]) do
		pp[0] := pp[0] + 1;
		if (t.getarg(pp[0], buf, 256) < 1) usage();
		pi[0] := 0;
	end
	arg := @buf::pi[0];
	pi[0] := str.length(buf)-1;
	return arg;
end

addstyle(s) do var k;
	k := str.length(s) + 1;
	if (Stp) Styles::(Stp-1) := '\n';
	if (k + Stp >= STYLEBUFLEN) error("too many styles");
	t.memcopy(@Styles::Stp, s, k);
	Stp := Stp+k;
end

do var abuf::256, p, i;
	init();
	p := 1;
	while (t.getarg(p, abuf, 256) > 0) do
		if (abuf::0 \= '-') usage();
		i := 1;
		while (abuf::i) do
			ie (abuf::i = 'b') do
				DoBanner := 1;
			end
			else ie (abuf::i = 'C') do
				str.ncopy(MAXWORD, Cbg, getarg(abuf, @p, @i));
				str.ncopy(MAXWORD, Ctxt, getarg(abuf, @p, @i));
				str.ncopy(MAXWORD, Clnk, getarg(abuf, @p, @i));
				DoColor := 1;
				DoFrame := 1;
			end
			else ie (abuf::i = 'e') do
				DoTOC := 0;
				DoFrame := 0;
			end
			else ie (abuf::i = 'f') do
				DoFrame := 1;
			end
			else ie (abuf::i = 'h') do
				str.ncopy(MAXWORD, Tmp, getarg(abuf, @p, @i));
				Hn := str.strtonum(Tmp, 10, 0);
				if (Hn < 1) usage();
			end
			else ie (abuf::i = 'i') do
				DoIndent := 1;
			end
			else ie (abuf::i = 'k') do
				KeywordFd := t.open(getarg(abuf, @p, @i),
					T3X.OREAD);
				if (KeywordFd < 0)
					error("could not open keyword file");
			end
			else ie (abuf::i = 'l') do
				LineLength := str.strtonum(getarg(abuf,
					@p, @i), 10, 0);
			end
			else ie (abuf::i = 'o') do
				Do1omit := 1;
			end
			else ie (abuf::i = 'r') do
				if (\DoHTML) DoRefmt := DoRefmt + 1;
				DoHTML := 0;
				DoTitle := 1;
			end
			else ie (abuf::i = 's') do
				DoShortIDs := 1;
			end
			else ie (abuf::i = 't') do
				DoTitle := 1;
			end
			else ie (abuf::i = 'x') do
				DoScatter := DoScatter + 1;
				DoFrame := 1;
			end
			else ie (abuf::i = 'S') do
				DoStandout := 1;
			end
			else ie (abuf::i = 'F') do
				str.ncopy(MAXWORD, Sfont,
					getarg(abuf, @p, @i));
			end
			else ie (abuf::i = 'H') do
				DoTblHeads := 1;
			end
			else ie (abuf::i = 'T') do
				str.ncopy(MAXWORD, Cth, getarg(abuf, @p, @i));
				str.ncopy(MAXWORD, Ctd, getarg(abuf, @p, @i));
				str.ncopy(MAXWORD, Ctb, getarg(abuf, @p, @i));
				DoColor := 1;
			end
			else ie (abuf::i = 'Y') do
				addstyle(getarg(abuf, @p, @i));
			end
			else do
				usage();
			end
			i := i+1;
		end
		p := p+1;
	end
	if (DoTOC /\ \DoScatter) do
		if (
		Out.open("txsdfmt.tmp", Outbuf, BUFLEN, IOSTREAM.FWRITE) = -1
		)
			error("cannot create workfile (txsdfmt.tmp)");
	end
	sdfmt();
	if (DoTOC /\ \DoScatter) do
		Out.close();
		Out.create(T3X.SYSOUT, Outbuf, BUFLEN, IOSTREAM.FWRITE);
	end
	if (DoTOC /\ \DoScatter /\ DoFrame) printhead(KeywordFd >= 0);
	if (DoTOC /\ \DoScatter /\ DoTitle) printtitle();
	if (DoTOC) do
		if (DoScatter) do
			nextfile("index.html", KeywordFd >= 0);
			if (DoTitle) printtitle();
		end
		if (DoHTML) writeindex();
		if (\DoScatter) do
			if (In.open("txsdfmt.tmp", Inbuf, BUFLEN,
				IOSTREAM.FREAD) = -1)
					error("cannot re-open txsdfmt.tmp");
			i := In.read(abuf, 256);
			while (i > 1) do
				Out.write(abuf, i);
				i := In.read(abuf, 256);
			end
		end
	end
	if (\DoScatter /\ DoFrame) printtail();
	t.remove("txsdfmt.tmp");
	fini();
end
